//
// Project NuriKit - Copyright 2023 SNU Compbio Lab.
// SPDX-License-Identifier: Apache-2.0
//

#include "nuri/core/element.h"

#include <algorithm>
#include <cstddef>
#include <cstdint>
#include <cstdlib>
#include <string_view>
#include <tuple>
#include <utility>
#include <vector>

#include <absl/log/absl_check.h>
#include <absl/strings/ascii.h>
#include <absl/strings/str_cat.h>

namespace nuri {
namespace {
  std::pair<std::int16_t, std::int16_t>
  get_period_group(const int atnum, bool &is_lanthanide_or_actinide) {
    // Dummy atom
    switch (atnum) {
    case 0:
      return { 0, 0 };
    case 1:
      return { 1, 1 };
    case 2:
      return { 1, 18 };
    default:
      break;
    }

    // 2nd - 3rd row
    if (atnum <= 18) {
      const auto div = std::div(atnum - 3, 8);
      const int period = div.quot + 2;
      int group = div.rem + 1;
      if (group >= 3) {
        group += 10;
      }
      return { period, group };
    }

    // 4th - 5th row
    if (atnum <= 54) {
      const auto div = std::div(atnum - 19, 18);
      const int period = div.quot + 4;
      const int group = div.rem + 1;
      return { period, group };
    }

    // 6th - 7th row
    const auto div = std::div(atnum - 55, 32);
    const int period = div.quot + 6;
    int group = div.rem + 1;
    if (group >= 3 && group <= 17) {
      // Lanthanides and actinides
      group = 3;
      is_lanthanide_or_actinide = true;
    } else if (group >= 18) {
      group -= 14;
    }
    return { period, group };
  }

  int get_valence_electrons(const Element &elem) {
    if (elem.lanthanide()) {
      if (elem.atomic_number() == 71) {
        // Lutetium has filled f-orbital
        return 3;
      }
      return elem.atomic_number() - 57 + 3;
    }

    if (elem.actinide()) {
      if (elem.atomic_number() == 103) {
        // Lawrencium has filled f-orbital
        return 3;
      }
      return elem.atomic_number() - 89 + 3;
    }

    if (elem.group() <= 12) {
      return elem.group();
    }

    // Helium
    if (elem.atomic_number() == 2) {
      return 2;
    }

    // Group 13-18
    return elem.group() - 10;
  }

  Element::Type get_type(const Element &elem) {
    if (elem.group() >= 17) {
      return Element::Type::kNonmetal;
    }

    switch (elem.atomic_number()) {
    case 0:
      return Element::Type::kUnknown;
    case 1:   // H
    case 6:   // C
    case 7:   // N
    case 8:   // O
    case 15:  // P
    case 16:  // S
    case 34:  // Se
      return Element::Type::kNonmetal;
    case 5:   // B
    case 14:  // Si
    case 32:  // Ge
    case 33:  // As
    case 51:  // Sb
    case 52:  // Te
    case 84:  // Po
      return Element::Type::kMetalloid;
    default:
      return Element::Type::kMetal;
    }
  }

  Element::State get_state(const Element &elem) {
    if (elem.group() == 18) {
      return Element::State::kGas;
    }

    switch (elem.atomic_number()) {
    case 0:
      return Element::State::kUnknown;
    case 35:  // Br
    case 80:  // Hg
      return Element::State::kLiquid;
    case 1:   // H
    case 7:   // N
    case 8:   // O
    case 9:   // F
    case 17:  // Cl
      return Element::State::kGas;
    default:
      return Element::State::kSolid;
    }
  }
}  // namespace

Element::Element(int atomic_number, std::string_view symbol,
                 std::string_view name, double atomic_weight, double cov_rad,
                 double vdw_rad, double eneg,
                 std::vector<Isotope> &&isotopes) noexcept
    : atomic_number_(atomic_number), flags_(static_cast<ElementFlags>(0)),
      symbol_(symbol), name_(name), atomic_weight_(atomic_weight),
      cov_rad_(cov_rad), vdw_rad_(vdw_rad), eneg_(eneg),
      isotopes_(std::move(isotopes)) {
  // GCOV_EXCL_START
  ABSL_DCHECK(!isotopes_.empty());
  // GCOV_EXCL_STOP

  bool lan_or_act = false;
  std::tie(period_, group_) = get_period_group(atomic_number_, lan_or_act);

  if (group_ < 3 || group_ > 12) {
    flags_ |= ElementFlags::kMainGroup;
  }
  if (lan_or_act) {
    flags_ |= period_ == 6 ? ElementFlags::kLanthanide
                           : ElementFlags::kActinide;
  }

  valence_electrons_ = static_cast<int16_t>(get_valence_electrons(*this));
  type_ = get_type(*this);
  state_ = get_state(*this);

  std::vector<Isotope>::iterator it;
  if (std::all_of(isotopes_.begin(), isotopes_.end(),
                  [](const Isotope &isotope) {
                    return isotope.abundance == 0.0;
                  })) {
    flags_ |= ElementFlags::kRadioactive;

    it = std::find_if(isotopes_.begin(), isotopes_.end(),
                      [&](const Isotope &isotope) {
                        // NOLINTNEXTLINE(clang-diagnostic-float-equal)
                        return isotope.mass_number == atomic_weight_;
                      });
  } else {
    it = std::max_element(isotopes_.begin(), isotopes_.end(),
                          [](const Isotope &lhs, const Isotope &rhs) {
                            return lhs.abundance < rhs.abundance;
                          });
  }

  // GCOV_EXCL_START
  ABSL_DCHECK(it != isotopes_.end());
  // GCOV_EXCL_STOP

  major_isotope_ = &*it;
}

// Format:
// { Z, Symbol, Name, Mass, r_cov, r_vdw, EN, Isotope list }
PeriodicTable::PeriodicTable() noexcept
  : elements_ {
  // clang-format off
// Wondering why Dummy atom radii is not 0.0? See the commit message at
// https://github.com/OpenChemistry/avogadrolibs/commit/c9c0e81e10384146f07f8a808a9103f11c91d46f
{ 0, "Xx", "Dummy", 0.0, 0.16, 0.60, -1.0, {
  { 0, 0, 0.0,           0.0 },
  { 0, 1, 1.00866491595, 0.0 },
} },
// 1st row
{ 1, "H", "Hydrogen", 1.0080, 0.31, 1.20, 2.20, {
  { 1, 1, 1.007825031898, 0.99985578 },
  { 1, 2, 2.014101777844, 0.00014578 },
  { 1, 3, 3.01604928132,  0.0        },
  { 1, 4, 4.026431867,    0.0        },
  { 1, 5, 5.035311492,    0.0        },
  { 1, 6, 6.044955437,    0.0        },
  { 1, 7, 7.052749,       0.0        },
} },
{ 2, "He", "Helium", 4.002602, 0.28, 1.43, -1.0, {
  { 2,  3,  3.01602932197, 0.000002 },
  { 2,  4,  4.00260325413, 0.999998 },
  { 2,  5,  5.012057224,   0.0      },
  { 2,  6,  6.018885889,   0.0      },
  { 2,  7,  7.027990652,   0.0      },
  { 2,  8,  8.033934388,   0.0      },
  { 2,  9,  9.043946414,   0.0      },
  { 2, 10, 10.052815306,   0.0      },
} },
// 2nd row
{ 3, "Li", "Lithium", 6.94, 1.28, 2.12, 0.98, {
  { 3,  3,  3.030775,      0.0       },
  { 3,  4,  4.027185561,   0.0       },
  { 3,  5,  5.012537800,   0.0       },
  { 3,  6,  6.01512288742, 0.0485171 },
  { 3,  7,  7.01600343426, 0.9515171 },
  { 3,  8,  8.022486244,   0.0       },
  { 3,  9,  9.026790191,   0.0       },
  { 3, 10, 10.035483453,   0.0       },
  { 3, 11, 11.043723581,   0.0       },
  { 3, 12, 12.052613942,   0.0       },
  { 3, 13, 13.061171503,   0.0       },
} },
{ 4, "Be", "Beryllium", 9.0121831, 0.96, 1.98, 1.57, {
  { 4,  5,  5.039870,    0.0 },
  { 4,  6,  6.019726409, 0.0 },
  { 4,  7,  7.016928714, 0.0 },
  { 4,  8,  8.005305102, 0.0 },
  { 4,  9,  9.012183062, 1.0 },
  { 4, 10, 10.013534692, 0.0 },
  { 4, 11, 11.021661080, 0.0 },
  { 4, 12, 12.026922082, 0.0 },
  { 4, 13, 13.036134506, 0.0 },
  { 4, 14, 14.042892920, 0.0 },
  { 4, 15, 15.053490215, 0.0 },
  { 4, 16, 16.061672036, 0.0 },
} },
{ 5, "B", "Boron", 10.81, 0.84, 1.91, 2.04, {
  { 5,  6,  6.050800,    0.0      },
  { 5,  7,  7.029712000, 0.0      },
  { 5,  8,  8.024607315, 0.0      },
  { 5,  9,  9.013329645, 0.0      },
  { 5, 10, 10.012936862, 0.196544 },
  { 5, 11, 11.009305166, 0.803544 },
  { 5, 12, 12.014352638, 0.0      },
  { 5, 13, 13.017779981, 0.0      },
  { 5, 14, 14.025404010, 0.0      },
  { 5, 15, 15.031087023, 0.0      },
  { 5, 16, 16.039841045, 0.0      },
  { 5, 17, 17.046931399, 0.0      },
  { 5, 18, 18.055601683, 0.0      },
  { 5, 19, 19.064166000, 0.0      },
  { 5, 20, 20.074505644, 0.0      },
  { 5, 21, 21.084147485, 0.0      },
} },
{ 6, "C", "Carbon", 12.011, 0.76, 1.77, 2.55, {
  { 6,  8,  8.037643039,   0.0     },
  { 6,  9,  9.031037202,   0.0     },
  { 6, 10, 10.016853217,   0.0     },
  { 6, 11, 11.011432597,   0.0     },
  { 6, 12, 12.0000000,     0.98946 },
  { 6, 13, 13.00335483534, 0.01066 },
  { 6, 14, 14.00324198862, 0.0     },
  { 6, 15, 15.010599256,   0.0     },
  { 6, 16, 16.014701255,   0.0     },
  { 6, 17, 17.022578650,   0.0     },
  { 6, 18, 18.026751930,   0.0     },
  { 6, 19, 19.034797594,   0.0     },
  { 6, 20, 20.040261732,   0.0     },
  { 6, 21, 21.049000,      0.0     },
  { 6, 22, 22.057553990,   0.0     },
  { 6, 23, 23.068890,      0.0     },
} },
{ 7, "N", "Nitrogen", 14.007, 0.71, 1.66, 3.04, {
  { 7, 10, 10.041653540,   0.0         },
  { 7, 11, 11.026157593,   0.0         },
  { 7, 12, 12.018613180,   0.0         },
  { 7, 13, 13.005738609,   0.0         },
  { 7, 14, 14.00307400425, 0.996205247 },
  { 7, 15, 15.00010889827, 0.003795247 },
  { 7, 16, 16.006101925,   0.0         },
  { 7, 17, 17.008448876,   0.0         },
  { 7, 18, 18.014077563,   0.0         },
  { 7, 19, 19.017022389,   0.0         },
  { 7, 20, 20.023367295,   0.0         },
  { 7, 21, 21.027087573,   0.0         },
  { 7, 22, 22.034100918,   0.0         },
  { 7, 23, 23.039421000,   0.0         },
  { 7, 24, 24.050390,      0.0         },
  { 7, 25, 25.060100,      0.0         },
} },
{ 8, "O", "Oxygen", 15.999, 0.66, 1.50, 3.44, {
  { 8, 11, 11.051249828,   0.0         },
  { 8, 12, 12.034367726,   0.0         },
  { 8, 13, 13.024815435,   0.0         },
  { 8, 14, 14.008596706,   0.0         },
  { 8, 15, 15.003065636,   0.0         },
  { 8, 16, 15.99491461926, 0.9975711   },
  { 8, 17, 16.99913175595, 0.000383596 },
  { 8, 18, 17.99915961214, 0.002045102 },
  { 8, 19, 19.003577969,   0.0         },
  { 8, 20, 20.004075357,   0.0         },
  { 8, 21, 21.008654948,   0.0         },
  { 8, 22, 22.009965744,   0.0         },
  { 8, 23, 23.015696686,   0.0         },
  { 8, 24, 24.019861000,   0.0         },
  { 8, 25, 25.029338919,   0.0         },
  { 8, 26, 26.037210155,   0.0         },
  { 8, 27, 27.047955,      0.0         },
  { 8, 28, 28.055910,      0.0         },
} },
{ 9, "F", "Fluorine", 18.998403162, 0.57, 1.46, 3.98, {
  { 9, 13, 13.045121,      0.0 },
  { 9, 14, 14.034315196,   0.0 },
  { 9, 15, 15.017785139,   0.0 },
  { 9, 16, 16.011460278,   0.0 },
  { 9, 17, 17.002095237,   0.0 },
  { 9, 18, 18.000937324,   0.0 },
  { 9, 19, 18.99840316207, 1.0 },
  { 9, 20, 19.999981252,   0.0 },
  { 9, 21, 20.999948893,   0.0 },
  { 9, 22, 22.002998812,   0.0 },
  { 9, 23, 23.003526875,   0.0 },
  { 9, 24, 24.008099370,   0.0 },
  { 9, 25, 25.012167727,   0.0 },
  { 9, 26, 26.020048065,   0.0 },
  { 9, 27, 27.026981897,   0.0 },
  { 9, 28, 28.035860448,   0.0 },
  { 9, 29, 29.043103000,   0.0 },
  { 9, 30, 30.052561,      0.0 },
  { 9, 31, 31.061023,      0.0 },
} },
{ 10, "Ne", "Neon", 20.1797, 0.58, 1.58, -1.0, {
  { 10, 15, 15.043172977,   0.0    },
  { 10, 16, 16.025750860,   0.0    },
  { 10, 17, 17.017713962,   0.0    },
  { 10, 18, 18.005708696,   0.0    },
  { 10, 19, 19.001880906,   0.0    },
  { 10, 20, 19.99244017525, 0.9048 },
  { 10, 21, 20.993846685,   0.0027 },
  { 10, 22, 21.991385113,   0.0925 },
  { 10, 23, 22.994466905,   0.0    },
  { 10, 24, 23.993610649,   0.0    },
  { 10, 25, 24.997814797,   0.0    },
  { 10, 26, 26.000516496,   0.0    },
  { 10, 27, 27.007569462,   0.0    },
  { 10, 28, 28.012130767,   0.0    },
  { 10, 29, 29.019753000,   0.0    },
  { 10, 30, 30.024992235,   0.0    },
  { 10, 31, 31.033474816,   0.0    },
  { 10, 32, 32.039720,      0.0    },
  { 10, 33, 33.049523,      0.0    },
  { 10, 34, 34.056728,      0.0    },
} },
// 3rd row
{ 11, "Na", "Sodium", 22.98976928, 1.66, 2.50, 0.93, {
  { 11, 17, 17.037273000,   0.0 },
  { 11, 18, 18.026879388,   0.0 },
  { 11, 19, 19.013880264,   0.0 },
  { 11, 20, 20.007354301,   0.0 },
  { 11, 21, 20.997654459,   0.0 },
  { 11, 22, 21.994437547,   0.0 },
  { 11, 23, 22.98976928195, 1.0 },
  { 11, 24, 23.990963012,   0.0 },
  { 11, 25, 24.989953974,   0.0 },
  { 11, 26, 25.992634649,   0.0 },
  { 11, 27, 26.994076408,   0.0 },
  { 11, 28, 27.998939000,   0.0 },
  { 11, 29, 29.002877091,   0.0 },
  { 11, 30, 30.009097931,   0.0 },
  { 11, 31, 31.013146654,   0.0 },
  { 11, 32, 32.020011024,   0.0 },
  { 11, 33, 33.025529000,   0.0 },
  { 11, 34, 34.034010000,   0.0 },
  { 11, 35, 35.040614,      0.0 },
  { 11, 36, 36.049279,      0.0 },
  { 11, 37, 37.057042,      0.0 },
  { 11, 38, 38.066458,      0.0 },
  { 11, 39, 39.075123,      0.0 },
} },
{ 12, "Mg", "Magnesium", 24.305, 1.41, 2.51, 1.31, {
  { 12, 19, 19.034179920, 0.0       },
  { 12, 20, 20.018763075, 0.0       },
  { 12, 21, 21.011705764, 0.0       },
  { 12, 22, 21.999570597, 0.0       },
  { 12, 23, 22.994123768, 0.0       },
  { 12, 24, 23.985041689, 0.7896549 },
  { 12, 25, 24.985836966, 0.1001113 },
  { 12, 26, 25.982592972, 0.1102538 },
  { 12, 27, 26.984340647, 0.0       },
  { 12, 28, 27.983875426, 0.0       },
  { 12, 29, 28.988607163, 0.0       },
  { 12, 30, 29.990465454, 0.0       },
  { 12, 31, 30.996648232, 0.0       },
  { 12, 32, 31.999110138, 0.0       },
  { 12, 33, 33.005327862, 0.0       },
  { 12, 34, 34.008935455, 0.0       },
  { 12, 35, 35.016790000, 0.0       },
  { 12, 36, 36.021879000, 0.0       },
  { 12, 37, 37.030286265, 0.0       },
  { 12, 38, 38.036580,    0.0       },
  { 12, 39, 39.045921,    0.0       },
  { 12, 40, 40.053194,    0.0       },
  { 12, 41, 41.062373,    0.0       },
} },
{ 13, "Al", "Aluminium", 26.9815384, 1.21, 2.25, 1.61, {
  { 13, 21, 21.029082,    0.0 },
  { 13, 22, 22.019540,    0.0 },
  { 13, 23, 23.007244351, 0.0 },
  { 13, 24, 23.999947598, 0.0 },
  { 13, 25, 24.990428308, 0.0 },
  { 13, 26, 25.986891876, 0.0 },
  { 13, 27, 26.981538408, 1.0 },
  { 13, 28, 27.981910009, 0.0 },
  { 13, 29, 28.980453164, 0.0 },
  { 13, 30, 29.982969171, 0.0 },
  { 13, 31, 30.983949754, 0.0 },
  { 13, 32, 31.988084338, 0.0 },
  { 13, 33, 32.990877685, 0.0 },
  { 13, 34, 33.996781924, 0.0 },
  { 13, 35, 34.999759816, 0.0 },
  { 13, 36, 36.006388000, 0.0 },
  { 13, 37, 37.010531000, 0.0 },
  { 13, 38, 38.017681,    0.0 },
  { 13, 39, 39.023070,    0.0 },
  { 13, 40, 40.030940,    0.0 },
  { 13, 41, 41.037134,    0.0 },
  { 13, 42, 42.045078,    0.0 },
  { 13, 43, 43.051820,    0.0 },
} },
{ 14, "Si", "Silicon", 28.085, 1.11, 2.19, 1.90, {
  { 14, 22, 22.036114,      0.0        },
  { 14, 23, 23.025711,      0.0        },
  { 14, 24, 24.011535430,   0.0        },
  { 14, 25, 25.004108798,   0.0        },
  { 14, 26, 25.992333818,   0.0        },
  { 14, 27, 26.986704687,   0.0        },
  { 14, 28, 27.97692653442, 0.92254537 },
  { 14, 29, 28.97649466434, 0.0467216  },
  { 14, 30, 29.973770137,   0.03073521 },
  { 14, 31, 30.975363196,   0.0        },
  { 14, 32, 31.974151538,   0.0        },
  { 14, 33, 32.977976964,   0.0        },
  { 14, 34, 33.978538045,   0.0        },
  { 14, 35, 34.984550111,   0.0        },
  { 14, 36, 35.986649271,   0.0        },
  { 14, 37, 36.992945191,   0.0        },
  { 14, 38, 37.995523000,   0.0        },
  { 14, 39, 39.002491000,   0.0        },
  { 14, 40, 40.006083641,   0.0        },
  { 14, 41, 41.014171,      0.0        },
  { 14, 42, 42.018078,      0.0        },
  { 14, 43, 43.026119,      0.0        },
  { 14, 44, 44.031466,      0.0        },
  { 14, 45, 45.039818,      0.0        },
} },
{ 15, "P", "Phosphorus", 30.973761998, 1.07, 1.90, 2.19, {
  { 15, 24, 24.036522,      0.0 },
  { 15, 25, 25.021675,      0.0 },
  { 15, 26, 26.011780,      0.0 },
  { 15, 27, 26.999292499,   0.0 },
  { 15, 28, 27.992326460,   0.0 },
  { 15, 29, 28.981800368,   0.0 },
  { 15, 30, 29.978313490,   0.0 },
  { 15, 31, 30.97376199768, 1.0 },
  { 15, 32, 31.973907643,   0.0 },
  { 15, 33, 32.971725692,   0.0 },
  { 15, 34, 33.973645886,   0.0 },
  { 15, 35, 34.973314045,   0.0 },
  { 15, 36, 35.978259610,   0.0 },
  { 15, 37, 36.979606942,   0.0 },
  { 15, 38, 37.984303105,   0.0 },
  { 15, 39, 38.986285865,   0.0 },
  { 15, 40, 39.991262221,   0.0 },
  { 15, 41, 40.994654000,   0.0 },
  { 15, 42, 42.001172140,   0.0 },
  { 15, 43, 43.005411,      0.0 },
  { 15, 44, 44.011927,      0.0 },
  { 15, 45, 45.017134,      0.0 },
  { 15, 46, 46.024520,      0.0 },
  { 15, 47, 47.030929,      0.0 },
} },
{ 16, "S", "Sulfur", 32.06, 1.05, 1.89, 2.58, {
  { 16, 26, 26.029716,      0.0        },
  { 16, 27, 27.018777,      0.0        },
  { 16, 28, 28.004372762,   0.0        },
  { 16, 29, 28.996678000,   0.0        },
  { 16, 30, 29.984906770,   0.0        },
  { 16, 31, 30.979557002,   0.0        },
  { 16, 32, 31.97207117354, 0.9485255  },
  { 16, 33, 32.97145890862, 0.007632   },
  { 16, 34, 33.967867011,   0.04365235 },
  { 16, 35, 34.969032321,   0.0        },
  { 16, 36, 35.967080692,   0.00015817 },
  { 16, 37, 36.971125500,   0.0        },
  { 16, 38, 37.971163300,   0.0        },
  { 16, 39, 38.975133850,   0.0        },
  { 16, 40, 39.975482561,   0.0        },
  { 16, 41, 40.979593451,   0.0        },
  { 16, 42, 41.981065100,   0.0        },
  { 16, 43, 42.986907635,   0.0        },
  { 16, 44, 43.990118846,   0.0        },
  { 16, 45, 44.996414,      0.0        },
  { 16, 46, 46.000687,      0.0        },
  { 16, 47, 47.007730,      0.0        },
  { 16, 48, 48.013301,      0.0        },
  { 16, 49, 49.021891,      0.0        },
} },
{ 17, "Cl", "Chlorine", 35.45, 1.02, 1.82, 3.16, {
  { 17, 28, 28.030349,    0.0    },
  { 17, 29, 29.015053,    0.0    },
  { 17, 30, 30.005018333, 0.0    },
  { 17, 31, 30.992448097, 0.0    },
  { 17, 32, 31.985684605, 0.0    },
  { 17, 33, 32.977451988, 0.0    },
  { 17, 34, 33.973762490, 0.0    },
  { 17, 35, 34.968852694, 0.7582 },
  { 17, 36, 35.968306822, 0.0    },
  { 17, 37, 36.965902573, 0.2422 },
  { 17, 38, 37.968010408, 0.0    },
  { 17, 39, 38.968008151, 0.0    },
  { 17, 40, 39.970415466, 0.0    },
  { 17, 41, 40.970684525, 0.0    },
  { 17, 42, 41.973342000, 0.0    },
  { 17, 43, 42.974063700, 0.0    },
  { 17, 44, 43.978014918, 0.0    },
  { 17, 45, 44.980394353, 0.0    },
  { 17, 46, 45.985254926, 0.0    },
  { 17, 47, 46.989715,    0.0    },
  { 17, 48, 47.995405,    0.0    },
  { 17, 49, 49.000794,    0.0    },
  { 17, 50, 50.008266,    0.0    },
  { 17, 51, 51.015341,    0.0    },
  { 17, 52, 52.024004,    0.0    },
} },
{ 18, "Ar", "Argon", 39.95, 1.06, 1.83, -1.0, {
  { 18, 29, 29.040761,      0.0        },
  { 18, 30, 30.023694,      0.0        },
  { 18, 31, 31.012158,      0.0        },
  { 18, 32, 31.997637824,   0.0        },
  { 18, 33, 32.989925545,   0.0        },
  { 18, 34, 33.980270092,   0.0        },
  { 18, 35, 34.975257719,   0.0        },
  { 18, 36, 35.967545106,   0.00333621 },
  { 18, 37, 36.966776301,   0.0        },
  { 18, 38, 37.962732102,   0.0006297  },
  { 18, 39, 38.964313037,   0.0        },
  { 18, 40, 39.96238312204, 0.99603525 },
  { 18, 41, 40.964500570,   0.0        },
  { 18, 42, 41.963045737,   0.0        },
  { 18, 43, 42.965636056,   0.0        },
  { 18, 44, 43.964923814,   0.0        },
  { 18, 45, 44.968039731,   0.0        },
  { 18, 46, 45.968039244,   0.0        },
  { 18, 47, 46.972767112,   0.0        },
  { 18, 48, 47.976001000,   0.0        },
  { 18, 49, 48.981685,      0.0        },
  { 18, 50, 49.985797,      0.0        },
  { 18, 51, 50.993033,      0.0        },
  { 18, 52, 51.998519,      0.0        },
  { 18, 53, 53.007290,      0.0        },
  { 18, 54, 54.013484,      0.0        },
} },
// 4th row
{ 19, "K", "Potassium", 39.0983, 2.03, 2.73, 0.82, {
  { 19, 31, 31.036780,      0.0      },
  { 19, 32, 32.023607,      0.0      },
  { 19, 33, 33.008095,      0.0      },
  { 19, 34, 33.998690,      0.0      },
  { 19, 35, 34.988005406,   0.0      },
  { 19, 36, 35.981301887,   0.0      },
  { 19, 37, 36.973375890,   0.0      },
  { 19, 38, 37.969081114,   0.0      },
  { 19, 39, 38.96370648482, 0.932581 },
  { 19, 40, 39.963998165,   0.000117 },
  { 19, 41, 40.96182525611, 0.067302 },
  { 19, 42, 41.962402305,   0.0      },
  { 19, 43, 42.960734701,   0.0      },
  { 19, 44, 43.961586984,   0.0      },
  { 19, 45, 44.960691491,   0.0      },
  { 19, 46, 45.961981584,   0.0      },
  { 19, 47, 46.961661612,   0.0      },
  { 19, 48, 47.965341184,   0.0      },
  { 19, 49, 48.968210753,   0.0      },
  { 19, 50, 49.972380015,   0.0      },
  { 19, 51, 50.975828664,   0.0      },
  { 19, 52, 51.981602000,   0.0      },
  { 19, 53, 52.986800000,   0.0      },
  { 19, 54, 53.994471,      0.0      },
  { 19, 55, 55.000505,      0.0      },
  { 19, 56, 56.008567,      0.0      },
  { 19, 57, 57.015169,      0.0      },
  { 19, 58, 58.023543,      0.0      },
  { 19, 59, 59.030864,      0.0      },
} },
{ 20, "Ca", "Calcium", 40.078, 1.76, 2.62, 1.00, {
  { 20, 33, 33.033312,    0.0     },
  { 20, 34, 34.015985,    0.0     },
  { 20, 35, 35.005572,    0.0     },
  { 20, 36, 35.993074388, 0.0     },
  { 20, 37, 36.985897849, 0.0     },
  { 20, 38, 37.976319223, 0.0     },
  { 20, 39, 38.970710811, 0.0     },
  { 20, 40, 39.962590850, 0.96941 },
  { 20, 41, 40.962277905, 0.0     },
  { 20, 42, 41.958617780, 0.00647 },
  { 20, 43, 42.958766381, 0.00135 },
  { 20, 44, 43.955481489, 0.02086 },
  { 20, 45, 44.956186270, 0.0     },
  { 20, 46, 45.953687726, 0.00004 },
  { 20, 47, 46.954541134, 0.0     },
  { 20, 48, 47.952522654, 0.00187 },
  { 20, 49, 48.955662625, 0.0     },
  { 20, 50, 49.957499215, 0.0     },
  { 20, 51, 50.960995663, 0.0     },
  { 20, 52, 51.963213646, 0.0     },
  { 20, 53, 52.968451000, 0.0     },
  { 20, 54, 53.972989000, 0.0     },
  { 20, 55, 54.979978000, 0.0     },
  { 20, 56, 55.985496000, 0.0     },
  { 20, 57, 56.992958,    0.0     },
  { 20, 58, 57.998357,    0.0     },
  { 20, 59, 59.006237,    0.0     },
  { 20, 60, 60.011809,    0.0     },
  { 20, 61, 61.020408,    0.0     },
} },
{ 21, "Sc", "Scandium", 44.955907, 1.70, 2.58, 1.36, {
  { 21, 35, 35.029093,    0.0 },
  { 21, 36, 36.017338,    0.0 },
  { 21, 37, 37.004058,    0.0 },
  { 21, 38, 37.995438,    0.0 },
  { 21, 39, 38.984784953, 0.0 },
  { 21, 40, 39.977967275, 0.0 },
  { 21, 41, 40.969251163, 0.0 },
  { 21, 42, 41.965516686, 0.0 },
  { 21, 43, 42.961150425, 0.0 },
  { 21, 44, 43.959402818, 0.0 },
  { 21, 45, 44.955907051, 1.0 },
  { 21, 46, 45.955167034, 0.0 },
  { 21, 47, 46.952402444, 0.0 },
  { 21, 48, 47.952222903, 0.0 },
  { 21, 49, 48.950013159, 0.0 },
  { 21, 50, 49.952187437, 0.0 },
  { 21, 51, 50.953568838, 0.0 },
  { 21, 52, 51.956496170, 0.0 },
  { 21, 53, 52.958379173, 0.0 },
  { 21, 54, 53.963029359, 0.0 },
  { 21, 55, 54.966889637, 0.0 },
  { 21, 56, 55.972607611, 0.0 },
  { 21, 57, 56.977048000, 0.0 },
  { 21, 58, 57.983382000, 0.0 },
  { 21, 59, 58.988374000, 0.0 },
  { 21, 60, 59.995115,    0.0 },
  { 21, 61, 61.000537,    0.0 },
  { 21, 62, 62.007848,    0.0 },
  { 21, 63, 63.014031,    0.0 },
} },
{ 22, "Ti", "Titanium", 47.867, 1.60, 2.46, 1.54, {
  { 22, 37, 37.027021,    0.0    },
  { 22, 38, 38.012206,    0.0    },
  { 22, 39, 39.002684,    0.0    },
  { 22, 40, 39.990345146, 0.0    },
  { 22, 41, 40.983148000, 0.0    },
  { 22, 42, 41.973049369, 0.0    },
  { 22, 43, 42.968528420, 0.0    },
  { 22, 44, 43.959689936, 0.0    },
  { 22, 45, 44.958120758, 0.0    },
  { 22, 46, 45.952626356, 0.0825 },
  { 22, 47, 46.951757491, 0.0744 },
  { 22, 48, 47.947940677, 0.7372 },
  { 22, 49, 48.947864391, 0.0541 },
  { 22, 50, 49.944785622, 0.0518 },
  { 22, 51, 50.946609468, 0.0    },
  { 22, 52, 51.946883509, 0.0    },
  { 22, 53, 52.949670714, 0.0    },
  { 22, 54, 53.950892000, 0.0    },
  { 22, 55, 54.955091000, 0.0    },
  { 22, 56, 55.957677675, 0.0    },
  { 22, 57, 56.963068098, 0.0    },
  { 22, 58, 57.966808519, 0.0    },
  { 22, 59, 58.972217,    0.0    },
  { 22, 60, 59.976275000, 0.0    },
  { 22, 61, 60.982426,    0.0    },
  { 22, 62, 61.986903,    0.0    },
  { 22, 63, 62.993709,    0.0    },
  { 22, 64, 63.998411,    0.0    },
  { 22, 65, 65.005593,    0.0    },
} },
{ 23, "V", "Vanadium", 50.9415, 1.53, 2.42, 1.63, {
  { 23, 39, 39.024230,    0.0     },
  { 23, 40, 40.013387,    0.0     },
  { 23, 41, 41.000333,    0.0     },
  { 23, 42, 41.991820,    0.0     },
  { 23, 43, 42.980766000, 0.0     },
  { 23, 44, 43.974440977, 0.0     },
  { 23, 45, 44.965768498, 0.0     },
  { 23, 46, 45.960197389, 0.0     },
  { 23, 47, 46.954903558, 0.0     },
  { 23, 48, 47.952250900, 0.0     },
  { 23, 49, 48.948510509, 0.0     },
  { 23, 50, 49.947156681, 0.00250 },
  { 23, 51, 50.943957664, 0.99750 },
  { 23, 52, 51.944773636, 0.0     },
  { 23, 53, 52.944334940, 0.0     },
  { 23, 54, 53.946432009, 0.0     },
  { 23, 55, 54.947262000, 0.0     },
  { 23, 56, 55.950420082, 0.0     },
  { 23, 57, 56.952297000, 0.0     },
  { 23, 58, 57.956595985, 0.0     },
  { 23, 59, 58.959623343, 0.0     },
  { 23, 60, 59.964479215, 0.0     },
  { 23, 61, 60.967603529, 0.0     },
  { 23, 62, 61.972932556, 0.0     },
  { 23, 63, 62.976661000, 0.0     },
  { 23, 64, 63.982480,    0.0     },
  { 23, 65, 64.986999,    0.0     },
  { 23, 66, 65.993237,    0.0     },
  { 23, 67, 66.998128,    0.0     },
} },
{ 24, "Cr", "Chromium", 51.9961, 1.39, 2.45, 1.66, {
  { 24, 41, 41.021911,    0.0     },
  { 24, 42, 42.007579,    0.0     },
  { 24, 43, 42.997885,    0.0     },
  { 24, 44, 43.985591000, 0.0     },
  { 24, 45, 44.979050000, 0.0     },
  { 24, 46, 45.968360969, 0.0     },
  { 24, 47, 46.962894995, 0.0     },
  { 24, 48, 47.954029431, 0.0     },
  { 24, 49, 48.951333720, 0.0     },
  { 24, 50, 49.946042209, 0.04345 },
  { 24, 51, 50.944765388, 0.0     },
  { 24, 52, 51.940504714, 0.83789 },
  { 24, 53, 52.940646304, 0.09501 },
  { 24, 54, 53.938877359, 0.02365 },
  { 24, 55, 54.940836637, 0.0     },
  { 24, 56, 55.940648977, 0.0     },
  { 24, 57, 56.943612112, 0.0     },
  { 24, 58, 57.944184501, 0.0     },
  { 24, 59, 58.948345426, 0.0     },
  { 24, 60, 59.949641656, 0.0     },
  { 24, 61, 60.954378130, 0.0     },
  { 24, 62, 61.956142920, 0.0     },
  { 24, 63, 62.961161000, 0.0     },
  { 24, 64, 63.963886000, 0.0     },
  { 24, 65, 64.969608,    0.0     },
  { 24, 66, 65.973011,    0.0     },
  { 24, 67, 66.979313,    0.0     },
  { 24, 68, 67.983156,    0.0     },
  { 24, 69, 68.989662,    0.0     },
  { 24, 70, 69.993945,    0.0     },
} },
{ 25, "Mn", "Manganese", 54.938043, 1.55, 2.45, 1.55, {
  { 25, 43, 43.018647,    0.0 },
  { 25, 44, 44.008009,    0.0 },
  { 25, 45, 44.994654,    0.0 },
  { 25, 46, 45.986669000, 0.0 },
  { 25, 47, 46.975774000, 0.0 },
  { 25, 48, 47.968548760, 0.0 },
  { 25, 49, 48.959613350, 0.0 },
  { 25, 50, 49.954238157, 0.0 },
  { 25, 51, 50.948208770, 0.0 },
  { 25, 52, 51.945559090, 0.0 },
  { 25, 53, 52.941287497, 0.0 },
  { 25, 54, 53.940355772, 0.0 },
  { 25, 55, 54.938043040, 1.0 },
  { 25, 56, 55.938902816, 0.0 },
  { 25, 57, 56.938285944, 0.0 },
  { 25, 58, 57.940066643, 0.0 },
  { 25, 59, 58.940391111, 0.0 },
  { 25, 60, 59.943136574, 0.0 },
  { 25, 61, 60.944452541, 0.0 },
  { 25, 62, 61.947907384, 0.0 },
  { 25, 63, 62.949664672, 0.0 },
  { 25, 64, 63.953849369, 0.0 },
  { 25, 65, 64.956019749, 0.0 },
  { 25, 66, 65.960546833, 0.0 },
  { 25, 67, 66.963950,    0.0 },
  { 25, 68, 67.968953,    0.0 },
  { 25, 69, 68.972775,    0.0 },
  { 25, 70, 69.978046,    0.0 },
  { 25, 71, 70.982158,    0.0 },
  { 25, 72, 71.988009,    0.0 },
  { 25, 73, 72.992807,    0.0 },
} },
{ 26, "Fe", "Iron", 55.845, 1.48, 2.44, 1.83, {
  { 26, 45, 45.015467,    0.0     },
  { 26, 46, 46.001299,    0.0     },
  { 26, 47, 46.992346,    0.0     },
  { 26, 48, 47.980667000, 0.0     },
  { 26, 49, 48.973429000, 0.0     },
  { 26, 50, 49.962988000, 0.0     },
  { 26, 51, 50.956855137, 0.0     },
  { 26, 52, 51.948113364, 0.0     },
  { 26, 53, 52.945305629, 0.0     },
  { 26, 54, 53.939608189, 0.05845 },
  { 26, 55, 54.938291158, 0.0     },
  { 26, 56, 55.934935537, 0.91754 },
  { 26, 57, 56.935391950, 0.02119 },
  { 26, 58, 57.933273575, 0.00282 },
  { 26, 59, 58.934873492, 0.0     },
  { 26, 60, 59.934070249, 0.0     },
  { 26, 61, 60.936746241, 0.0     },
  { 26, 62, 61.936791809, 0.0     },
  { 26, 63, 62.940272698, 0.0     },
  { 26, 64, 63.940987761, 0.0     },
  { 26, 65, 64.945015323, 0.0     },
  { 26, 66, 65.946249958, 0.0     },
  { 26, 67, 66.950930000, 0.0     },
  { 26, 68, 67.952875,    0.0     },
  { 26, 69, 68.957918,    0.0     },
  { 26, 70, 69.960397,    0.0     },
  { 26, 71, 70.965722,    0.0     },
  { 26, 72, 71.968599,    0.0     },
  { 26, 73, 72.974246,    0.0     },
  { 26, 74, 73.977821,    0.0     },
  { 26, 75, 74.984219,    0.0     },
  { 26, 76, 75.988631,    0.0     },
} },
{ 27, "Co", "Cobalt", 58.933194, 1.29, 2.40, 1.88, {
  { 27, 47, 47.011401,    0.0 },
  { 27, 48, 48.001857,    0.0 },
  { 27, 49, 48.989501,    0.0 },
  { 27, 50, 49.981117000, 0.0 },
  { 27, 51, 50.970647000, 0.0 },
  { 27, 52, 51.963130224, 0.0 },
  { 27, 53, 52.954203278, 0.0 },
  { 27, 54, 53.948459075, 0.0 },
  { 27, 55, 54.941996416, 0.0 },
  { 27, 56, 55.939838032, 0.0 },
  { 27, 57, 56.936289819, 0.0 },
  { 27, 58, 57.935751292, 0.0 },
  { 27, 59, 58.933193524, 1.0 },
  { 27, 60, 59.933815536, 0.0 },
  { 27, 61, 60.932476031, 0.0 },
  { 27, 62, 61.934058198, 0.0 },
  { 27, 63, 62.933599630, 0.0 },
  { 27, 64, 63.935810176, 0.0 },
  { 27, 65, 64.936462071, 0.0 },
  { 27, 66, 65.939442943, 0.0 },
  { 27, 67, 66.940609625, 0.0 },
  { 27, 68, 67.944559401, 0.0 },
  { 27, 69, 68.945909000, 0.0 },
  { 27, 70, 69.950053400, 0.0 },
  { 27, 71, 70.952366923, 0.0 },
  { 27, 72, 71.956736,    0.0 },
  { 27, 73, 72.959238,    0.0 },
  { 27, 74, 73.963993,    0.0 },
  { 27, 75, 74.967192,    0.0 },
  { 27, 76, 75.972453,    0.0 },
  { 27, 77, 76.976479,    0.0 },
  { 27, 78, 77.983553,    0.0 },
} },
{ 28, "Ni", "Nickel", 58.6934, 1.24, 2.40, 1.91, {
  { 28, 48, 48.019515,    0.0      },
  { 28, 49, 49.009157,    0.0      },
  { 28, 50, 49.996286,    0.0      },
  { 28, 51, 50.987493,    0.0      },
  { 28, 52, 51.975781000, 0.0      },
  { 28, 53, 52.968190000, 0.0      },
  { 28, 54, 53.957833000, 0.0      },
  { 28, 55, 54.951329846, 0.0      },
  { 28, 56, 55.942127761, 0.0      },
  { 28, 57, 56.939791394, 0.0      },
  { 28, 58, 57.935341650, 0.680769 },
  { 28, 59, 58.934345442, 0.0      },
  { 28, 60, 59.930785129, 0.262231 },
  { 28, 61, 60.931054819, 0.011399 },
  { 28, 62, 61.928344753, 0.036345 },
  { 28, 63, 62.929669021, 0.0      },
  { 28, 64, 63.927966228, 0.009256 },
  { 28, 65, 64.930084585, 0.0      },
  { 28, 66, 65.929139333, 0.0      },
  { 28, 67, 66.931569413, 0.0      },
  { 28, 68, 67.931868787, 0.0      },
  { 28, 69, 68.935610267, 0.0      },
  { 28, 70, 69.936431300, 0.0      },
  { 28, 71, 70.940518962, 0.0      },
  { 28, 72, 71.941785924, 0.0      },
  { 28, 73, 72.946206681, 0.0      },
  { 28, 74, 73.947718,    0.0      },
  { 28, 75, 74.952506,    0.0      },
  { 28, 76, 75.954707,    0.0      },
  { 28, 77, 76.959903,    0.0      },
  { 28, 78, 77.962555,    0.0      },
  { 28, 79, 78.969769,    0.0      },
  { 28, 80, 79.975051,    0.0      },
  { 28, 81, 80.982727,    0.0      },
  { 28, 82, 81.988492,    0.0      },
} },
{ 29, "Cu", "Copper", 63.546, 1.32, 2.38, 1.90, {
  { 29, 52, 51.997982,    0.0    },
  { 29, 53, 52.985894,    0.0    },
  { 29, 54, 53.977198,    0.0    },
  { 29, 55, 54.966038000, 0.0    },
  { 29, 56, 55.958529278, 0.0    },
  { 29, 57, 56.949211686, 0.0    },
  { 29, 58, 57.944532283, 0.0    },
  { 29, 59, 58.939496713, 0.0    },
  { 29, 60, 59.937363787, 0.0    },
  { 29, 61, 60.933457375, 0.0    },
  { 29, 62, 61.932594803, 0.0    },
  { 29, 63, 62.929597119, 0.6915 },
  { 29, 64, 63.929764001, 0.0    },
  { 29, 65, 64.927789476, 0.3085 },
  { 29, 66, 65.928868804, 0.0    },
  { 29, 67, 66.927729490, 0.0    },
  { 29, 68, 67.929610887, 0.0    },
  { 29, 69, 68.929429267, 0.0    },
  { 29, 70, 69.932392078, 0.0    },
  { 29, 71, 70.932676831, 0.0    },
  { 29, 72, 71.935820306, 0.0    },
  { 29, 73, 72.936674376, 0.0    },
  { 29, 74, 73.939874860, 0.0    },
  { 29, 75, 74.941523817, 0.0    },
  { 29, 76, 75.945268974, 0.0    },
  { 29, 77, 76.947543599, 0.0    },
  { 29, 78, 77.951916524, 0.0    },
  { 29, 79, 78.954473100, 0.0    },
  { 29, 80, 79.960623,    0.0    },
  { 29, 81, 80.965743,    0.0    },
  { 29, 82, 81.972378,    0.0    },
  { 29, 83, 82.978110,    0.0    },
  { 29, 84, 83.985271,    0.0    },
} },
{ 30, "Zn", "Zinc", 65.38, 1.22, 2.39, 1.65, {
  { 30, 54, 53.993879,    0.0    },
  { 30, 55, 54.984681,    0.0    },
  { 30, 56, 55.972743,    0.0    },
  { 30, 57, 56.965056,    0.0    },
  { 30, 58, 57.954590296, 0.0    },
  { 30, 59, 58.949311886, 0.0    },
  { 30, 60, 59.941841317, 0.0    },
  { 30, 61, 60.939506964, 0.0    },
  { 30, 62, 61.934333359, 0.0    },
  { 30, 63, 62.933211140, 0.0    },
  { 30, 64, 63.929141776, 0.4917 },
  { 30, 65, 64.929240534, 0.0    },
  { 30, 66, 65.926033639, 0.2773 },
  { 30, 67, 66.927127422, 0.0404 },
  { 30, 68, 67.924844232, 0.1845 },
  { 30, 69, 68.926550360, 0.0    },
  { 30, 70, 69.925319175, 0.0061 },
  { 30, 71, 70.927719578, 0.0    },
  { 30, 72, 71.926842806, 0.0    },
  { 30, 73, 72.929582580, 0.0    },
  { 30, 74, 73.929407260, 0.0    },
  { 30, 75, 74.932840244, 0.0    },
  { 30, 76, 75.933114956, 0.0    },
  { 30, 77, 76.936887197, 0.0    },
  { 30, 78, 77.938289204, 0.0    },
  { 30, 79, 78.942638067, 0.0    },
  { 30, 80, 79.944552929, 0.0    },
  { 30, 81, 80.950402617, 0.0    },
  { 30, 82, 81.954574097, 0.0    },
  { 30, 83, 82.961041,    0.0    },
  { 30, 84, 83.965829,    0.0    },
  { 30, 85, 84.973054,    0.0    },
  { 30, 86, 85.978463,    0.0    },
} },
{ 31, "Ga", "Gallium", 69.723, 1.22, 2.32, 1.81, {
  { 31, 56, 55.995878,    0.0     },
  { 31, 57, 56.983457,    0.0     },
  { 31, 58, 57.974729,    0.0     },
  { 31, 59, 58.963757,    0.0     },
  { 31, 60, 59.957498,    0.0     },
  { 31, 61, 60.949398861, 0.0     },
  { 31, 62, 61.944189639, 0.0     },
  { 31, 63, 62.939294194, 0.0     },
  { 31, 64, 63.936840366, 0.0     },
  { 31, 65, 64.932734424, 0.0     },
  { 31, 66, 65.931589766, 0.0     },
  { 31, 67, 66.928202276, 0.0     },
  { 31, 68, 67.927980161, 0.0     },
  { 31, 69, 68.925573528, 0.60108 },
  { 31, 70, 69.926021914, 0.0     },
  { 31, 71, 70.924702554, 0.39892 },
  { 31, 72, 71.926367452, 0.0     },
  { 31, 73, 72.925174680, 0.0     },
  { 31, 74, 73.926945725, 0.0     },
  { 31, 75, 74.926504484, 0.0     },
  { 31, 76, 75.928827624, 0.0     },
  { 31, 77, 76.929154299, 0.0     },
  { 31, 78, 77.931610854, 0.0     },
  { 31, 79, 78.932851582, 0.0     },
  { 31, 80, 79.936420773, 0.0     },
  { 31, 81, 80.938133841, 0.0     },
  { 31, 82, 81.943176531, 0.0     },
  { 31, 83, 82.947120300, 0.0     },
  { 31, 84, 83.952663000, 0.0     },
  { 31, 85, 84.957333000, 0.0     },
  { 31, 86, 85.963757,    0.0     },
  { 31, 87, 86.969007,    0.0     },
  { 31, 88, 87.975963,    0.0     },
} },
{ 32, "Ge", "Germanium", 72.630, 1.20, 2.29, 2.01, {
  { 32, 58, 57.991863,    0.0    },
  { 32, 59, 58.982426,    0.0    },
  { 32, 60, 59.970445,    0.0    },
  { 32, 61, 60.963725,    0.0    },
  { 32, 62, 61.954761,    0.0    },
  { 32, 63, 62.949628000, 0.0    },
  { 32, 64, 63.941689912, 0.0    },
  { 32, 65, 64.939368136, 0.0    },
  { 32, 66, 65.933862124, 0.0    },
  { 32, 67, 66.932716999, 0.0    },
  { 32, 68, 67.928095305, 0.0    },
  { 32, 69, 68.927964467, 0.0    },
  { 32, 70, 69.924248542, 0.2052 },
  { 32, 71, 70.924952120, 0.0    },
  { 32, 72, 71.922075824, 0.2745 },
  { 32, 73, 72.923458954, 0.0776 },
  { 32, 74, 73.921177760, 0.3652 },
  { 32, 75, 74.922858370, 0.0    },
  { 32, 76, 75.921402725, 0.0775 },
  { 32, 77, 76.923549843, 0.0    },
  { 32, 78, 77.922852911, 0.0    },
  { 32, 79, 78.925359506, 0.0    },
  { 32, 80, 79.925350773, 0.0    },
  { 32, 81, 80.928832941, 0.0    },
  { 32, 82, 81.929774031, 0.0    },
  { 32, 83, 82.934539100, 0.0    },
  { 32, 84, 83.937575090, 0.0    },
  { 32, 85, 84.942969658, 0.0    },
  { 32, 86, 85.946967000, 0.0    },
  { 32, 87, 86.953204,    0.0    },
  { 32, 88, 87.957574,    0.0    },
  { 32, 89, 88.964530,    0.0    },
  { 32, 90, 89.969436,    0.0    },
} },
{ 33, "As", "Arsenic", 74.921595, 1.19, 1.88, 2.18, {
  { 33, 60, 59.993945,    0.0 },
  { 33, 61, 60.981535,    0.0 },
  { 33, 62, 61.973784,    0.0 },
  { 33, 63, 62.964036,    0.0 },
  { 33, 64, 63.957560,    0.0 },
  { 33, 65, 64.949611000, 0.0 },
  { 33, 66, 65.944148778, 0.0 },
  { 33, 67, 66.939251110, 0.0 },
  { 33, 68, 67.936774127, 0.0 },
  { 33, 69, 68.932246289, 0.0 },
  { 33, 70, 69.930934642, 0.0 },
  { 33, 71, 70.927113594, 0.0 },
  { 33, 72, 71.926752291, 0.0 },
  { 33, 73, 72.923829086, 0.0 },
  { 33, 74, 73.923928596, 0.0 },
  { 33, 75, 74.921594562, 1.0 },
  { 33, 76, 75.922392011, 0.0 },
  { 33, 77, 76.920647555, 0.0 },
  { 33, 78, 77.921827771, 0.0 },
  { 33, 79, 78.920948419, 0.0 },
  { 33, 80, 79.922474440, 0.0 },
  { 33, 81, 80.922132288, 0.0 },
  { 33, 82, 81.924738731, 0.0 },
  { 33, 83, 82.925206900, 0.0 },
  { 33, 84, 83.929303290, 0.0 },
  { 33, 85, 84.932163658, 0.0 },
  { 33, 86, 85.936701532, 0.0 },
  { 33, 87, 86.940291716, 0.0 },
  { 33, 88, 87.945840,    0.0 },
  { 33, 89, 88.950048,    0.0 },
  { 33, 90, 89.955995,    0.0 },
  { 33, 91, 90.960816,    0.0 },
  { 33, 92, 91.967386,    0.0 },
} },
{ 34, "Se", "Selenium", 78.971, 1.20, 1.82, 2.55, {
  { 34, 63, 62.981911,    0.0    },
  { 34, 64, 63.971165,    0.0    },
  { 34, 65, 64.964552,    0.0    },
  { 34, 66, 65.955276,    0.0    },
  { 34, 67, 66.949994000, 0.0    },
  { 34, 68, 67.941825236, 0.0    },
  { 34, 69, 68.939414845, 0.0    },
  { 34, 70, 69.933515521, 0.0    },
  { 34, 71, 70.932209431, 0.0    },
  { 34, 72, 71.927140506, 0.0    },
  { 34, 73, 72.926754881, 0.0    },
  { 34, 74, 73.922475933, 0.0086 },
  { 34, 75, 74.922522870, 0.0    },
  { 34, 76, 75.919213702, 0.0923 },
  { 34, 77, 76.919914150, 0.0760 },
  { 34, 78, 77.917309244, 0.2369 },
  { 34, 79, 78.918499252, 0.0    },
  { 34, 80, 79.916521761, 0.4980 },
  { 34, 81, 80.917993019, 0.0    },
  { 34, 82, 81.916699531, 0.0882 },
  { 34, 83, 82.919118604, 0.0    },
  { 34, 84, 83.918466761, 0.0    },
  { 34, 85, 84.922260758, 0.0    },
  { 34, 86, 85.924311732, 0.0    },
  { 34, 87, 86.928688616, 0.0    },
  { 34, 88, 87.931417490, 0.0    },
  { 34, 89, 88.936669058, 0.0    },
  { 34, 90, 89.940096000, 0.0    },
  { 34, 91, 90.945700000, 0.0    },
  { 34, 92, 91.949840,    0.0    },
  { 34, 93, 92.956135,    0.0    },
  { 34, 94, 93.960490,    0.0    },
  { 34, 95, 94.967300,    0.0    },
} },
{ 35, "Br", "Bromine", 79.904, 1.20, 1.86, 2.96, {
  { 35, 65, 64.982297,    0.0     },
  { 35, 66, 65.974697,    0.0     },
  { 35, 67, 66.965078,    0.0     },
  { 35, 68, 67.958356,    0.0     },
  { 35, 69, 68.950338410, 0.0     },
  { 35, 70, 69.944792321, 0.0     },
  { 35, 71, 70.939342153, 0.0     },
  { 35, 72, 71.936594606, 0.0     },
  { 35, 73, 72.931673441, 0.0     },
  { 35, 74, 73.929910279, 0.0     },
  { 35, 75, 74.925810566, 0.0     },
  { 35, 76, 75.924541574, 0.0     },
  { 35, 77, 76.921379193, 0.0     },
  { 35, 78, 77.921145858, 0.0     },
  { 35, 79, 78.918337574, 0.50659 },
  { 35, 80, 79.918529784, 0.0     },
  { 35, 81, 80.916288197, 0.49359 },
  { 35, 82, 81.916801752, 0.0     },
  { 35, 83, 82.915175285, 0.0     },
  { 35, 84, 83.916496417, 0.0     },
  { 35, 85, 84.915645758, 0.0     },
  { 35, 86, 85.918805432, 0.0     },
  { 35, 87, 86.920674016, 0.0     },
  { 35, 88, 87.924083290, 0.0     },
  { 35, 89, 88.926704558, 0.0     },
  { 35, 90, 89.931292848, 0.0     },
  { 35, 91, 90.934398617, 0.0     },
  { 35, 92, 91.939631595, 0.0     },
  { 35, 93, 92.943220000, 0.0     },
  { 35, 94, 93.948846,    0.0     },
  { 35, 95, 94.952925,    0.0     },
  { 35, 96, 95.958980,    0.0     },
  { 35, 97, 96.963499,    0.0     },
  { 35, 98, 97.969887,    0.0     },
} },
{ 36, "Kr", "Krypton", 83.798, 1.16, 2.25, 3.00, {
  { 36,  67,  66.983305,      0.0     },
  { 36,  68,  67.972489,      0.0     },
  { 36,  69,  68.965496,      0.0     },
  { 36,  70,  69.955877,      0.0     },
  { 36,  71,  70.950265695,   0.0     },
  { 36,  72,  71.942092406,   0.0     },
  { 36,  73,  72.939289193,   0.0     },
  { 36,  74,  73.933084016,   0.0     },
  { 36,  75,  74.930945744,   0.0     },
  { 36,  76,  75.925910743,   0.0     },
  { 36,  77,  76.924669999,   0.0     },
  { 36,  78,  77.920366341,   0.00355 },
  { 36,  79,  78.920082919,   0.0     },
  { 36,  80,  79.916377940,   0.02286 },
  { 36,  81,  80.916589703,   0.0     },
  { 36,  82,  81.91348115368, 0.11593 },
  { 36,  83,  82.914126516,   0.11500 },
  { 36,  84,  83.91149772708, 0.56987 },
  { 36,  85,  84.912527260,   0.0     },
  { 36,  86,  85.91061062468, 0.17279 },
  { 36,  87,  86.913354759,   0.0     },
  { 36,  88,  87.914447879,   0.0     },
  { 36,  89,  88.917835449,   0.0     },
  { 36,  90,  89.919527929,   0.0     },
  { 36,  91,  90.923806309,   0.0     },
  { 36,  92,  91.926173092,   0.0     },
  { 36,  93,  92.931147172,   0.0     },
  { 36,  94,  93.934140452,   0.0     },
  { 36,  95,  94.939710922,   0.0     },
  { 36,  96,  95.943014473,   0.0     },
  { 36,  97,  96.949088782,   0.0     },
  { 36,  98,  97.952635,      0.0     },
  { 36,  99,  98.958776,      0.0     },
  { 36, 100,  99.962995,      0.0     },
  { 36, 101, 100.969318,      0.0     },
} },
// 5th row
{ 37, "Rb", "Rubidium", 85.4678, 2.20, 3.21, 0.82, {
  { 37,  71,  70.965335,      0.0    },
  { 37,  72,  71.958851,      0.0    },
  { 37,  73,  72.950604506,   0.0    },
  { 37,  74,  73.944265867,   0.0    },
  { 37,  75,  74.938573200,   0.0    },
  { 37,  76,  75.935073031,   0.0    },
  { 37,  77,  76.930401599,   0.0    },
  { 37,  78,  77.928141866,   0.0    },
  { 37,  79,  78.923990095,   0.0    },
  { 37,  80,  79.922516442,   0.0    },
  { 37,  81,  80.918993900,   0.0    },
  { 37,  82,  81.918209023,   0.0    },
  { 37,  83,  82.915114181,   0.0    },
  { 37,  84,  83.914375223,   0.0    },
  { 37,  85,  84.91178973604, 0.7217 },
  { 37,  86,  85.911167443,   0.0    },
  { 37,  87,  86.909180529,   0.2783 },
  { 37,  88,  87.911315590,   0.0    },
  { 37,  89,  88.912278136,   0.0    },
  { 37,  90,  89.914797557,   0.0    },
  { 37,  91,  90.916537261,   0.0    },
  { 37,  92,  91.919728477,   0.0    },
  { 37,  93,  92.922039334,   0.0    },
  { 37,  94,  93.926394819,   0.0    },
  { 37,  95,  94.929263849,   0.0    },
  { 37,  96,  95.934133398,   0.0    },
  { 37,  97,  96.937177117,   0.0    },
  { 37,  98,  97.941632317,   0.0    },
  { 37,  99,  98.945119190,   0.0    },
  { 37, 100,  99.950331532,   0.0    },
  { 37, 101, 100.954302000,   0.0    },
  { 37, 102, 101.960008000,   0.0    },
  { 37, 103, 102.964401,      0.0    },
  { 37, 104, 103.970531,      0.0    },
} },
{ 38, "Sr", "Strontium", 87.62, 1.95, 2.84, 0.95, {
  { 38,  73,  72.965700,      0.0    },
  { 38,  74,  73.956170,      0.0    },
  { 38,  75,  74.949952767,   0.0    },
  { 38,  76,  75.941762760,   0.0    },
  { 38,  77,  76.937945454,   0.0    },
  { 38,  78,  77.932179979,   0.0    },
  { 38,  79,  78.929704692,   0.0    },
  { 38,  80,  79.924517538,   0.0    },
  { 38,  81,  80.923211393,   0.0    },
  { 38,  82,  81.918399845,   0.0    },
  { 38,  83,  82.917554372,   0.0    },
  { 38,  84,  83.913419118,   0.0056 },
  { 38,  85,  84.912932041,   0.0    },
  { 38,  86,  85.90926072473, 0.0986 },
  { 38,  87,  86.90887749454, 0.0700 },
  { 38,  88,  87.905612253,   0.8258 },
  { 38,  89,  88.907450808,   0.0    },
  { 38,  90,  89.907727870,   0.0    },
  { 38,  91,  90.910195942,   0.0    },
  { 38,  92,  91.911038222,   0.0    },
  { 38,  93,  92.914024314,   0.0    },
  { 38,  94,  93.915355641,   0.0    },
  { 38,  95,  94.919358282,   0.0    },
  { 38,  96,  95.921719045,   0.0    },
  { 38,  97,  96.926375621,   0.0    },
  { 38,  98,  97.928692636,   0.0    },
  { 38,  99,  98.932883604,   0.0    },
  { 38, 100,  99.935783270,   0.0    },
  { 38, 101, 100.940606264,   0.0    },
  { 38, 102, 101.944004679,   0.0    },
  { 38, 103, 102.949243,      0.0    },
  { 38, 104, 103.953022,      0.0    },
  { 38, 105, 104.959001,      0.0    },
  { 38, 106, 105.963177,      0.0    },
  { 38, 107, 106.969672,      0.0    },
} },
{ 39, "Y", "Yttrium", 88.905838, 1.90, 2.75, 1.22, {
  { 39,  75,  74.965840,    0.0 },
  { 39,  76,  75.958937,    0.0 },
  { 39,  77,  76.950146,    0.0 },
  { 39,  78,  77.943990,    0.0 },
  { 39,  79,  78.937946000, 0.0 },
  { 39,  80,  79.934354750, 0.0 },
  { 39,  81,  80.929454283, 0.0 },
  { 39,  82,  81.926930189, 0.0 },
  { 39,  83,  82.922484026, 0.0 },
  { 39,  84,  83.920671060, 0.0 },
  { 39,  85,  84.916433039, 0.0 },
  { 39,  86,  85.914886095, 0.0 },
  { 39,  87,  86.910876100, 0.0 },
  { 39,  88,  87.909501274, 0.0 },
  { 39,  89,  88.905838156, 1.0 },
  { 39,  90,  89.907141749, 0.0 },
  { 39,  91,  90.907298048, 0.0 },
  { 39,  92,  91.908945752, 0.0 },
  { 39,  93,  92.909578434, 0.0 },
  { 39,  94,  93.911592062, 0.0 },
  { 39,  95,  94.912819697, 0.0 },
  { 39,  96,  95.915909305, 0.0 },
  { 39,  97,  96.918286702, 0.0 },
  { 39,  98,  97.922394841, 0.0 },
  { 39,  99,  98.924160839, 0.0 },
  { 39, 100,  99.927727678, 0.0 },
  { 39, 101, 100.930160817, 0.0 },
  { 39, 102, 101.934328471, 0.0 },
  { 39, 103, 102.937243796, 0.0 },
  { 39, 104, 103.941943,    0.0 },
  { 39, 105, 104.945711,    0.0 },
  { 39, 106, 105.950842,    0.0 },
  { 39, 107, 106.954943,    0.0 },
  { 39, 108, 107.960515,    0.0 },
  { 39, 109, 108.965131,    0.0 },
} },
{ 40, "Zr", "Zirconium", 91.224, 1.75, 2.52, 1.33, {
  { 40,  77,  76.966076,    0.0    },
  { 40,  78,  77.956146,    0.0    },
  { 40,  79,  78.949790,    0.0    },
  { 40,  80,  79.941213,    0.0    },
  { 40,  81,  80.938245000, 0.0    },
  { 40,  82,  81.931707497, 0.0    },
  { 40,  83,  82.929240926, 0.0    },
  { 40,  84,  83.923325663, 0.0    },
  { 40,  85,  84.921443199, 0.0    },
  { 40,  86,  85.916296814, 0.0    },
  { 40,  87,  86.914817338, 0.0    },
  { 40,  88,  87.910220715, 0.0    },
  { 40,  89,  88.908879751, 0.0    },
  { 40,  90,  89.904698755, 0.5145 },
  { 40,  91,  90.905640205, 0.1122 },
  { 40,  92,  91.905035336, 0.1715 },
  { 40,  93,  92.906470661, 0.0    },
  { 40,  94,  93.906312523, 0.1738 },
  { 40,  95,  94.908040276, 0.0    },
  { 40,  96,  95.908277615, 0.0280 },
  { 40,  97,  96.910963802, 0.0    },
  { 40,  98,  97.912740448, 0.0    },
  { 40,  99,  98.916675081, 0.0    },
  { 40, 100,  99.918010499, 0.0    },
  { 40, 101, 100.921458454, 0.0    },
  { 40, 102, 101.923154181, 0.0    },
  { 40, 103, 102.927204054, 0.0    },
  { 40, 104, 103.929449193, 0.0    },
  { 40, 105, 104.934021832, 0.0    },
  { 40, 106, 105.936930,    0.0    },
  { 40, 107, 106.942007,    0.0    },
  { 40, 108, 107.945303,    0.0    },
  { 40, 109, 108.950907,    0.0    },
  { 40, 110, 109.954675,    0.0    },
  { 40, 111, 110.960837,    0.0    },
  { 40, 112, 111.965196,    0.0    },
  { 40, 113, 112.971723,    0.0    },
} },
{ 41, "Nb", "Niobium", 92.90637, 1.64, 2.56, 1.6, {
  { 41,  79,  78.966022,    0.0 },
  { 41,  80,  79.958754,    0.0 },
  { 41,  81,  80.950230,    0.0 },
  { 41,  82,  81.944380,    0.0 },
  { 41,  83,  82.938150000, 0.0 },
  { 41,  84,  83.934305711, 0.0 },
  { 41,  85,  84.928845836, 0.0 },
  { 41,  86,  85.925781536, 0.0 },
  { 41,  87,  86.920692473, 0.0 },
  { 41,  88,  87.918226476, 0.0 },
  { 41,  89,  88.913444696, 0.0 },
  { 41,  90,  89.911259201, 0.0 },
  { 41,  91,  90.906990256, 0.0 },
  { 41,  92,  91.907188580, 0.0 },
  { 41,  93,  92.906373170, 1.0 },
  { 41,  94,  93.907279001, 0.0 },
  { 41,  95,  94.906831110, 0.0 },
  { 41,  96,  95.908101586, 0.0 },
  { 41,  97,  96.908101622, 0.0 },
  { 41,  98,  97.910332645, 0.0 },
  { 41,  99,  98.911609377, 0.0 },
  { 41, 100,  99.914340578, 0.0 },
  { 41, 101, 100.915306508, 0.0 },
  { 41, 102, 101.918090447, 0.0 },
  { 41, 103, 102.919453416, 0.0 },
  { 41, 104, 103.922907728, 0.0 },
  { 41, 105, 104.924942577, 0.0 },
  { 41, 106, 105.928928505, 0.0 },
  { 41, 107, 106.931589685, 0.0 },
  { 41, 108, 107.936075604, 0.0 },
  { 41, 109, 108.939141000, 0.0 },
  { 41, 110, 109.943843000, 0.0 },
  { 41, 111, 110.947439,    0.0 },
  { 41, 112, 111.952689,    0.0 },
  { 41, 113, 112.956833,    0.0 },
  { 41, 114, 113.962469,    0.0 },
  { 41, 115, 114.966849,    0.0 },
  { 41, 116, 115.972914,    0.0 },
} },
{ 42, "Mo", "Molybdenum", 95.95, 1.54, 2.45, 2.16, {
  { 42,  81,  80.966226,    0.0     },
  { 42,  82,  81.956661,    0.0     },
  { 42,  83,  82.950252,    0.0     },
  { 42,  84,  83.941846,    0.0     },
  { 42,  85,  84.938260736, 0.0     },
  { 42,  86,  85.931174092, 0.0     },
  { 42,  87,  86.928196198, 0.0     },
  { 42,  88,  87.921967779, 0.0     },
  { 42,  89,  88.919468149, 0.0     },
  { 42,  90,  89.913931270, 0.0     },
  { 42,  91,  90.911745190, 0.0     },
  { 42,  92,  91.906807153, 0.14649 },
  { 42,  93,  92.906808772, 0.0     },
  { 42,  94,  93.905083586, 0.09187 },
  { 42,  95,  94.905837436, 0.15873 },
  { 42,  96,  95.904674770, 0.16673 },
  { 42,  97,  96.906016903, 0.09582 },
  { 42,  98,  97.905403609, 0.24292 },
  { 42,  99,  98.907707299, 0.0     },
  { 42, 100,  99.907467982, 0.09744 },
  { 42, 101, 100.910337648, 0.0     },
  { 42, 102, 101.910293725, 0.0     },
  { 42, 103, 102.913091954, 0.0     },
  { 42, 104, 103.913747443, 0.0     },
  { 42, 105, 104.916981989, 0.0     },
  { 42, 106, 105.918273231, 0.0     },
  { 42, 107, 106.922119770, 0.0     },
  { 42, 108, 107.924047508, 0.0     },
  { 42, 109, 108.928438318, 0.0     },
  { 42, 110, 109.930717956, 0.0     },
  { 42, 111, 110.935651966, 0.0     },
  { 42, 112, 111.938293,    0.0     },
  { 42, 113, 112.943478,    0.0     },
  { 42, 114, 113.946666,    0.0     },
  { 42, 115, 114.952174,    0.0     },
  { 42, 116, 115.955759,    0.0     },
  { 42, 117, 116.961686,    0.0     },
  { 42, 118, 117.965249,    0.0     },
  { 42, 119, 118.971465,    0.0     },
} },
{ 43, "Tc", "Technetium", 97, 1.47, 2.44, 1.9, {
  { 43,  83,  82.966377,    0.0 },
  { 43,  84,  83.959527,    0.0 },
  { 43,  85,  84.950778,    0.0 },
  { 43,  86,  85.944637,    0.0 },
  { 43,  87,  86.938067185, 0.0 },
  { 43,  88,  87.933794211, 0.0 },
  { 43,  89,  88.927648649, 0.0 },
  { 43,  90,  89.924073919, 0.0 },
  { 43,  91,  90.918424972, 0.0 },
  { 43,  92,  91.915269777, 0.0 },
  { 43,  93,  92.910245147, 0.0 },
  { 43,  94,  93.909652319, 0.0 },
  { 43,  95,  94.907652281, 0.0 },
  { 43,  96,  95.907866675, 0.0 },
  { 43,  97,  96.906360720, 0.0 },
  { 43,  98,  97.907211206, 0.0 },
  { 43,  99,  98.906249681, 0.0 },
  { 43, 100,  99.907652715, 0.0 },
  { 43, 101, 100.907305271, 0.0 },
  { 43, 102, 101.909207239, 0.0 },
  { 43, 103, 102.909173960, 0.0 },
  { 43, 104, 103.911433718, 0.0 },
  { 43, 105, 104.911662024, 0.0 },
  { 43, 106, 105.914356674, 0.0 },
  { 43, 107, 106.915458437, 0.0 },
  { 43, 108, 107.918493493, 0.0 },
  { 43, 109, 108.920254107, 0.0 },
  { 43, 110, 109.923741263, 0.0 },
  { 43, 111, 110.925898966, 0.0 },
  { 43, 112, 111.929941658, 0.0 },
  { 43, 113, 112.932569032, 0.0 },
  { 43, 114, 113.937090000, 0.0 },
  { 43, 115, 114.940100,    0.0 },
  { 43, 116, 115.945020,    0.0 },
  { 43, 117, 116.948320,    0.0 },
  { 43, 118, 117.953526,    0.0 },
  { 43, 119, 118.956876,    0.0 },
  { 43, 120, 119.962426,    0.0 },
  { 43, 121, 120.966140,    0.0 },
  { 43, 122, 121.971760,    0.0 },
} },
{ 44, "Ru", "Ruthenium", 101.07, 1.46, 2.46, 2.2, {
  { 44,  85,  84.967117,    0.0    },
  { 44,  86,  85.957305,    0.0    },
  { 44,  87,  86.950907,    0.0    },
  { 44,  88,  87.941664,    0.0    },
  { 44,  89,  88.937337849, 0.0    },
  { 44,  90,  89.930344378, 0.0    },
  { 44,  91,  90.926741530, 0.0    },
  { 44,  92,  91.920234373, 0.0    },
  { 44,  93,  92.917104442, 0.0    },
  { 44,  94,  93.911342860, 0.0    },
  { 44,  95,  94.910404415, 0.0    },
  { 44,  96,  95.907588910, 0.0554 },
  { 44,  97,  96.907545776, 0.0    },
  { 44,  98,  97.905286709, 0.0187 },
  { 44,  99,  98.905930284, 0.1276 },
  { 44, 100,  99.904210460, 0.1260 },
  { 44, 101, 100.905573086, 0.1706 },
  { 44, 102, 101.904340312, 0.3155 },
  { 44, 103, 102.906314846, 0.0    },
  { 44, 104, 103.905425312, 0.1862 },
  { 44, 105, 104.907745478, 0.0    },
  { 44, 106, 105.907328181, 0.0    },
  { 44, 107, 106.909969837, 0.0    },
  { 44, 108, 107.910185793, 0.0    },
  { 44, 109, 108.913323707, 0.0    },
  { 44, 110, 109.914038501, 0.0    },
  { 44, 111, 110.917567566, 0.0    },
  { 44, 112, 111.918806922, 0.0    },
  { 44, 113, 112.922846729, 0.0    },
  { 44, 114, 113.924614430, 0.0    },
  { 44, 115, 114.929033049, 0.0    },
  { 44, 116, 115.931219191, 0.0    },
  { 44, 117, 116.936135000, 0.0    },
  { 44, 118, 117.938808,    0.0    },
  { 44, 119, 118.944090,    0.0    },
  { 44, 120, 119.946623,    0.0    },
  { 44, 121, 120.952098,    0.0    },
  { 44, 122, 121.955147,    0.0    },
  { 44, 123, 122.960762,    0.0    },
  { 44, 124, 123.963940,    0.0    },
  { 44, 125, 124.969544,    0.0    },
} },
{ 45, "Rh", "Rhodium", 102.90549, 1.42, 2.44, 2.28, {
  { 45,  88,  87.960429,    0.0 },
  { 45,  89,  88.950992,    0.0 },
  { 45,  90,  89.944569,    0.0 },
  { 45,  91,  90.937123,    0.0 },
  { 45,  92,  91.932367692, 0.0 },
  { 45,  93,  92.925912778, 0.0 },
  { 45,  94,  93.921730450, 0.0 },
  { 45,  95,  94.915897893, 0.0 },
  { 45,  96,  95.914451705, 0.0 },
  { 45,  97,  96.911327872, 0.0 },
  { 45,  98,  97.910707734, 0.0 },
  { 45,  99,  98.908121241, 0.0 },
  { 45, 100,  99.908114147, 0.0 },
  { 45, 101, 100.906158903, 0.0 },
  { 45, 102, 101.906834282, 0.0 },
  { 45, 103, 102.905494081, 1.0 },
  { 45, 104, 103.906645309, 0.0 },
  { 45, 105, 104.905687787, 0.0 },
  { 45, 106, 105.907285879, 0.0 },
  { 45, 107, 106.906747975, 0.0 },
  { 45, 108, 107.908715304, 0.0 },
  { 45, 109, 108.908749555, 0.0 },
  { 45, 110, 109.911079745, 0.0 },
  { 45, 111, 110.911643164, 0.0 },
  { 45, 112, 111.914405199, 0.0 },
  { 45, 113, 112.915440212, 0.0 },
  { 45, 114, 113.918721680, 0.0 },
  { 45, 115, 114.920311649, 0.0 },
  { 45, 116, 115.924062060, 0.0 },
  { 45, 117, 116.926036291, 0.0 },
  { 45, 118, 117.930341116, 0.0 },
  { 45, 119, 118.932556951, 0.0 },
  { 45, 120, 119.937069,    0.0 },
  { 45, 121, 120.939613000, 0.0 },
  { 45, 122, 121.944305,    0.0 },
  { 45, 123, 122.947192,    0.0 },
  { 45, 124, 123.952002,    0.0 },
  { 45, 125, 124.955094,    0.0 },
  { 45, 126, 125.960064,    0.0 },
  { 45, 127, 126.963789,    0.0 },
  { 45, 128, 127.970649,    0.0 },
} },
{ 46, "Pd", "Palladium", 106.42, 1.39, 2.15, 2.20, {
  { 46,  90,  89.957370,    0.0    },
  { 46,  91,  90.950435,    0.0    },
  { 46,  92,  91.941192225, 0.0    },
  { 46,  93,  92.936680426, 0.0    },
  { 46,  94,  93.929036286, 0.0    },
  { 46,  95,  94.924888506, 0.0    },
  { 46,  96,  95.918213739, 0.0    },
  { 46,  97,  96.916471985, 0.0    },
  { 46,  98,  97.912698335, 0.0    },
  { 46,  99,  98.911773073, 0.0    },
  { 46, 100,  99.908520438, 0.0    },
  { 46, 101, 100.908284824, 0.0    },
  { 46, 102, 101.905632292, 0.0102 },
  { 46, 103, 102.906111074, 0.0    },
  { 46, 104, 103.904030393, 0.1114 },
  { 46, 105, 104.905079479, 0.2233 },
  { 46, 106, 105.903480287, 0.2733 },
  { 46, 107, 106.905128058, 0.0    },
  { 46, 108, 107.903891806, 0.2646 },
  { 46, 109, 108.905950576, 0.0    },
  { 46, 110, 109.905172878, 0.1172 },
  { 46, 111, 110.907690358, 0.0    },
  { 46, 112, 111.907330557, 0.0    },
  { 46, 113, 112.910261912, 0.0    },
  { 46, 114, 113.910369430, 0.0    },
  { 46, 115, 114.913659333, 0.0    },
  { 46, 116, 115.914297872, 0.0    },
  { 46, 117, 116.917955584, 0.0    },
  { 46, 118, 117.919067273, 0.0    },
  { 46, 119, 118.923341138, 0.0    },
  { 46, 120, 119.924551745, 0.0    },
  { 46, 121, 120.928950342, 0.0    },
  { 46, 122, 121.930631693, 0.0    },
  { 46, 123, 122.935126000, 0.0    },
  { 46, 124, 123.937305,    0.0    },
  { 46, 125, 124.942072,    0.0    },
  { 46, 126, 125.944401,    0.0    },
  { 46, 127, 126.949307,    0.0    },
  { 46, 128, 127.952345,    0.0    },
  { 46, 129, 128.959334,    0.0    },
  { 46, 130, 129.964863,    0.0    },
  { 46, 131, 130.972367,    0.0    },
} },
{ 47, "Ag", "Silver", 107.8682, 1.45, 2.53, 1.93, {
  { 47,  92,  91.959710,    0.0     },
  { 47,  93,  92.950188,    0.0     },
  { 47,  94,  93.943744,    0.0     },
  { 47,  95,  94.935688,    0.0     },
  { 47,  96,  95.930743903, 0.0     },
  { 47,  97,  96.923881400, 0.0     },
  { 47,  98,  97.921559970, 0.0     },
  { 47,  99,  98.917645766, 0.0     },
  { 47, 100,  99.916115443, 0.0     },
  { 47, 101, 100.912683951, 0.0     },
  { 47, 102, 101.911704538, 0.0     },
  { 47, 103, 102.908960558, 0.0     },
  { 47, 104, 103.908623715, 0.0     },
  { 47, 105, 104.906525604, 0.0     },
  { 47, 106, 105.906663499, 0.0     },
  { 47, 107, 106.905091509, 0.51839 },
  { 47, 108, 107.905950245, 0.0     },
  { 47, 109, 108.904755778, 0.48161 },
  { 47, 110, 109.906110724, 0.0     },
  { 47, 111, 110.905296827, 0.0     },
  { 47, 112, 111.907048548, 0.0     },
  { 47, 113, 112.906572865, 0.0     },
  { 47, 114, 113.908823029, 0.0     },
  { 47, 115, 114.908767445, 0.0     },
  { 47, 116, 115.911386809, 0.0     },
  { 47, 117, 116.911774086, 0.0     },
  { 47, 118, 117.914595484, 0.0     },
  { 47, 119, 118.915570309, 0.0     },
  { 47, 120, 119.918784765, 0.0     },
  { 47, 121, 120.920125279, 0.0     },
  { 47, 122, 121.923664446, 0.0     },
  { 47, 123, 122.925315060, 0.0     },
  { 47, 124, 123.928899227, 0.0     },
  { 47, 125, 124.930735000, 0.0     },
  { 47, 126, 125.934814,    0.0     },
  { 47, 127, 126.937037,    0.0     },
  { 47, 128, 127.941266,    0.0     },
  { 47, 129, 128.944315,    0.0     },
  { 47, 130, 129.950727,    0.0     },
  { 47, 131, 130.956253,    0.0     },
  { 47, 132, 131.963070,    0.0     },
  { 47, 133, 132.968781,    0.0     },
} },
{ 48, "Cd", "Cadmium", 112.414, 1.44, 2.49, 1.69, {
  { 48,  94,  93.956586,    0.0     },
  { 48,  95,  94.949483,    0.0     },
  { 48,  96,  95.940341,    0.0     },
  { 48,  97,  96.934799343, 0.0     },
  { 48,  98,  97.927389315, 0.0     },
  { 48,  99,  98.924925845, 0.0     },
  { 48, 100,  99.920348829, 0.0     },
  { 48, 101, 100.918586209, 0.0     },
  { 48, 102, 101.914481797, 0.0     },
  { 48, 103, 102.913416922, 0.0     },
  { 48, 104, 103.909856228, 0.0     },
  { 48, 105, 104.909463893, 0.0     },
  { 48, 106, 105.906459791, 0.01245 },
  { 48, 107, 106.906612049, 0.0     },
  { 48, 108, 107.904183588, 0.00888 },
  { 48, 109, 108.904986697, 0.0     },
  { 48, 110, 109.903007470, 0.12470 },
  { 48, 111, 110.904183776, 0.12795 },
  { 48, 112, 111.902763896, 0.24109 },
  { 48, 113, 112.904408105, 0.12227 },
  { 48, 114, 113.903364998, 0.28754 },
  { 48, 115, 114.905437426, 0.0     },
  { 48, 116, 115.904763230, 0.07512 },
  { 48, 117, 116.907226039, 0.0     },
  { 48, 118, 117.906921956, 0.0     },
  { 48, 119, 118.909847052, 0.0     },
  { 48, 120, 119.909868065, 0.0     },
  { 48, 121, 120.912963660, 0.0     },
  { 48, 122, 121.913459050, 0.0     },
  { 48, 123, 122.916892460, 0.0     },
  { 48, 124, 123.917659772, 0.0     },
  { 48, 125, 124.921257590, 0.0     },
  { 48, 126, 125.922430290, 0.0     },
  { 48, 127, 126.926203291, 0.0     },
  { 48, 128, 127.927816778, 0.0     },
  { 48, 129, 128.932235597, 0.0     },
  { 48, 130, 129.934387563, 0.0     },
  { 48, 131, 130.940727740, 0.0     },
  { 48, 132, 131.945823136, 0.0     },
  { 48, 133, 132.952614,    0.0     },
  { 48, 134, 133.957638,    0.0     },
  { 48, 135, 134.964766,    0.0     },
} },
{ 49, "In", "Indium", 114.818, 1.42, 2.43, 1.78, {
  { 49,  96,  95.959109,    0.0     },
  { 49,  97,  96.949125,    0.0     },
  { 49,  98,  97.942129,    0.0     },
  { 49,  99,  98.934110,    0.0     },
  { 49, 100,  99.931101929, 0.0     },
  { 49, 101, 100.926414025, 0.0     },
  { 49, 102, 101.924105911, 0.0     },
  { 49, 103, 102.919878830, 0.0     },
  { 49, 104, 103.918214538, 0.0     },
  { 49, 105, 104.914502322, 0.0     },
  { 49, 106, 105.913463596, 0.0     },
  { 49, 107, 106.910287497, 0.0     },
  { 49, 108, 107.909693654, 0.0     },
  { 49, 109, 108.907149679, 0.0     },
  { 49, 110, 109.907170674, 0.0     },
  { 49, 111, 110.905107236, 0.0     },
  { 49, 112, 111.905538718, 0.0     },
  { 49, 113, 112.904060451, 0.04281 },
  { 49, 114, 113.904916405, 0.0     },
  { 49, 115, 114.903878772, 0.95719 },
  { 49, 116, 115.905259992, 0.0     },
  { 49, 117, 116.904515729, 0.0     },
  { 49, 118, 117.906356705, 0.0     },
  { 49, 119, 118.905851622, 0.0     },
  { 49, 120, 119.907967489, 0.0     },
  { 49, 121, 120.907852778, 0.0     },
  { 49, 122, 121.910282458, 0.0     },
  { 49, 123, 122.910435252, 0.0     },
  { 49, 124, 123.913184873, 0.0     },
  { 49, 125, 124.913673841, 0.0     },
  { 49, 126, 125.916468202, 0.0     },
  { 49, 127, 126.917466040, 0.0     },
  { 49, 128, 127.920353637, 0.0     },
  { 49, 129, 128.921808534, 0.0     },
  { 49, 130, 129.924952257, 0.0     },
  { 49, 131, 130.926972839, 0.0     },
  { 49, 132, 131.932998444, 0.0     },
  { 49, 133, 132.938067,    0.0     },
  { 49, 134, 133.944208,    0.0     },
  { 49, 135, 134.949425,    0.0     },
  { 49, 136, 135.956017,    0.0     },
  { 49, 137, 136.961535,    0.0     },
} },
{ 50, "Sn", "Tin", 118.710, 1.39, 2.42, 1.96, {
  { 50,  99,  98.948495,    0.0    },
  { 50, 100,  99.938648944, 0.0    },
  { 50, 101, 100.935259252, 0.0    },
  { 50, 102, 101.930289525, 0.0    },
  { 50, 103, 102.927973,    0.0    },
  { 50, 104, 103.923105195, 0.0    },
  { 50, 105, 104.921268421, 0.0    },
  { 50, 106, 105.916957394, 0.0    },
  { 50, 107, 106.915713649, 0.0    },
  { 50, 108, 107.911894290, 0.0    },
  { 50, 109, 108.911292857, 0.0    },
  { 50, 110, 109.907844835, 0.0    },
  { 50, 111, 110.907741143, 0.0    },
  { 50, 112, 111.904824894, 0.0097 },
  { 50, 113, 112.905175857, 0.0    },
  { 50, 114, 113.902780130, 0.0066 },
  { 50, 115, 114.903344695, 0.0034 },
  { 50, 116, 115.901742825, 0.1454 },
  { 50, 117, 116.902954036, 0.0768 },
  { 50, 118, 117.901606630, 0.2422 },
  { 50, 119, 118.903311266, 0.0859 },
  { 50, 120, 119.902202557, 0.3258 },
  { 50, 121, 120.904243488, 0.0    },
  { 50, 122, 121.903445494, 0.0463 },
  { 50, 123, 122.905727065, 0.0    },
  { 50, 124, 123.905279619, 0.0579 },
  { 50, 125, 124.907789370, 0.0    },
  { 50, 126, 125.907658958, 0.0    },
  { 50, 127, 126.910391726, 0.0    },
  { 50, 128, 127.910507828, 0.0    },
  { 50, 129, 128.913482440, 0.0    },
  { 50, 130, 129.913974531, 0.0    },
  { 50, 131, 130.917053067, 0.0    },
  { 50, 132, 131.917823898, 0.0    },
  { 50, 133, 132.923913753, 0.0    },
  { 50, 134, 133.928680430, 0.0    },
  { 50, 135, 134.934908603, 0.0    },
  { 50, 136, 135.939699,    0.0    },
  { 50, 137, 136.946162,    0.0    },
  { 50, 138, 137.951143,    0.0    },
  { 50, 139, 138.957799,    0.0    },
  { 50, 140, 139.962973,    0.0    },
} },
{ 51, "Sb", "Antimony", 121.760, 1.39, 2.47, 2.05, {
  { 51, 102, 101.945142,    0.0    },
  { 51, 103, 102.939162,    0.0    },
  { 51, 104, 103.936344,    0.0    },
  { 51, 105, 104.931276547, 0.0    },
  { 51, 106, 105.928637979, 0.0    },
  { 51, 107, 106.924150621, 0.0    },
  { 51, 108, 107.922226731, 0.0    },
  { 51, 109, 108.918141203, 0.0    },
  { 51, 110, 109.916854283, 0.0    },
  { 51, 111, 110.913218187, 0.0    },
  { 51, 112, 111.912399903, 0.0    },
  { 51, 113, 112.909374664, 0.0    },
  { 51, 114, 113.909289155, 0.0    },
  { 51, 115, 114.906598000, 0.0    },
  { 51, 116, 115.906792732, 0.0    },
  { 51, 117, 116.904841519, 0.0    },
  { 51, 118, 117.905532194, 0.0    },
  { 51, 119, 118.903944062, 0.0    },
  { 51, 120, 119.905080308, 0.0    },
  { 51, 121, 120.903811353, 0.5721 },
  { 51, 122, 121.905169335, 0.0    },
  { 51, 123, 122.904215292, 0.4279 },
  { 51, 124, 123.905937065, 0.0    },
  { 51, 125, 124.905254264, 0.0    },
  { 51, 126, 125.907253158, 0.0    },
  { 51, 127, 126.906925557, 0.0    },
  { 51, 128, 127.909146121, 0.0    },
  { 51, 129, 128.909146623, 0.0    },
  { 51, 130, 129.911662686, 0.0    },
  { 51, 131, 130.911989339, 0.0    },
  { 51, 132, 131.914508013, 0.0    },
  { 51, 133, 132.915272128, 0.0    },
  { 51, 134, 133.920537334, 0.0    },
  { 51, 135, 134.925184354, 0.0    },
  { 51, 136, 135.930749009, 0.0    },
  { 51, 137, 136.935522519, 0.0    },
  { 51, 138, 137.941331,    0.0    },
  { 51, 139, 138.946269,    0.0    },
  { 51, 140, 139.952345,    0.0    },
  { 51, 141, 140.957552,    0.0    },
  { 51, 142, 141.963918,    0.0    },
} },
{ 52, "Te", "Tellurium", 127.60, 1.38, 1.99, 2.1, {
  { 52, 104, 103.946723408, 0.0    },
  { 52, 105, 104.943304516, 0.0    },
  { 52, 106, 105.937498521, 0.0    },
  { 52, 107, 106.934882,    0.0    },
  { 52, 108, 107.929380469, 0.0    },
  { 52, 109, 108.927304532, 0.0    },
  { 52, 110, 109.922458102, 0.0    },
  { 52, 111, 110.921000587, 0.0    },
  { 52, 112, 111.916727848, 0.0    },
  { 52, 113, 112.915891000, 0.0    },
  { 52, 114, 113.912087820, 0.0    },
  { 52, 115, 114.911902000, 0.0    },
  { 52, 116, 115.908465558, 0.0    },
  { 52, 117, 116.908646227, 0.0    },
  { 52, 118, 117.905860104, 0.0    },
  { 52, 119, 118.906405699, 0.0    },
  { 52, 120, 119.904065779, 0.0009 },
  { 52, 121, 120.904945065, 0.0    },
  { 52, 122, 121.903044708, 0.0255 },
  { 52, 123, 122.904271022, 0.0089 },
  { 52, 124, 123.902818341, 0.0474 },
  { 52, 125, 124.904431178, 0.0707 },
  { 52, 126, 125.903312144, 0.1884 },
  { 52, 127, 126.905226993, 0.0    },
  { 52, 128, 127.904461237, 0.3174 },
  { 52, 129, 128.906596419, 0.0    },
  { 52, 130, 129.906222745, 0.3408 },
  { 52, 131, 130.908522210, 0.0    },
  { 52, 132, 131.908546713, 0.0    },
  { 52, 133, 132.910963330, 0.0    },
  { 52, 134, 133.911396376, 0.0    },
  { 52, 135, 134.916554715, 0.0    },
  { 52, 136, 135.920101180, 0.0    },
  { 52, 137, 136.925599354, 0.0    },
  { 52, 138, 137.929472452, 0.0    },
  { 52, 139, 138.935367191, 0.0    },
  { 52, 140, 139.939487057, 0.0    },
  { 52, 141, 140.945604,    0.0    },
  { 52, 142, 141.950027,    0.0    },
  { 52, 143, 142.956489,    0.0    },
  { 52, 144, 143.961116,    0.0    },
  { 52, 145, 144.967783,    0.0    },
} },
{ 53, "I", "Iodine", 126.90447, 1.39, 2.04, 2.66, {
  { 53, 106, 105.953516,    0.0 },
  { 53, 107, 106.946935,    0.0 },
  { 53, 108, 107.943348,    0.0 },
  { 53, 109, 108.938086022, 0.0 },
  { 53, 110, 109.935085102, 0.0 },
  { 53, 111, 110.930269236, 0.0 },
  { 53, 112, 111.928004548, 0.0 },
  { 53, 113, 112.923650062, 0.0 },
  { 53, 114, 113.922018900, 0.0 },
  { 53, 115, 114.918048000, 0.0 },
  { 53, 116, 115.916885513, 0.0 },
  { 53, 117, 116.913645649, 0.0 },
  { 53, 118, 117.913074000, 0.0 },
  { 53, 119, 118.910060910, 0.0 },
  { 53, 120, 119.910093729, 0.0 },
  { 53, 121, 120.907411492, 0.0 },
  { 53, 122, 121.907590094, 0.0 },
  { 53, 123, 122.905589753, 0.0 },
  { 53, 124, 123.906210297, 0.0 },
  { 53, 125, 124.904630610, 0.0 },
  { 53, 126, 125.905624205, 0.0 },
  { 53, 127, 126.904472592, 1.0 },
  { 53, 128, 127.905809355, 0.0 },
  { 53, 129, 128.904983643, 0.0 },
  { 53, 130, 129.906670168, 0.0 },
  { 53, 131, 130.906126375, 0.0 },
  { 53, 132, 131.907993511, 0.0 },
  { 53, 133, 132.907828400, 0.0 },
  { 53, 134, 133.909775660, 0.0 },
  { 53, 135, 134.910059355, 0.0 },
  { 53, 136, 135.914604693, 0.0 },
  { 53, 137, 136.918028178, 0.0 },
  { 53, 138, 137.922726392, 0.0 },
  { 53, 139, 138.926493400, 0.0 },
  { 53, 140, 139.931715914, 0.0 },
  { 53, 141, 140.935666081, 0.0 },
  { 53, 142, 141.941166595, 0.0 },
  { 53, 143, 142.945475,    0.0 },
  { 53, 144, 143.951336,    0.0 },
  { 53, 145, 144.955845,    0.0 },
  { 53, 146, 145.961846,    0.0 },
  { 53, 147, 146.966505,    0.0 },
} },
{ 54, "Xe", "Xenon", 131.293, 1.40, 2.06, 2.6, {
  { 54, 108, 107.954232285,   0.0     },
  { 54, 109, 108.950434955,   0.0     },
  { 54, 110, 109.944258759,   0.0     },
  { 54, 111, 110.941470,      0.0     },
  { 54, 112, 111.935559068,   0.0     },
  { 54, 113, 112.933221663,   0.0     },
  { 54, 114, 113.927980329,   0.0     },
  { 54, 115, 114.926293943,   0.0     },
  { 54, 116, 115.921580955,   0.0     },
  { 54, 117, 116.920358758,   0.0     },
  { 54, 118, 117.916178678,   0.0     },
  { 54, 119, 118.915410641,   0.0     },
  { 54, 120, 119.911784267,   0.0     },
  { 54, 121, 120.911453012,   0.0     },
  { 54, 122, 121.908367655,   0.0     },
  { 54, 123, 122.908482235,   0.0     },
  { 54, 124, 123.905885174,   0.00095 },
  { 54, 125, 124.906387640,   0.0     },
  { 54, 126, 125.904297422,   0.00089 },
  { 54, 127, 126.905183636,   0.0     },
  { 54, 128, 127.90353075341, 0.01910 },
  { 54, 129, 128.90478085742, 0.26401 },
  { 54, 130, 129.903509346,   0.04071 },
  { 54, 131, 130.90508412808, 0.21232 },
  { 54, 132, 131.90415508346, 0.26909 },
  { 54, 133, 132.905910748,   0.0     },
  { 54, 134, 133.905393030,   0.10436 },
  { 54, 135, 134.907231441,   0.0     },
  { 54, 136, 135.907214474,   0.08857 },
  { 54, 137, 136.911557771,   0.0     },
  { 54, 138, 137.914146268,   0.0     },
  { 54, 139, 138.918792200,   0.0     },
  { 54, 140, 139.921645814,   0.0     },
  { 54, 141, 140.926787181,   0.0     },
  { 54, 142, 141.929973095,   0.0     },
  { 54, 143, 142.935369550,   0.0     },
  { 54, 144, 143.938945076,   0.0     },
  { 54, 145, 144.944719631,   0.0     },
  { 54, 146, 145.948518245,   0.0     },
  { 54, 147, 146.954482,      0.0     },
  { 54, 148, 147.958508,      0.0     },
  { 54, 149, 148.964573,      0.0     },
  { 54, 150, 149.968878,      0.0     },
} },
// 6th row; default VdW radius = 2.5 (61, 84-86)
{ 55, "Cs", "Caesium", 132.90545196, 2.44, 3.48, 0.79, {
  { 55, 111, 110.953945,    0.0 },
  { 55, 112, 111.950172,    0.0 },
  { 55, 113, 112.944428484, 0.0 },
  { 55, 114, 113.941292244, 0.0 },
  { 55, 115, 114.935910,    0.0 },
  { 55, 116, 115.933395,    0.0 },
  { 55, 117, 116.928616723, 0.0 },
  { 55, 118, 117.926559517, 0.0 },
  { 55, 119, 118.922377327, 0.0 },
  { 55, 120, 119.920677277, 0.0 },
  { 55, 121, 120.917227235, 0.0 },
  { 55, 122, 121.916108144, 0.0 },
  { 55, 123, 122.912996060, 0.0 },
  { 55, 124, 123.912247366, 0.0 },
  { 55, 125, 124.909725953, 0.0 },
  { 55, 126, 125.909445821, 0.0 },
  { 55, 127, 126.907417527, 0.0 },
  { 55, 128, 127.907748452, 0.0 },
  { 55, 129, 128.906065910, 0.0 },
  { 55, 130, 129.906709281, 0.0 },
  { 55, 131, 130.905468457, 0.0 },
  { 55, 132, 131.906437740, 0.0 },
  { 55, 133, 132.905451958, 1.0 },
  { 55, 134, 133.906718501, 0.0 },
  { 55, 135, 134.905976907, 0.0 },
  { 55, 136, 135.907311431, 0.0 },
  { 55, 137, 136.907089296, 0.0 },
  { 55, 138, 137.911017119, 0.0 },
  { 55, 139, 138.913363822, 0.0 },
  { 55, 140, 139.917283707, 0.0 },
  { 55, 141, 140.920045279, 0.0 },
  { 55, 142, 141.924299514, 0.0 },
  { 55, 143, 142.927347346, 0.0 },
  { 55, 144, 143.932075402, 0.0 },
  { 55, 145, 144.935528927, 0.0 },
  { 55, 146, 145.940621867, 0.0 },
  { 55, 147, 146.944261512, 0.0 },
  { 55, 148, 147.949639026, 0.0 },
  { 55, 149, 148.953516,    0.0 },
  { 55, 150, 149.959023,    0.0 },
  { 55, 151, 150.963199,    0.0 },
  { 55, 152, 151.968728,    0.0 },
} },
{ 56, "Ba", "Barium", 137.327, 2.15, 3.03, 0.89, {
  { 56, 113, 112.957370,    0.0    },
  { 56, 114, 113.950718489, 0.0    },
  { 56, 115, 114.947482,    0.0    },
  { 56, 116, 115.941621,    0.0    },
  { 56, 117, 116.938316403, 0.0    },
  { 56, 118, 117.933226,    0.0    },
  { 56, 119, 118.930659683, 0.0    },
  { 56, 120, 119.926044997, 0.0    },
  { 56, 121, 120.924052286, 0.0    },
  { 56, 122, 121.919904000, 0.0    },
  { 56, 123, 122.918781060, 0.0    },
  { 56, 124, 123.915093627, 0.0    },
  { 56, 125, 124.914471840, 0.0    },
  { 56, 126, 125.911250202, 0.0    },
  { 56, 127, 126.911091272, 0.0    },
  { 56, 128, 127.908352446, 0.0    },
  { 56, 129, 128.908683409, 0.0    },
  { 56, 130, 129.906326002, 0.0011 },
  { 56, 131, 130.906946315, 0.0    },
  { 56, 132, 131.905061231, 0.0010 },
  { 56, 133, 132.906007443, 0.0    },
  { 56, 134, 133.904508249, 0.0242 },
  { 56, 135, 134.905688447, 0.0659 },
  { 56, 136, 135.904575800, 0.0785 },
  { 56, 137, 136.905827207, 0.1123 },
  { 56, 138, 137.905247059, 0.7170 },
  { 56, 139, 138.908841164, 0.0    },
  { 56, 140, 139.910608231, 0.0    },
  { 56, 141, 140.914403653, 0.0    },
  { 56, 142, 141.916432904, 0.0    },
  { 56, 143, 142.920625149, 0.0    },
  { 56, 144, 143.922954821, 0.0    },
  { 56, 145, 144.927518400, 0.0    },
  { 56, 146, 145.930363200, 0.0    },
  { 56, 147, 146.935303900, 0.0    },
  { 56, 148, 147.938223000, 0.0    },
  { 56, 149, 148.943284000, 0.0    },
  { 56, 150, 149.946441100, 0.0    },
  { 56, 151, 150.951755,    0.0    },
  { 56, 152, 151.955330,    0.0    },
  { 56, 153, 152.960848,    0.0    },
  { 56, 154, 153.964659,    0.0    },
} },
{ 57, "La", "Lanthanum", 138.90547, 2.07, 2.98, 1.10, {
  { 57, 116, 115.957005,    0.0       },
  { 57, 117, 116.950326,    0.0       },
  { 57, 118, 117.946731,    0.0       },
  { 57, 119, 118.940934,    0.0       },
  { 57, 120, 119.938196,    0.0       },
  { 57, 121, 120.933236,    0.0       },
  { 57, 122, 121.930710,    0.0       },
  { 57, 123, 122.926300,    0.0       },
  { 57, 124, 123.924574275, 0.0       },
  { 57, 125, 124.920815931, 0.0       },
  { 57, 126, 125.919512667, 0.0       },
  { 57, 127, 126.916375083, 0.0       },
  { 57, 128, 127.915592123, 0.0       },
  { 57, 129, 128.912695592, 0.0       },
  { 57, 130, 129.912369413, 0.0       },
  { 57, 131, 130.910070000, 0.0       },
  { 57, 132, 131.910119047, 0.0       },
  { 57, 133, 132.908218000, 0.0       },
  { 57, 134, 133.908514011, 0.0       },
  { 57, 135, 134.906984427, 0.0       },
  { 57, 136, 135.907634962, 0.0       },
  { 57, 137, 136.906450438, 0.0       },
  { 57, 138, 137.907124041, 0.0008881 },
  { 57, 139, 138.906362927, 0.9991119 },
  { 57, 140, 139.909487285, 0.0       },
  { 57, 141, 140.910971155, 0.0       },
  { 57, 142, 141.914090760, 0.0       },
  { 57, 143, 142.916079482, 0.0       },
  { 57, 144, 143.919645589, 0.0       },
  { 57, 145, 144.921808065, 0.0       },
  { 57, 146, 145.925688017, 0.0       },
  { 57, 147, 146.928417800, 0.0       },
  { 57, 148, 147.932679400, 0.0       },
  { 57, 149, 148.935351259, 0.0       },
  { 57, 150, 149.939547500, 0.0       },
  { 57, 151, 150.942769000, 0.0       },
  { 57, 152, 151.947085,    0.0       },
  { 57, 153, 152.950553,    0.0       },
  { 57, 154, 153.955416,    0.0       },
  { 57, 155, 154.959280,    0.0       },
  { 57, 156, 155.964519,    0.0       },
  { 57, 157, 156.968792,    0.0       },
} },
{ 58, "Ce", "Cerium", 140.116, 2.04, 2.88, 1.12, {
  { 58, 119, 118.952957,    0.0     },
  { 58, 120, 119.946613,    0.0     },
  { 58, 121, 120.943435,    0.0     },
  { 58, 122, 121.937870,    0.0     },
  { 58, 123, 122.935280,    0.0     },
  { 58, 124, 123.930310,    0.0     },
  { 58, 125, 124.928440,    0.0     },
  { 58, 126, 125.923971000, 0.0     },
  { 58, 127, 126.922727000, 0.0     },
  { 58, 128, 127.918911000, 0.0     },
  { 58, 129, 128.918102000, 0.0     },
  { 58, 130, 129.914736000, 0.0     },
  { 58, 131, 130.914429465, 0.0     },
  { 58, 132, 131.911466226, 0.0     },
  { 58, 133, 132.911520402, 0.0     },
  { 58, 134, 133.908928142, 0.0     },
  { 58, 135, 134.909160662, 0.0     },
  { 58, 136, 135.907129256, 0.00185 },
  { 58, 137, 136.907762416, 0.0     },
  { 58, 138, 137.905994180, 0.00251 },
  { 58, 139, 138.906647029, 0.0     },
  { 58, 140, 139.905448433, 0.88450 },
  { 58, 141, 140.908285991, 0.0     },
  { 58, 142, 141.909250208, 0.11114 },
  { 58, 143, 142.912391953, 0.0     },
  { 58, 144, 143.913652763, 0.0     },
  { 58, 145, 144.917265113, 0.0     },
  { 58, 146, 145.918812294, 0.0     },
  { 58, 147, 146.922689900, 0.0     },
  { 58, 148, 147.924424186, 0.0     },
  { 58, 149, 148.928426900, 0.0     },
  { 58, 150, 149.930384032, 0.0     },
  { 58, 151, 150.934272200, 0.0     },
  { 58, 152, 151.936682,    0.0     },
  { 58, 153, 152.941052,    0.0     },
  { 58, 154, 153.943940,    0.0     },
  { 58, 155, 154.948706,    0.0     },
  { 58, 156, 155.951884,    0.0     },
  { 58, 157, 156.957133,    0.0     },
  { 58, 158, 157.960773,    0.0     },
  { 58, 159, 158.966355,    0.0     },
} },
{ 59, "Pr", "Praseodymium", 140.90766, 2.03, 2.92, 1.13, {
  { 59, 121, 120.955393,    0.0 },
  { 59, 122, 121.951927,    0.0 },
  { 59, 123, 122.946076,    0.0 },
  { 59, 124, 123.942940,    0.0 },
  { 59, 125, 124.937659,    0.0 },
  { 59, 126, 125.935240,    0.0 },
  { 59, 127, 126.930710,    0.0 },
  { 59, 128, 127.928791000, 0.0 },
  { 59, 129, 128.925095000, 0.0 },
  { 59, 130, 129.923590000, 0.0 },
  { 59, 131, 130.920234960, 0.0 },
  { 59, 132, 131.919240000, 0.0 },
  { 59, 133, 132.916330558, 0.0 },
  { 59, 134, 133.915696729, 0.0 },
  { 59, 135, 134.913111772, 0.0 },
  { 59, 136, 135.912677470, 0.0 },
  { 59, 137, 136.910679183, 0.0 },
  { 59, 138, 137.910757495, 0.0 },
  { 59, 139, 138.908932700, 0.0 },
  { 59, 140, 139.909085600, 0.0 },
  { 59, 141, 140.907659604, 1.0 },
  { 59, 142, 141.910051640, 0.0 },
  { 59, 143, 142.910822624, 0.0 },
  { 59, 144, 143.913310682, 0.0 },
  { 59, 145, 144.914517987, 0.0 },
  { 59, 146, 145.917687630, 0.0 },
  { 59, 147, 146.919007438, 0.0 },
  { 59, 148, 147.922129992, 0.0 },
  { 59, 149, 148.923736100, 0.0 },
  { 59, 150, 149.926676391, 0.0 },
  { 59, 151, 150.928309066, 0.0 },
  { 59, 152, 151.931552900, 0.0 },
  { 59, 153, 152.933903511, 0.0 },
  { 59, 154, 153.937885165, 0.0 },
  { 59, 155, 154.940509193, 0.0 },
  { 59, 156, 155.944766900, 0.0 },
  { 59, 157, 156.948003100, 0.0 },
  { 59, 158, 157.952603,    0.0 },
  { 59, 159, 158.956232,    0.0 },
  { 59, 160, 159.961138,    0.0 },
  { 59, 161, 160.965121,    0.0 },
} },
{ 60, "Nd", "Neodymium", 144.242, 2.01, 2.95, 1.14, {
  { 60, 124, 123.951873,    0.0     },
  { 60, 125, 124.948395,    0.0     },
  { 60, 126, 125.942694,    0.0     },
  { 60, 127, 126.939978,    0.0     },
  { 60, 128, 127.935018,    0.0     },
  { 60, 129, 128.933038,    0.0     },
  { 60, 130, 129.928506000, 0.0     },
  { 60, 131, 130.927248020, 0.0     },
  { 60, 132, 131.923321237, 0.0     },
  { 60, 133, 132.922348000, 0.0     },
  { 60, 134, 133.918790207, 0.0     },
  { 60, 135, 134.918181318, 0.0     },
  { 60, 136, 135.914976061, 0.0     },
  { 60, 137, 136.914563099, 0.0     },
  { 60, 138, 137.911950938, 0.0     },
  { 60, 139, 138.911951208, 0.0     },
  { 60, 140, 139.909546130, 0.0     },
  { 60, 141, 140.909616690, 0.0     },
  { 60, 142, 141.907728824, 0.27152 },
  { 60, 143, 142.909819815, 0.12174 },
  { 60, 144, 143.910092798, 0.23798 },
  { 60, 145, 144.912579151, 0.08293 },
  { 60, 146, 145.913122459, 0.17189 },
  { 60, 147, 146.916105969, 0.0     },
  { 60, 148, 147.916899027, 0.05756 },
  { 60, 149, 148.920154583, 0.0     },
  { 60, 150, 149.920901322, 0.05638 },
  { 60, 151, 150.923839363, 0.0     },
  { 60, 152, 151.924691242, 0.0     },
  { 60, 153, 152.927717868, 0.0     },
  { 60, 154, 153.929597404, 0.0     },
  { 60, 155, 154.933135598, 0.0     },
  { 60, 156, 155.935370358, 0.0     },
  { 60, 157, 156.939351074, 0.0     },
  { 60, 158, 157.942205620, 0.0     },
  { 60, 159, 158.946619085, 0.0     },
  { 60, 160, 159.949839172, 0.0     },
  { 60, 161, 160.954664,    0.0     },
  { 60, 162, 161.958121,    0.0     },
  { 60, 163, 162.963414,    0.0     },
} },
{ 61, "Pm", "Promethium", 145, 1.99, 2.5, -1.0, {
  { 61, 126, 125.957327,    0.0 },
  { 61, 127, 126.951358,    0.0 },
  { 61, 128, 127.948234,    0.0 },
  { 61, 129, 128.942909,    0.0 },
  { 61, 130, 129.940451,    0.0 },
  { 61, 131, 130.935834,    0.0 },
  { 61, 132, 131.933840,    0.0 },
  { 61, 133, 132.929782000, 0.0 },
  { 61, 134, 133.928326000, 0.0 },
  { 61, 135, 134.924785000, 0.0 },
  { 61, 136, 135.923595949, 0.0 },
  { 61, 137, 136.920479519, 0.0 },
  { 61, 138, 137.919576119, 0.0 },
  { 61, 139, 138.916799228, 0.0 },
  { 61, 140, 139.916035918, 0.0 },
  { 61, 141, 140.913555081, 0.0 },
  { 61, 142, 141.912890982, 0.0 },
  { 61, 143, 142.910938068, 0.0 },
  { 61, 144, 143.912596208, 0.0 },
  { 61, 145, 144.912755748, 0.0 },
  { 61, 146, 145.914702240, 0.0 },
  { 61, 147, 146.915144944, 0.0 },
  { 61, 148, 147.917481091, 0.0 },
  { 61, 149, 148.918341507, 0.0 },
  { 61, 150, 149.920990014, 0.0 },
  { 61, 151, 150.921216613, 0.0 },
  { 61, 152, 151.923505185, 0.0 },
  { 61, 153, 152.924156252, 0.0 },
  { 61, 154, 153.926712791, 0.0 },
  { 61, 155, 154.928136951, 0.0 },
  { 61, 156, 155.931114059, 0.0 },
  { 61, 157, 156.933121298, 0.0 },
  { 61, 158, 157.936546948, 0.0 },
  { 61, 159, 158.939286409, 0.0 },
  { 61, 160, 159.943215272, 0.0 },
  { 61, 161, 160.946229837, 0.0 },
  { 61, 162, 161.950574,    0.0 },
  { 61, 163, 162.953881,    0.0 },
  { 61, 164, 163.958819,    0.0 },
  { 61, 165, 164.962780,    0.0 },
} },
{ 62, "Sm", "Samarium", 150.36, 1.98, 2.90, 1.17, {
  { 62, 128, 127.957971,    0.0    },
  { 62, 129, 128.954557,    0.0    },
  { 62, 130, 129.948792,    0.0    },
  { 62, 131, 130.946022,    0.0    },
  { 62, 132, 131.940805,    0.0    },
  { 62, 133, 132.938560,    0.0    },
  { 62, 134, 133.934110,    0.0    },
  { 62, 135, 134.932520000, 0.0    },
  { 62, 136, 135.928275553, 0.0    },
  { 62, 137, 136.927007959, 0.0    },
  { 62, 138, 137.923243988, 0.0    },
  { 62, 139, 138.922296631, 0.0    },
  { 62, 140, 139.918994714, 0.0    },
  { 62, 141, 140.918481545, 0.0    },
  { 62, 142, 141.915209415, 0.0    },
  { 62, 143, 142.914634848, 0.0    },
  { 62, 144, 143.912006285, 0.0308 },
  { 62, 145, 144.913417157, 0.0    },
  { 62, 146, 145.913046835, 0.0    },
  { 62, 147, 146.914904401, 0.1500 },
  { 62, 148, 147.914829233, 0.1125 },
  { 62, 149, 148.917191211, 0.1382 },
  { 62, 150, 149.917281993, 0.0737 },
  { 62, 151, 150.919938859, 0.0    },
  { 62, 152, 151.919738646, 0.2674 },
  { 62, 153, 152.922103576, 0.0    },
  { 62, 154, 153.922215756, 0.2274 },
  { 62, 155, 154.924646645, 0.0    },
  { 62, 156, 155.925538191, 0.0    },
  { 62, 157, 156.928418598, 0.0    },
  { 62, 158, 157.929949262, 0.0    },
  { 62, 159, 158.933217130, 0.0    },
  { 62, 160, 159.935337032, 0.0    },
  { 62, 161, 160.939160062, 0.0    },
  { 62, 162, 161.941621687, 0.0    },
  { 62, 163, 162.945679085, 0.0    },
  { 62, 164, 163.948550061, 0.0    },
  { 62, 165, 164.953290,    0.0    },
  { 62, 166, 165.956575,    0.0    },
  { 62, 167, 166.962072,    0.0    },
  { 62, 168, 167.966033,    0.0    },
} },
{ 63, "Eu", "Europium", 151.964, 1.98, 2.87, -1.0, {
  { 63, 130, 129.964022,    0.0    },
  { 63, 131, 130.957634,    0.0    },
  { 63, 132, 131.954696,    0.0    },
  { 63, 133, 132.949290,    0.0    },
  { 63, 134, 133.946537,    0.0    },
  { 63, 135, 134.941870,    0.0    },
  { 63, 136, 135.939620,    0.0    },
  { 63, 137, 136.935430719, 0.0    },
  { 63, 138, 137.933709000, 0.0    },
  { 63, 139, 138.929792307, 0.0    },
  { 63, 140, 139.928087633, 0.0    },
  { 63, 141, 140.924931734, 0.0    },
  { 63, 142, 141.923446719, 0.0    },
  { 63, 143, 142.920298678, 0.0    },
  { 63, 144, 143.918819481, 0.0    },
  { 63, 145, 144.916272659, 0.0    },
  { 63, 146, 145.917210852, 0.0    },
  { 63, 147, 146.916752440, 0.0    },
  { 63, 148, 147.918091288, 0.0    },
  { 63, 149, 148.917936875, 0.0    },
  { 63, 150, 149.919707092, 0.0    },
  { 63, 151, 150.919856606, 0.4781 },
  { 63, 152, 151.921750980, 0.0    },
  { 63, 153, 152.921236789, 0.5219 },
  { 63, 154, 153.922985699, 0.0    },
  { 63, 155, 154.922899847, 0.0    },
  { 63, 156, 155.924762976, 0.0    },
  { 63, 157, 156.925432556, 0.0    },
  { 63, 158, 157.927782192, 0.0    },
  { 63, 159, 158.929099512, 0.0    },
  { 63, 160, 159.931836982, 0.0    },
  { 63, 161, 160.933663991, 0.0    },
  { 63, 162, 161.936958329, 0.0    },
  { 63, 163, 162.939265510, 0.0    },
  { 63, 164, 163.942852943, 0.0    },
  { 63, 165, 164.945540070, 0.0    },
  { 63, 166, 165.949813,    0.0    },
  { 63, 167, 166.953011,    0.0    },
  { 63, 168, 167.957863,    0.0    },
  { 63, 169, 168.961717,    0.0    },
  { 63, 170, 169.966870,    0.0    },
} },
{ 64, "Gd", "Gadolinium", 157.25, 1.96, 2.83, 1.20, {
  { 64, 133, 132.961288,    0.0    },
  { 64, 134, 133.955416,    0.0    },
  { 64, 135, 134.952496,    0.0    },
  { 64, 136, 135.947300,    0.0    },
  { 64, 137, 136.945020,    0.0    },
  { 64, 138, 137.940247,    0.0    },
  { 64, 139, 138.938130,    0.0    },
  { 64, 140, 139.933674000, 0.0    },
  { 64, 141, 140.932126000, 0.0    },
  { 64, 142, 141.928116000, 0.0    },
  { 64, 143, 142.926750678, 0.0    },
  { 64, 144, 143.922963000, 0.0    },
  { 64, 145, 144.921710051, 0.0    },
  { 64, 146, 145.918318513, 0.0    },
  { 64, 147, 146.919101014, 0.0    },
  { 64, 148, 147.918121414, 0.0    },
  { 64, 149, 148.919347666, 0.0    },
  { 64, 150, 149.918663949, 0.0    },
  { 64, 151, 150.920354922, 0.0    },
  { 64, 152, 151.919798414, 0.0020 },
  { 64, 153, 152.921756945, 0.0    },
  { 64, 154, 153.920872974, 0.0218 },
  { 64, 155, 154.922629356, 0.1480 },
  { 64, 156, 155.922130120, 0.2047 },
  { 64, 157, 156.923967424, 0.1565 },
  { 64, 158, 157.924111200, 0.2484 },
  { 64, 159, 158.926395822, 0.0    },
  { 64, 160, 159.927061202, 0.2186 },
  { 64, 161, 160.929676267, 0.0    },
  { 64, 162, 161.930991812, 0.0    },
  { 64, 163, 162.934096640, 0.0    },
  { 64, 164, 163.935916193, 0.0    },
  { 64, 165, 164.939317080, 0.0    },
  { 64, 166, 165.941630413, 0.0    },
  { 64, 167, 166.945490012, 0.0    },
  { 64, 168, 167.948309,    0.0    },
  { 64, 169, 168.952882,    0.0    },
  { 64, 170, 169.956146,    0.0    },
  { 64, 171, 170.961127,    0.0    },
  { 64, 172, 171.964605,    0.0    },
} },
{ 65, "Tb", "Terbium", 158.925354, 1.94, 2.79, -1.0, {
  { 65, 135, 134.964516,    0.0 },
  { 65, 136, 135.961460,    0.0 },
  { 65, 137, 136.956020,    0.0 },
  { 65, 138, 137.953193,    0.0 },
  { 65, 139, 138.948330,    0.0 },
  { 65, 140, 139.945805048, 0.0 },
  { 65, 141, 140.941448000, 0.0 },
  { 65, 142, 141.939280858, 0.0 },
  { 65, 143, 142.935137332, 0.0 },
  { 65, 144, 143.933045000, 0.0 },
  { 65, 145, 144.928717001, 0.0 },
  { 65, 146, 145.927252739, 0.0 },
  { 65, 147, 146.924054620, 0.0 },
  { 65, 148, 147.924275476, 0.0 },
  { 65, 149, 148.923253792, 0.0 },
  { 65, 150, 149.923664799, 0.0 },
  { 65, 151, 150.923108970, 0.0 },
  { 65, 152, 151.924081855, 0.0 },
  { 65, 153, 152.923441694, 0.0 },
  { 65, 154, 153.924683681, 0.0 },
  { 65, 155, 154.923509511, 0.0 },
  { 65, 156, 155.924754209, 0.0 },
  { 65, 157, 156.924031888, 0.0 },
  { 65, 158, 157.925419942, 0.0 },
  { 65, 159, 158.925353707, 1.0 },
  { 65, 160, 159.927174553, 0.0 },
  { 65, 161, 160.927576806, 0.0 },
  { 65, 162, 161.929275400, 0.0 },
  { 65, 163, 162.930653609, 0.0 },
  { 65, 164, 163.933327561, 0.0 },
  { 65, 165, 164.934955198, 0.0 },
  { 65, 166, 165.937939727, 0.0 },
  { 65, 167, 166.940007046, 0.0 },
  { 65, 168, 167.943337074, 0.0 },
  { 65, 169, 168.945807,    0.0 },
  { 65, 170, 169.949855,    0.0 },
  { 65, 171, 170.953011,    0.0 },
  { 65, 172, 171.957391,    0.0 },
  { 65, 173, 172.960805,    0.0 },
  { 65, 174, 173.965679,    0.0 },
} },
{ 66, "Dy", "Dysprosium", 162.500, 1.92, 2.87, 1.22, {
  { 66, 138, 137.962500,    0.0     },
  { 66, 139, 138.959527,    0.0     },
  { 66, 140, 139.954020,    0.0     },
  { 66, 141, 140.951280,    0.0     },
  { 66, 142, 141.946194,    0.0     },
  { 66, 143, 142.943994332, 0.0     },
  { 66, 144, 143.939269512, 0.0     },
  { 66, 145, 144.937473992, 0.0     },
  { 66, 146, 145.932844526, 0.0     },
  { 66, 147, 146.931082712, 0.0     },
  { 66, 148, 147.927149944, 0.0     },
  { 66, 149, 148.927327516, 0.0     },
  { 66, 150, 149.925593068, 0.0     },
  { 66, 151, 150.926191279, 0.0     },
  { 66, 152, 151.924725274, 0.0     },
  { 66, 153, 152.925771729, 0.0     },
  { 66, 154, 153.924428920, 0.0     },
  { 66, 155, 154.925758049, 0.0     },
  { 66, 156, 155.924283593, 0.00056 },
  { 66, 157, 156.925469555, 0.0     },
  { 66, 158, 157.924414817, 0.00095 },
  { 66, 159, 158.925745938, 0.0     },
  { 66, 160, 159.925203578, 0.02329 },
  { 66, 161, 160.926939425, 0.18889 },
  { 66, 162, 161.926804507, 0.25475 },
  { 66, 163, 162.928737221, 0.24896 },
  { 66, 164, 163.929180819, 0.28260 },
  { 66, 165, 164.931709402, 0.0     },
  { 66, 166, 165.932812810, 0.0     },
  { 66, 167, 166.935682415, 0.0     },
  { 66, 168, 167.937134977, 0.0     },
  { 66, 169, 168.940315231, 0.0     },
  { 66, 170, 169.942340,    0.0     },
  { 66, 171, 170.946312,    0.0     },
  { 66, 172, 171.948728,    0.0     },
  { 66, 173, 172.953043,    0.0     },
  { 66, 174, 173.955845,    0.0     },
  { 66, 175, 174.960569,    0.0     },
  { 66, 176, 175.963918,    0.0     },
} },
{ 67, "Ho", "Holmium", 164.930329, 1.92, 2.81, 1.23, {
  { 67, 140, 139.968526,    0.0 },
  { 67, 141, 140.963108,    0.0 },
  { 67, 142, 141.960010,    0.0 },
  { 67, 143, 142.954860,    0.0 },
  { 67, 144, 143.952109712, 0.0 },
  { 67, 145, 144.947267392, 0.0 },
  { 67, 146, 145.944993503, 0.0 },
  { 67, 147, 146.940142293, 0.0 },
  { 67, 148, 147.937743925, 0.0 },
  { 67, 149, 148.933820457, 0.0 },
  { 67, 150, 149.933498353, 0.0 },
  { 67, 151, 150.931698176, 0.0 },
  { 67, 152, 151.931717618, 0.0 },
  { 67, 153, 152.930206671, 0.0 },
  { 67, 154, 153.930606776, 0.0 },
  { 67, 155, 154.929103363, 0.0 },
  { 67, 156, 155.929641634, 0.0 },
  { 67, 157, 156.928251974, 0.0 },
  { 67, 158, 157.928944910, 0.0 },
  { 67, 159, 158.927718683, 0.0 },
  { 67, 160, 159.928735538, 0.0 },
  { 67, 161, 160.927861815, 0.0 },
  { 67, 162, 161.929102543, 0.0 },
  { 67, 163, 162.928740260, 0.0 },
  { 67, 164, 163.930240548, 0.0 },
  { 67, 165, 164.930329116, 1.0 },
  { 67, 166, 165.932291209, 0.0 },
  { 67, 167, 166.933140254, 0.0 },
  { 67, 168, 167.935523766, 0.0 },
  { 67, 169, 168.936879890, 0.0 },
  { 67, 170, 169.939626548, 0.0 },
  { 67, 171, 170.941472713, 0.0 },
  { 67, 172, 171.944730,    0.0 },
  { 67, 173, 172.947020,    0.0 },
  { 67, 174, 173.950757,    0.0 },
  { 67, 175, 174.953516,    0.0 },
  { 67, 176, 175.957713,    0.0 },
  { 67, 177, 176.961052,    0.0 },
  { 67, 178, 177.965507,    0.0 },
} },
{ 68, "Er", "Erbium", 167.259, 1.89, 2.83, 1.24, {
  { 68, 142, 141.970016,    0.0     },
  { 68, 143, 142.966548,    0.0     },
  { 68, 144, 143.960700,    0.0     },
  { 68, 145, 144.957874,    0.0     },
  { 68, 146, 145.952418357, 0.0     },
  { 68, 147, 146.949964456, 0.0     },
  { 68, 148, 147.944735026, 0.0     },
  { 68, 149, 148.942306000, 0.0     },
  { 68, 150, 149.937915524, 0.0     },
  { 68, 151, 150.937448567, 0.0     },
  { 68, 152, 151.935050347, 0.0     },
  { 68, 153, 152.935086350, 0.0     },
  { 68, 154, 153.932790799, 0.0     },
  { 68, 155, 154.933215710, 0.0     },
  { 68, 156, 155.931065926, 0.0     },
  { 68, 157, 156.931922652, 0.0     },
  { 68, 158, 157.929893474, 0.0     },
  { 68, 159, 158.930690790, 0.0     },
  { 68, 160, 159.929077193, 0.0     },
  { 68, 161, 160.930003530, 0.0     },
  { 68, 162, 161.928787299, 0.00139 },
  { 68, 163, 162.930039908, 0.0     },
  { 68, 164, 163.929207739, 0.01601 },
  { 68, 165, 164.930733482, 0.0     },
  { 68, 166, 165.930301067, 0.33503 },
  { 68, 167, 166.932056192, 0.22869 },
  { 68, 168, 167.932378282, 0.26978 },
  { 68, 169, 168.934598444, 0.0     },
  { 68, 170, 169.935471933, 0.14910 },
  { 68, 171, 170.938037372, 0.0     },
  { 68, 172, 171.939363461, 0.0     },
  { 68, 173, 172.942400,    0.0     },
  { 68, 174, 173.944230,    0.0     },
  { 68, 175, 174.947770,    0.0     },
  { 68, 176, 175.949940,    0.0     },
  { 68, 177, 176.953990,    0.0     },
  { 68, 178, 177.956779,    0.0     },
  { 68, 179, 178.961267,    0.0     },
  { 68, 180, 179.964380,    0.0     },
} },
{ 69, "Tm", "Thulium", 168.934219, 1.90, 2.79, 1.25, {
  { 69, 144, 143.976211,    0.0 },
  { 69, 145, 144.970389,    0.0 },
  { 69, 146, 145.966661,    0.0 },
  { 69, 147, 146.961379887, 0.0 },
  { 69, 148, 147.958384026, 0.0 },
  { 69, 149, 148.952828,    0.0 },
  { 69, 150, 149.950090,    0.0 },
  { 69, 151, 150.945494433, 0.0 },
  { 69, 152, 151.944476000, 0.0 },
  { 69, 153, 152.942058023, 0.0 },
  { 69, 154, 153.941570062, 0.0 },
  { 69, 155, 154.939209576, 0.0 },
  { 69, 156, 155.938985746, 0.0 },
  { 69, 157, 156.936973000, 0.0 },
  { 69, 158, 157.936979525, 0.0 },
  { 69, 159, 158.934975000, 0.0 },
  { 69, 160, 159.935264177, 0.0 },
  { 69, 161, 160.933549000, 0.0 },
  { 69, 162, 161.934001211, 0.0 },
  { 69, 163, 162.932658282, 0.0 },
  { 69, 164, 163.933538019, 0.0 },
  { 69, 165, 164.932441843, 0.0 },
  { 69, 166, 165.933562136, 0.0 },
  { 69, 167, 166.932857206, 0.0 },
  { 69, 168, 167.934178457, 0.0 },
  { 69, 169, 168.934218956, 1.0 },
  { 69, 170, 169.935807093, 0.0 },
  { 69, 171, 170.936435162, 0.0 },
  { 69, 172, 171.938406959, 0.0 },
  { 69, 173, 172.939606630, 0.0 },
  { 69, 174, 173.942174061, 0.0 },
  { 69, 175, 174.943842310, 0.0 },
  { 69, 176, 175.946997707, 0.0 },
  { 69, 177, 176.948932,    0.0 },
  { 69, 178, 177.952506,    0.0 },
  { 69, 179, 178.955018,    0.0 },
  { 69, 180, 179.959023,    0.0 },
  { 69, 181, 180.961954,    0.0 },
  { 69, 182, 181.966194,    0.0 },
} },
{ 70, "Yb", "Ytterbium", 173.045, 1.87, 2.80, -1.0, {
  { 70, 148, 147.967547,    0.0     },
  { 70, 149, 148.964219,    0.0     },
  { 70, 150, 149.958314,    0.0     },
  { 70, 151, 150.955402453, 0.0     },
  { 70, 152, 151.950326699, 0.0     },
  { 70, 153, 152.949372,    0.0     },
  { 70, 154, 153.946395696, 0.0     },
  { 70, 155, 154.945783216, 0.0     },
  { 70, 156, 155.942817096, 0.0     },
  { 70, 157, 156.942651368, 0.0     },
  { 70, 158, 157.939871202, 0.0     },
  { 70, 159, 158.940060257, 0.0     },
  { 70, 160, 159.937559210, 0.0     },
  { 70, 161, 160.937912384, 0.0     },
  { 70, 162, 161.935779342, 0.0     },
  { 70, 163, 162.936345406, 0.0     },
  { 70, 164, 163.934500743, 0.0     },
  { 70, 165, 164.935270241, 0.0     },
  { 70, 166, 165.933876439, 0.0     },
  { 70, 167, 166.934954069, 0.0     },
  { 70, 168, 167.933891297, 0.00126 },
  { 70, 169, 168.935184208, 0.0     },
  { 70, 170, 169.934767242, 0.03023 },
  { 70, 171, 170.936331515, 0.14216 },
  { 70, 172, 171.936386654, 0.21754 },
  { 70, 173, 172.938216211, 0.16098 },
  { 70, 174, 173.938867545, 0.31896 },
  { 70, 175, 174.941281907, 0.0     },
  { 70, 176, 175.942574706, 0.12887 },
  { 70, 177, 176.945263846, 0.0     },
  { 70, 178, 177.946669400, 0.0     },
  { 70, 179, 178.949930,    0.0     },
  { 70, 180, 179.951991,    0.0     },
  { 70, 181, 180.955890,    0.0     },
  { 70, 182, 181.958239,    0.0     },
  { 70, 183, 182.962426,    0.0     },
  { 70, 184, 183.965002,    0.0     },
  { 70, 185, 184.969425,    0.0     },
} },
{ 71, "Lu", "Lutetium", 174.9668, 1.87, 2.74, 1.27, {
  { 71, 150, 149.973407,    0.0     },
  { 71, 151, 150.967471,    0.0     },
  { 71, 152, 151.964120,    0.0     },
  { 71, 153, 152.958802248, 0.0     },
  { 71, 154, 153.957416,    0.0     },
  { 71, 155, 154.954326005, 0.0     },
  { 71, 156, 155.953086606, 0.0     },
  { 71, 157, 156.950144807, 0.0     },
  { 71, 158, 157.949315620, 0.0     },
  { 71, 159, 158.946635615, 0.0     },
  { 71, 160, 159.946033000, 0.0     },
  { 71, 161, 160.943572000, 0.0     },
  { 71, 162, 161.943282776, 0.0     },
  { 71, 163, 162.941179000, 0.0     },
  { 71, 164, 163.941339000, 0.0     },
  { 71, 165, 164.939406758, 0.0     },
  { 71, 166, 165.939859000, 0.0     },
  { 71, 167, 166.938243000, 0.0     },
  { 71, 168, 167.938729798, 0.0     },
  { 71, 169, 168.937645845, 0.0     },
  { 71, 170, 169.938479230, 0.0     },
  { 71, 171, 170.937918591, 0.0     },
  { 71, 172, 171.939091320, 0.0     },
  { 71, 173, 172.938935722, 0.0     },
  { 71, 174, 173.940342840, 0.0     },
  { 71, 175, 174.940777211, 0.97401 },
  { 71, 176, 175.942691711, 0.02599 },
  { 71, 177, 176.943763570, 0.0     },
  { 71, 178, 177.945960065, 0.0     },
  { 71, 179, 178.947332985, 0.0     },
  { 71, 180, 179.949890744, 0.0     },
  { 71, 181, 180.951908000, 0.0     },
  { 71, 182, 181.955158,    0.0     },
  { 71, 183, 182.957363000, 0.0     },
  { 71, 184, 183.961030,    0.0     },
  { 71, 185, 184.963542,    0.0     },
  { 71, 186, 185.967450,    0.0     },
  { 71, 187, 186.970188,    0.0     },
  { 71, 188, 187.974428,    0.0     },
} },
{ 72, "Hf", "Hafnium", 178.486, 1.75, 2.63, 1.3, {
  { 72, 153, 152.970692,    0.0     },
  { 72, 154, 153.964863,    0.0     },
  { 72, 155, 154.963167,    0.0     },
  { 72, 156, 155.959399083, 0.0     },
  { 72, 157, 156.958288,    0.0     },
  { 72, 158, 157.954801217, 0.0     },
  { 72, 159, 158.953995837, 0.0     },
  { 72, 160, 159.950682728, 0.0     },
  { 72, 161, 160.950277927, 0.0     },
  { 72, 162, 161.947215526, 0.0     },
  { 72, 163, 162.947107211, 0.0     },
  { 72, 164, 163.944370709, 0.0     },
  { 72, 165, 164.944567000, 0.0     },
  { 72, 166, 165.942180000, 0.0     },
  { 72, 167, 166.942600000, 0.0     },
  { 72, 168, 167.940568000, 0.0     },
  { 72, 169, 168.941259000, 0.0     },
  { 72, 170, 169.939609000, 0.0     },
  { 72, 171, 170.940492000, 0.0     },
  { 72, 172, 171.939449716, 0.0     },
  { 72, 173, 172.940513000, 0.0     },
  { 72, 174, 173.940048377, 0.00161 },
  { 72, 175, 174.941511424, 0.0     },
  { 72, 176, 175.941409797, 0.0524  },
  { 72, 177, 176.943230187, 0.1858  },
  { 72, 178, 177.943708322, 0.2728  },
  { 72, 179, 178.945825705, 0.1363  },
  { 72, 180, 179.946559537, 0.3512  },
  { 72, 181, 180.949110834, 0.0     },
  { 72, 182, 181.950563684, 0.0     },
  { 72, 183, 182.953533203, 0.0     },
  { 72, 184, 183.955448507, 0.0     },
  { 72, 185, 184.958862000, 0.0     },
  { 72, 186, 185.960897000, 0.0     },
  { 72, 187, 186.964573,    0.0     },
  { 72, 188, 187.966903,    0.0     },
  { 72, 189, 188.970853,    0.0     },
  { 72, 190, 189.973376,    0.0     },
} },
{ 73, "Ta", "Tantalum", 180.94788, 1.70, 2.53, 1.5, {
  { 73, 155, 154.974248,    0.0       },
  { 73, 156, 155.972087,    0.0       },
  { 73, 157, 156.968227445, 0.0       },
  { 73, 158, 157.966593,    0.0       },
  { 73, 159, 158.963028046, 0.0       },
  { 73, 160, 159.961541678, 0.0       },
  { 73, 161, 160.958369489, 0.0       },
  { 73, 162, 161.957292907, 0.0       },
  { 73, 163, 162.954337194, 0.0       },
  { 73, 164, 163.953534000, 0.0       },
  { 73, 165, 164.950780287, 0.0       },
  { 73, 166, 165.950512000, 0.0       },
  { 73, 167, 166.948093000, 0.0       },
  { 73, 168, 167.948047000, 0.0       },
  { 73, 169, 168.946011000, 0.0       },
  { 73, 170, 169.946175000, 0.0       },
  { 73, 171, 170.944476000, 0.0       },
  { 73, 172, 171.944895000, 0.0       },
  { 73, 173, 172.943750000, 0.0       },
  { 73, 174, 173.944454000, 0.0       },
  { 73, 175, 174.943737000, 0.0       },
  { 73, 176, 175.944857000, 0.0       },
  { 73, 177, 176.944481940, 0.0       },
  { 73, 178, 177.945680,    0.0       },
  { 73, 179, 178.945939050, 0.0       },
  { 73, 180, 179.947467589, 0.0001176 },
  { 73, 181, 180.947998528, 0.9998824 },
  { 73, 182, 181.950154612, 0.0       },
  { 73, 183, 182.951375380, 0.0       },
  { 73, 184, 183.954009958, 0.0       },
  { 73, 185, 184.955561317, 0.0       },
  { 73, 186, 185.958553036, 0.0       },
  { 73, 187, 186.960391000, 0.0       },
  { 73, 188, 187.963596,    0.0       },
  { 73, 189, 188.965690,    0.0       },
  { 73, 190, 189.969168,    0.0       },
  { 73, 191, 190.971530,    0.0       },
  { 73, 192, 191.975201,    0.0       },
  { 73, 193, 192.977660,    0.0       },
  { 73, 194, 193.981610,    0.0       },
} },
{ 74, "W", "Tungsten", 183.84, 1.62, 2.57, 2.36, {
  { 74, 157, 156.978862,    0.0    },
  { 74, 158, 157.974565,    0.0    },
  { 74, 159, 158.972696,    0.0    },
  { 74, 160, 159.968513946, 0.0    },
  { 74, 161, 160.967249,    0.0    },
  { 74, 162, 161.963500341, 0.0    },
  { 74, 163, 162.962524251, 0.0    },
  { 74, 164, 163.958952445, 0.0    },
  { 74, 165, 164.958280663, 0.0    },
  { 74, 166, 165.955031952, 0.0    },
  { 74, 167, 166.954811080, 0.0    },
  { 74, 168, 167.951805459, 0.0    },
  { 74, 169, 168.951778689, 0.0    },
  { 74, 170, 169.949231235, 0.0    },
  { 74, 171, 170.949451000, 0.0    },
  { 74, 172, 171.947292000, 0.0    },
  { 74, 173, 172.947689000, 0.0    },
  { 74, 174, 173.946079000, 0.0    },
  { 74, 175, 174.946717000, 0.0    },
  { 74, 176, 175.945634000, 0.0    },
  { 74, 177, 176.946643000, 0.0    },
  { 74, 178, 177.945885791, 0.0    },
  { 74, 179, 178.947079378, 0.0    },
  { 74, 180, 179.946713304, 0.0012 },
  { 74, 181, 180.948218733, 0.0    },
  { 74, 182, 181.948205636, 0.2650 },
  { 74, 183, 182.950224416, 0.1431 },
  { 74, 184, 183.950933180, 0.3064 },
  { 74, 185, 184.953421206, 0.0    },
  { 74, 186, 185.954365140, 0.2843 },
  { 74, 187, 186.957161249, 0.0    },
  { 74, 188, 187.958488325, 0.0    },
  { 74, 189, 188.961557,    0.0    },
  { 74, 190, 189.963103542, 0.0    },
  { 74, 191, 190.966531000, 0.0    },
  { 74, 192, 191.968202,    0.0    },
  { 74, 193, 192.971884,    0.0    },
  { 74, 194, 193.973795,    0.0    },
  { 74, 195, 194.977735,    0.0    },
  { 74, 196, 195.979882,    0.0    },
  { 74, 197, 196.984036,    0.0    },
} },
{ 75, "Re", "Rhenium", 186.207, 1.51, 2.49, 1.9, {
  { 75, 159, 158.984106,    0.0    },
  { 75, 160, 159.981880,    0.0    },
  { 75, 161, 160.977624313, 0.0    },
  { 75, 162, 161.975896,    0.0    },
  { 75, 163, 162.972085434, 0.0    },
  { 75, 164, 163.970507122, 0.0    },
  { 75, 165, 164.967085831, 0.0    },
  { 75, 166, 165.965821216, 0.0    },
  { 75, 167, 166.962604,    0.0    },
  { 75, 168, 167.961572607, 0.0    },
  { 75, 169, 168.958765979, 0.0    },
  { 75, 170, 169.958234844, 0.0    },
  { 75, 171, 170.955716000, 0.0    },
  { 75, 172, 171.955376165, 0.0    },
  { 75, 173, 172.953243000, 0.0    },
  { 75, 174, 173.953115000, 0.0    },
  { 75, 175, 174.951381000, 0.0    },
  { 75, 176, 175.951623000, 0.0    },
  { 75, 177, 176.950328000, 0.0    },
  { 75, 178, 177.950989000, 0.0    },
  { 75, 179, 178.949989686, 0.0    },
  { 75, 180, 179.950791568, 0.0    },
  { 75, 181, 180.950061507, 0.0    },
  { 75, 182, 181.951211560, 0.0    },
  { 75, 183, 182.950821306, 0.0    },
  { 75, 184, 183.952528073, 0.0    },
  { 75, 185, 184.952958320, 0.3740 },
  { 75, 186, 185.954989172, 0.0    },
  { 75, 187, 186.955752217, 0.6260 },
  { 75, 188, 187.958113658, 0.0    },
  { 75, 189, 188.959227764, 0.0    },
  { 75, 190, 189.961800064, 0.0    },
  { 75, 191, 190.963123322, 0.0    },
  { 75, 192, 191.966088000, 0.0    },
  { 75, 193, 192.967545000, 0.0    },
  { 75, 194, 193.970735,    0.0    },
  { 75, 195, 194.972560,    0.0    },
  { 75, 196, 195.975996,    0.0    },
  { 75, 197, 196.978153,    0.0    },
  { 75, 198, 197.981760,    0.0    },
  { 75, 199, 198.984187,    0.0    },
} },
{ 76, "Os", "Osmium", 190.23, 1.44, 2.48, 2.2, {
  { 76, 161, 160.989054,    0.0    },
  { 76, 162, 161.984434,    0.0    },
  { 76, 163, 162.982462,    0.0    },
  { 76, 164, 163.978073158, 0.0    },
  { 76, 165, 164.976654,    0.0    },
  { 76, 166, 165.972698135, 0.0    },
  { 76, 167, 166.971552304, 0.0    },
  { 76, 168, 167.967799050, 0.0    },
  { 76, 169, 168.967017521, 0.0    },
  { 76, 170, 169.963579273, 0.0    },
  { 76, 171, 170.963180402, 0.0    },
  { 76, 172, 171.960017309, 0.0    },
  { 76, 173, 172.959808387, 0.0    },
  { 76, 174, 173.957063192, 0.0    },
  { 76, 175, 174.956945126, 0.0    },
  { 76, 176, 175.954770315, 0.0    },
  { 76, 177, 176.954957902, 0.0    },
  { 76, 178, 177.953253334, 0.0    },
  { 76, 179, 178.953815985, 0.0    },
  { 76, 180, 179.952381665, 0.0    },
  { 76, 181, 180.953247188, 0.0    },
  { 76, 182, 181.952110154, 0.0    },
  { 76, 183, 182.953125028, 0.0    },
  { 76, 184, 183.952492919, 0.0002 },
  { 76, 185, 184.954045969, 0.0    },
  { 76, 186, 185.953837569, 0.0159 },
  { 76, 187, 186.955749569, 0.0196 },
  { 76, 188, 187.955837292, 0.1324 },
  { 76, 189, 188.958145949, 0.1615 },
  { 76, 190, 189.958445442, 0.2626 },
  { 76, 191, 190.960928105, 0.0    },
  { 76, 192, 191.961478765, 0.4078 },
  { 76, 193, 192.964149637, 0.0    },
  { 76, 194, 193.965179407, 0.0    },
  { 76, 195, 194.968318000, 0.0    },
  { 76, 196, 195.969643261, 0.0    },
  { 76, 197, 196.973076,    0.0    },
  { 76, 198, 197.974664,    0.0    },
  { 76, 199, 198.978239,    0.0    },
  { 76, 200, 199.980086,    0.0    },
  { 76, 201, 200.984069,    0.0    },
  { 76, 202, 201.986548,    0.0    },
  { 76, 203, 202.992195,    0.0    },
} },
{ 77, "Ir", "Iridium", 192.217, 1.41, 2.41, 2.20, {
  { 77, 163, 162.994299,    0.0    },
  { 77, 164, 163.991966,    0.0    },
  { 77, 165, 164.987552,    0.0    },
  { 77, 166, 165.985716,    0.0    },
  { 77, 167, 166.981671973, 0.0    },
  { 77, 168, 167.979960978, 0.0    },
  { 77, 169, 168.976281743, 0.0    },
  { 77, 170, 169.975113,    0.0    },
  { 77, 171, 170.971645520, 0.0    },
  { 77, 172, 171.970607035, 0.0    },
  { 77, 173, 172.967505477, 0.0    },
  { 77, 174, 173.966949939, 0.0    },
  { 77, 175, 174.964149519, 0.0    },
  { 77, 176, 175.963626261, 0.0    },
  { 77, 177, 176.961301500, 0.0    },
  { 77, 178, 177.961079395, 0.0    },
  { 77, 179, 178.959117594, 0.0    },
  { 77, 180, 179.959229446, 0.0    },
  { 77, 181, 180.957634691, 0.0    },
  { 77, 182, 181.958076296, 0.0    },
  { 77, 183, 182.956841231, 0.0    },
  { 77, 184, 183.957476000, 0.0    },
  { 77, 185, 184.956698000, 0.0    },
  { 77, 186, 185.957946754, 0.0    },
  { 77, 187, 186.957542000, 0.0    },
  { 77, 188, 187.958834999, 0.0    },
  { 77, 189, 188.958722602, 0.0    },
  { 77, 190, 189.960543374, 0.0    },
  { 77, 191, 190.960591455, 0.3723 },
  { 77, 192, 191.962602414, 0.0    },
  { 77, 193, 192.962923753, 0.6277 },
  { 77, 194, 193.965075703, 0.0    },
  { 77, 195, 194.965976898, 0.0    },
  { 77, 196, 195.968399669, 0.0    },
  { 77, 197, 196.969657217, 0.0    },
  { 77, 198, 197.972399,    0.0    },
  { 77, 199, 198.973807097, 0.0    },
  { 77, 200, 199.976844,    0.0    },
  { 77, 201, 200.978701,    0.0    },
  { 77, 202, 201.982136,    0.0    },
  { 77, 203, 202.984573,    0.0    },
  { 77, 204, 203.989726,    0.0    },
  { 77, 205, 204.993988,    0.0    },
} },
{ 78, "Pt", "Platinum", 195.084, 1.36, 2.29, 2.28, {
  { 78, 165, 164.999658,    0.0     },
  { 78, 166, 165.994866,    0.0     },
  { 78, 167, 166.992750,    0.0     },
  { 78, 168, 167.988180196, 0.0     },
  { 78, 169, 168.986619,    0.0     },
  { 78, 170, 169.982502087, 0.0     },
  { 78, 171, 170.981248868, 0.0     },
  { 78, 172, 171.977341059, 0.0     },
  { 78, 173, 172.976449922, 0.0     },
  { 78, 174, 173.972820431, 0.0     },
  { 78, 175, 174.972400593, 0.0     },
  { 78, 176, 175.968938162, 0.0     },
  { 78, 177, 176.968469541, 0.0     },
  { 78, 178, 177.965649288, 0.0     },
  { 78, 179, 178.965358742, 0.0     },
  { 78, 180, 179.963038010, 0.0     },
  { 78, 181, 180.963089946, 0.0     },
  { 78, 182, 181.961171605, 0.0     },
  { 78, 183, 182.961595895, 0.0     },
  { 78, 184, 183.959921929, 0.0     },
  { 78, 185, 184.960613659, 0.0     },
  { 78, 186, 185.959350845, 0.0     },
  { 78, 187, 186.960616646, 0.0     },
  { 78, 188, 187.959397521, 0.0     },
  { 78, 189, 188.960848485, 0.0     },
  { 78, 190, 189.959949823, 0.00012 },
  { 78, 191, 190.961676261, 0.0     },
  { 78, 192, 191.961042667, 0.00782 },
  { 78, 193, 192.962984546, 0.0     },
  { 78, 194, 193.962683498, 0.32864 },
  { 78, 195, 194.964794325, 0.33775 },
  { 78, 196, 195.964954648, 0.25211 },
  { 78, 197, 196.967343030, 0.0     },
  { 78, 198, 197.967896718, 0.07356 },
  { 78, 199, 198.970597022, 0.0     },
  { 78, 200, 199.971444609, 0.0     },
  { 78, 201, 200.974513305, 0.0     },
  { 78, 202, 201.975639000, 0.0     },
  { 78, 203, 202.979055,    0.0     },
  { 78, 204, 203.981084,    0.0     },
  { 78, 205, 204.986237,    0.0     },
  { 78, 206, 205.990080,    0.0     },
  { 78, 207, 206.995556,    0.0     },
  { 78, 208, 207.999463,    0.0     },
} },
{ 79, "Au", "Gold", 196.966570, 1.36, 2.32, 2.54, {
  { 79, 168, 168.002716,    0.0 },
  { 79, 169, 168.998080,    0.0 },
  { 79, 170, 169.996024,    0.0 },
  { 79, 171, 170.991881533, 0.0 },
  { 79, 172, 171.989996704, 0.0 },
  { 79, 173, 172.986224263, 0.0 },
  { 79, 174, 173.984908,    0.0 },
  { 79, 175, 174.981316375, 0.0 },
  { 79, 176, 175.980116925, 0.0 },
  { 79, 177, 176.976869701, 0.0 },
  { 79, 178, 177.976056714, 0.0 },
  { 79, 179, 178.973173666, 0.0 },
  { 79, 180, 179.972489738, 0.0 },
  { 79, 181, 180.970079102, 0.0 },
  { 79, 182, 181.969614433, 0.0 },
  { 79, 183, 182.967588106, 0.0 },
  { 79, 184, 183.967451523, 0.0 },
  { 79, 185, 184.965798871, 0.0 },
  { 79, 186, 185.965952703, 0.0 },
  { 79, 187, 186.964542147, 0.0 },
  { 79, 188, 187.965247966, 0.0 },
  { 79, 189, 188.963948286, 0.0 },
  { 79, 190, 189.964751746, 0.0 },
  { 79, 191, 190.963716452, 0.0 },
  { 79, 192, 191.964817615, 0.0 },
  { 79, 193, 192.964138442, 0.0 },
  { 79, 194, 193.965419051, 0.0 },
  { 79, 195, 194.965037823, 0.0 },
  { 79, 196, 195.966571213, 0.0 },
  { 79, 197, 196.966570103, 1.0 },
  { 79, 198, 197.968243714, 0.0 },
  { 79, 199, 198.968766573, 0.0 },
  { 79, 200, 199.970756558, 0.0 },
  { 79, 201, 200.971657678, 0.0 },
  { 79, 202, 201.973856000, 0.0 },
  { 79, 203, 202.975154492, 0.0 },
  { 79, 204, 203.978110,    0.0 },
  { 79, 205, 204.980064,    0.0 },
  { 79, 206, 205.984766,    0.0 },
  { 79, 207, 206.988577,    0.0 },
  { 79, 208, 207.993655,    0.0 },
  { 79, 209, 208.997606,    0.0 },
  { 79, 210, 210.002877,    0.0 },
} },
{ 80, "Hg", "Mercury", 200.592, 1.32, 2.45, 2.00, {
  { 80, 170, 170.005814,    0.0    },
  { 80, 171, 171.003585,    0.0    },
  { 80, 172, 171.998860581, 0.0    },
  { 80, 173, 172.997143,    0.0    },
  { 80, 174, 173.992870575, 0.0    },
  { 80, 175, 174.991444451, 0.0    },
  { 80, 176, 175.987348670, 0.0    },
  { 80, 177, 176.986284590, 0.0    },
  { 80, 178, 177.982484756, 0.0    },
  { 80, 179, 178.981821759, 0.0    },
  { 80, 180, 179.978260180, 0.0    },
  { 80, 181, 180.977819368, 0.0    },
  { 80, 182, 181.974689173, 0.0    },
  { 80, 183, 182.974444652, 0.0    },
  { 80, 184, 183.971717709, 0.0    },
  { 80, 185, 184.971890696, 0.0    },
  { 80, 186, 185.969362061, 0.0    },
  { 80, 187, 186.969813540, 0.0    },
  { 80, 188, 187.967580738, 0.0    },
  { 80, 189, 188.968194776, 0.0    },
  { 80, 190, 189.966322250, 0.0    },
  { 80, 191, 190.967158301, 0.0    },
  { 80, 192, 191.965634263, 0.0    },
  { 80, 193, 192.966653395, 0.0    },
  { 80, 194, 193.965449108, 0.0    },
  { 80, 195, 194.966705809, 0.0    },
  { 80, 196, 195.965833445, 0.0015 },
  { 80, 197, 196.967213715, 0.0    },
  { 80, 198, 197.966769177, 0.1004 },
  { 80, 199, 198.968280994, 0.1694 },
  { 80, 200, 199.968326941, 0.2314 },
  { 80, 201, 200.970303054, 0.1317 },
  { 80, 202, 201.970643604, 0.2974 },
  { 80, 203, 202.972872396, 0.0    },
  { 80, 204, 203.973494037, 0.0682 },
  { 80, 205, 204.976073151, 0.0    },
  { 80, 206, 205.977513837, 0.0    },
  { 80, 207, 206.982300000, 0.0    },
  { 80, 208, 207.985759000, 0.0    },
  { 80, 209, 208.990757,    0.0    },
  { 80, 210, 209.994310,    0.0    },
  { 80, 211, 210.999581,    0.0    },
  { 80, 212, 212.003242,    0.0    },
  { 80, 213, 213.008803,    0.0    },
  { 80, 214, 214.012636,    0.0    },
  { 80, 215, 215.018368,    0.0    },
  { 80, 216, 216.022459,    0.0    },
} },
{ 81, "Tl", "Thallium", 204.38, 1.45, 2.47, 1.62, {
  { 81, 176, 176.000627731, 0.0       },
  { 81, 177, 176.996414252, 0.0       },
  { 81, 178, 177.995047,    0.0       },
  { 81, 179, 178.991122185, 0.0       },
  { 81, 180, 179.989918950, 0.0       },
  { 81, 181, 180.986259978, 0.0       },
  { 81, 182, 181.985692649, 0.0       },
  { 81, 183, 182.982192843, 0.0       },
  { 81, 184, 183.981874973, 0.0       },
  { 81, 185, 184.978789189, 0.0       },
  { 81, 186, 185.978654787, 0.0       },
  { 81, 187, 186.975904740, 0.0       },
  { 81, 188, 187.976020886, 0.0       },
  { 81, 189, 188.973573525, 0.0       },
  { 81, 190, 189.973841771, 0.0       },
  { 81, 191, 190.971784093, 0.0       },
  { 81, 192, 191.972225000, 0.0       },
  { 81, 193, 192.970501994, 0.0       },
  { 81, 194, 193.971081408, 0.0       },
  { 81, 195, 194.969774052, 0.0       },
  { 81, 196, 195.970481189, 0.0       },
  { 81, 197, 196.969560492, 0.0       },
  { 81, 198, 197.970446669, 0.0       },
  { 81, 199, 198.969877000, 0.0       },
  { 81, 200, 199.970963608, 0.0       },
  { 81, 201, 200.970820235, 0.0       },
  { 81, 202, 201.972108874, 0.0       },
  { 81, 203, 202.972344098, 0.2951544 },
  { 81, 204, 203.973863420, 0.0       },
  { 81, 205, 204.974427318, 0.7048544 },
  { 81, 206, 205.976110108, 0.0       },
  { 81, 207, 206.977418605, 0.0       },
  { 81, 208, 207.982018006, 0.0       },
  { 81, 209, 208.985351713, 0.0       },
  { 81, 210, 209.990072942, 0.0       },
  { 81, 211, 210.993475000, 0.0       },
  { 81, 212, 211.998335,    0.0       },
  { 81, 213, 213.001915000, 0.0       },
  { 81, 214, 214.006940,    0.0       },
  { 81, 215, 215.010768,    0.0       },
  { 81, 216, 216.015964,    0.0       },
  { 81, 217, 217.020032,    0.0       },
  { 81, 218, 218.025454,    0.0       },
} },
{ 82, "Pb", "Lead", 207.2, 1.46, 2.60, 2.33, {
  { 82, 178, 178.003836171, 0.0    },
  { 82, 179, 179.002202492, 0.0    },
  { 82, 180, 179.997916177, 0.0    },
  { 82, 181, 180.996660600, 0.0    },
  { 82, 182, 181.992673537, 0.0    },
  { 82, 183, 182.991862527, 0.0    },
  { 82, 184, 183.988135634, 0.0    },
  { 82, 185, 184.987610000, 0.0    },
  { 82, 186, 185.984239409, 0.0    },
  { 82, 187, 186.983910842, 0.0    },
  { 82, 188, 187.980879079, 0.0    },
  { 82, 189, 188.980843658, 0.0    },
  { 82, 190, 189.978081872, 0.0    },
  { 82, 191, 190.978216455, 0.0    },
  { 82, 192, 191.975789598, 0.0    },
  { 82, 193, 192.976135914, 0.0    },
  { 82, 194, 193.974011788, 0.0    },
  { 82, 195, 194.974516167, 0.0    },
  { 82, 196, 195.972787552, 0.0    },
  { 82, 197, 196.973434737, 0.0    },
  { 82, 198, 197.972015450, 0.0    },
  { 82, 199, 198.972912620, 0.0    },
  { 82, 200, 199.971818546, 0.0    },
  { 82, 201, 200.972870431, 0.0    },
  { 82, 202, 201.972151613, 0.0    },
  { 82, 203, 202.973390617, 0.0    },
  { 82, 204, 203.973043506, 0.0146 },
  { 82, 205, 204.974481682, 0.0    },
  { 82, 206, 205.974465210, 0.2413 },
  { 82, 207, 206.975896821, 0.2215 },
  { 82, 208, 207.976652005, 0.5247 },
  { 82, 209, 208.981089978, 0.0    },
  { 82, 210, 209.984188381, 0.0    },
  { 82, 211, 210.988735288, 0.0    },
  { 82, 212, 211.991895891, 0.0    },
  { 82, 213, 212.996560796, 0.0    },
  { 82, 214, 213.999803521, 0.0    },
  { 82, 215, 215.004661591, 0.0    },
  { 82, 216, 216.008062,    0.0    },
  { 82, 217, 217.013162,    0.0    },
  { 82, 218, 218.016779,    0.0    },
  { 82, 219, 219.022136,    0.0    },
  { 82, 220, 220.025905,    0.0    },
} },
{ 83, "Bi", "Bismuth", 208.98040, 1.48, 2.54, 2.02, {
  { 83, 184, 184.001347,    0.0 },
  { 83, 185, 184.997600,    0.0 },
  { 83, 186, 185.996623169, 0.0 },
  { 83, 187, 186.993147272, 0.0 },
  { 83, 188, 187.992276064, 0.0 },
  { 83, 189, 188.989195139, 0.0 },
  { 83, 190, 189.988624828, 0.0 },
  { 83, 191, 190.985786972, 0.0 },
  { 83, 192, 191.985470077, 0.0 },
  { 83, 193, 192.982947220, 0.0 },
  { 83, 194, 193.982798581, 0.0 },
  { 83, 195, 194.980648759, 0.0 },
  { 83, 196, 195.980666509, 0.0 },
  { 83, 197, 196.978864927, 0.0 },
  { 83, 198, 197.979201316, 0.0 },
  { 83, 199, 198.977672841, 0.0 },
  { 83, 200, 199.978131290, 0.0 },
  { 83, 201, 200.976995017, 0.0 },
  { 83, 202, 201.977723042, 0.0 },
  { 83, 203, 202.976892077, 0.0 },
  { 83, 204, 203.977835687, 0.0 },
  { 83, 205, 204.977385182, 0.0 },
  { 83, 206, 205.978498843, 0.0 },
  { 83, 207, 206.978470551, 0.0 },
  { 83, 208, 207.979742060, 0.0 },
  { 83, 209, 208.980398599, 1.0 },
  { 83, 210, 209.984120237, 0.0 },
  { 83, 211, 210.987268715, 0.0 },
  { 83, 212, 211.991285030, 0.0 },
  { 83, 213, 212.994383570, 0.0 },
  { 83, 214, 213.998710909, 0.0 },
  { 83, 215, 215.001749095, 0.0 },
  { 83, 216, 216.006305985, 0.0 },
  { 83, 217, 217.009372000, 0.0 },
  { 83, 218, 218.014188000, 0.0 },
  { 83, 219, 219.017520,    0.0 },
  { 83, 220, 220.022501,    0.0 },
  { 83, 221, 221.025980,    0.0 },
  { 83, 222, 222.031079,    0.0 },
  { 83, 223, 223.034611,    0.0 },
  { 83, 224, 224.039796,    0.0 },
} },
{ 84, "Po", "Polonium", 209, 1.40, 2.5, 2.0, {
  { 84, 186, 186.004403174, 0.0 },
  { 84, 187, 187.003031482, 0.0 },
  { 84, 188, 187.999415586, 0.0 },
  { 84, 189, 188.998473425, 0.0 },
  { 84, 190, 189.995101731, 0.0 },
  { 84, 191, 190.994558494, 0.0 },
  { 84, 192, 191.991340274, 0.0 },
  { 84, 193, 192.991062421, 0.0 },
  { 84, 194, 193.988186058, 0.0 },
  { 84, 195, 194.988065781, 0.0 },
  { 84, 196, 195.985540722, 0.0 },
  { 84, 197, 196.985621939, 0.0 },
  { 84, 198, 197.983388753, 0.0 },
  { 84, 199, 198.983640445, 0.0 },
  { 84, 200, 199.981812355, 0.0 },
  { 84, 201, 200.982263799, 0.0 },
  { 84, 202, 201.980738934, 0.0 },
  { 84, 203, 202.981416072, 0.0 },
  { 84, 204, 203.980310078, 0.0 },
  { 84, 205, 204.981190006, 0.0 },
  { 84, 206, 205.980473662, 0.0 },
  { 84, 207, 206.981593334, 0.0 },
  { 84, 208, 207.981246035, 0.0 },
  { 84, 209, 208.982430361, 0.0 },
  { 84, 210, 209.982873686, 0.0 },
  { 84, 211, 210.986653171, 0.0 },
  { 84, 212, 211.988867982, 0.0 },
  { 84, 213, 212.992857154, 0.0 },
  { 84, 214, 213.995201287, 0.0 },
  { 84, 215, 214.999418385, 0.0 },
  { 84, 216, 216.001913416, 0.0 },
  { 84, 217, 217.006316145, 0.0 },
  { 84, 218, 218.008971234, 0.0 },
  { 84, 219, 219.013614000, 0.0 },
  { 84, 220, 220.016386000, 0.0 },
  { 84, 221, 221.021228000, 0.0 },
  { 84, 222, 222.024140000, 0.0 },
  { 84, 223, 223.029070,    0.0 },
  { 84, 224, 224.032110,    0.0 },
  { 84, 225, 225.037123,    0.0 },
  { 84, 226, 226.040310,    0.0 },
  { 84, 227, 227.045390,    0.0 },
} },
{ 85, "At", "Astatine", 210, 1.50, 2.5, 2.2, {
  { 85, 191, 191.004148081, 0.0 },
  { 85, 192, 192.003140912, 0.0 },
  { 85, 193, 192.999927725, 0.0 },
  { 85, 194, 193.999230816, 0.0 },
  { 85, 195, 194.996274480, 0.0 },
  { 85, 196, 195.995799034, 0.0 },
  { 85, 197, 196.993177353, 0.0 },
  { 85, 198, 197.992797864, 0.0 },
  { 85, 199, 198.990527715, 0.0 },
  { 85, 200, 199.990351099, 0.0 },
  { 85, 201, 200.988417058, 0.0 },
  { 85, 202, 201.988625686, 0.0 },
  { 85, 203, 202.986942904, 0.0 },
  { 85, 204, 203.987251393, 0.0 },
  { 85, 205, 204.986060546, 0.0 },
  { 85, 206, 205.986645768, 0.0 },
  { 85, 207, 206.985799715, 0.0 },
  { 85, 208, 207.986613011, 0.0 },
  { 85, 209, 208.986168701, 0.0 },
  { 85, 210, 209.987147423, 0.0 },
  { 85, 211, 210.987496226, 0.0 },
  { 85, 212, 211.990737301, 0.0 },
  { 85, 213, 212.992936593, 0.0 },
  { 85, 214, 213.996372331, 0.0 },
  { 85, 215, 214.998651002, 0.0 },
  { 85, 216, 216.002422643, 0.0 },
  { 85, 217, 217.004717794, 0.0 },
  { 85, 218, 218.008695941, 0.0 },
  { 85, 219, 219.011160587, 0.0 },
  { 85, 220, 220.015433000, 0.0 },
  { 85, 221, 221.018017000, 0.0 },
  { 85, 222, 222.022494000, 0.0 },
  { 85, 223, 223.025151000, 0.0 },
  { 85, 224, 224.029749000, 0.0 },
  { 85, 225, 225.032528,    0.0 },
  { 85, 226, 226.037209,    0.0 },
  { 85, 227, 227.040183,    0.0 },
  { 85, 228, 228.044960,    0.0 },
  { 85, 229, 229.048191,    0.0 },
} },
{ 86, "Rn", "Radon", 222, 1.50, 2.5, -1.0, {
  { 86, 193, 193.009707973, 0.0 },
  { 86, 194, 194.006145636, 0.0 },
  { 86, 195, 195.005421703, 0.0 },
  { 86, 196, 196.002120431, 0.0 },
  { 86, 197, 197.001621446, 0.0 },
  { 86, 198, 197.998679197, 0.0 },
  { 86, 199, 198.998325436, 0.0 },
  { 86, 200, 199.995705335, 0.0 },
  { 86, 201, 200.995590511, 0.0 },
  { 86, 202, 201.993263982, 0.0 },
  { 86, 203, 202.993361155, 0.0 },
  { 86, 204, 203.991443729, 0.0 },
  { 86, 205, 204.991723228, 0.0 },
  { 86, 206, 205.990195409, 0.0 },
  { 86, 207, 206.990730224, 0.0 },
  { 86, 208, 207.989634513, 0.0 },
  { 86, 209, 208.990401389, 0.0 },
  { 86, 210, 209.989688862, 0.0 },
  { 86, 211, 210.990600767, 0.0 },
  { 86, 212, 211.990703946, 0.0 },
  { 86, 213, 212.993885147, 0.0 },
  { 86, 214, 213.995362650, 0.0 },
  { 86, 215, 214.998745037, 0.0 },
  { 86, 216, 216.000271942, 0.0 },
  { 86, 217, 217.003927632, 0.0 },
  { 86, 218, 218.005601123, 0.0 },
  { 86, 219, 219.009478683, 0.0 },
  { 86, 220, 220.011392443, 0.0 },
  { 86, 221, 221.015535637, 0.0 },
  { 86, 222, 222.017576017, 0.0 },
  { 86, 223, 223.021889283, 0.0 },
  { 86, 224, 224.024095803, 0.0 },
  { 86, 225, 225.028485572, 0.0 },
  { 86, 226, 226.030861380, 0.0 },
  { 86, 227, 227.035304393, 0.0 },
  { 86, 228, 228.037835415, 0.0 },
  { 86, 229, 229.042257272, 0.0 },
  { 86, 230, 230.045271,    0.0 },
  { 86, 231, 231.049973,    0.0 },
} },
// 7th row; default VdW radius = 3.0 (87-88, 100-118)
{ 87, "Fr", "Francium", 223, 2.60, 3.0, 0.7, {
  { 87, 197, 197.011008086, 0.0 },
  { 87, 198, 198.010282081, 0.0 },
  { 87, 199, 199.007269384, 0.0 },
  { 87, 200, 200.006584666, 0.0 },
  { 87, 201, 201.003852491, 0.0 },
  { 87, 202, 202.003329637, 0.0 },
  { 87, 203, 203.000940867, 0.0 },
  { 87, 204, 204.000651972, 0.0 },
  { 87, 205, 204.998593854, 0.0 },
  { 87, 206, 205.998661441, 0.0 },
  { 87, 207, 206.996941450, 0.0 },
  { 87, 208, 207.997139082, 0.0 },
  { 87, 209, 208.995939701, 0.0 },
  { 87, 210, 209.996410596, 0.0 },
  { 87, 211, 210.995555189, 0.0 },
  { 87, 212, 211.996225420, 0.0 },
  { 87, 213, 212.996184410, 0.0 },
  { 87, 214, 213.998971193, 0.0 },
  { 87, 215, 215.000341534, 0.0 },
  { 87, 216, 216.003189523, 0.0 },
  { 87, 217, 217.004631980, 0.0 },
  { 87, 218, 218.007578620, 0.0 },
  { 87, 219, 219.009250664, 0.0 },
  { 87, 220, 220.012326789, 0.0 },
  { 87, 221, 221.014253714, 0.0 },
  { 87, 222, 222.017582615, 0.0 },
  { 87, 223, 223.019734241, 0.0 },
  { 87, 224, 224.023348096, 0.0 },
  { 87, 225, 225.025572466, 0.0 },
  { 87, 226, 226.029544512, 0.0 },
  { 87, 227, 227.031865413, 0.0 },
  { 87, 228, 228.035839433, 0.0 },
  { 87, 229, 229.038291443, 0.0 },
  { 87, 230, 230.042390787, 0.0 },
  { 87, 231, 231.045175353, 0.0 },
  { 87, 232, 232.049461219, 0.0 },
  { 87, 233, 233.052517833, 0.0 },
} },
{ 88, "Ra", "Radium", 226, 2.21, 3.0, 0.9, {
  { 88, 201, 201.012814699, 0.0 },
  { 88, 202, 202.009742305, 0.0 },
  { 88, 203, 203.009233907, 0.0 },
  { 88, 204, 204.006506855, 0.0 },
  { 88, 205, 205.006230692, 0.0 },
  { 88, 206, 206.003827842, 0.0 },
  { 88, 207, 207.003772420, 0.0 },
  { 88, 208, 208.001855012, 0.0 },
  { 88, 209, 209.001994902, 0.0 },
  { 88, 210, 210.000475406, 0.0 },
  { 88, 211, 211.000893049, 0.0 },
  { 88, 212, 211.999786619, 0.0 },
  { 88, 213, 213.000370971, 0.0 },
  { 88, 214, 214.000099560, 0.0 },
  { 88, 215, 215.002718208, 0.0 },
  { 88, 216, 216.003533534, 0.0 },
  { 88, 217, 217.006322676, 0.0 },
  { 88, 218, 218.007134297, 0.0 },
  { 88, 219, 219.010084715, 0.0 },
  { 88, 220, 220.011027542, 0.0 },
  { 88, 221, 221.013917293, 0.0 },
  { 88, 222, 222.015373371, 0.0 },
  { 88, 223, 223.018500648, 0.0 },
  { 88, 224, 224.020210361, 0.0 },
  { 88, 225, 225.023610502, 0.0 },
  { 88, 226, 226.025408186, 0.0 },
  { 88, 227, 227.029176205, 0.0 },
  { 88, 228, 228.031068574, 0.0 },
  { 88, 229, 229.034956703, 0.0 },
  { 88, 230, 230.037054776, 0.0 },
  { 88, 231, 231.041027085, 0.0 },
  { 88, 232, 232.043475267, 0.0 },
  { 88, 233, 233.047594570, 0.0 },
  { 88, 234, 234.050382100, 0.0 },
  { 88, 235, 235.054890,    0.0 },
} },
{ 89, "Ac", "Actinium", 227, 2.15, 2.8, 1.1, {
  { 89, 205, 205.015144152, 0.0 },
  { 89, 206, 206.014476477, 0.0 },
  { 89, 207, 207.011965967, 0.0 },
  { 89, 208, 208.011552251, 0.0 },
  { 89, 209, 209.009495375, 0.0 },
  { 89, 210, 210.009408625, 0.0 },
  { 89, 211, 211.007668846, 0.0 },
  { 89, 212, 212.007836442, 0.0 },
  { 89, 213, 213.006592665, 0.0 },
  { 89, 214, 214.006906400, 0.0 },
  { 89, 215, 215.006474061, 0.0 },
  { 89, 216, 216.008749101, 0.0 },
  { 89, 217, 217.009342325, 0.0 },
  { 89, 218, 218.011648860, 0.0 },
  { 89, 219, 219.012420425, 0.0 },
  { 89, 220, 220.014754527, 0.0 },
  { 89, 221, 221.015599721, 0.0 },
  { 89, 222, 222.017844232, 0.0 },
  { 89, 223, 223.019135982, 0.0 },
  { 89, 224, 224.021722249, 0.0 },
  { 89, 225, 225.023228601, 0.0 },
  { 89, 226, 226.026096999, 0.0 },
  { 89, 227, 227.027750594, 0.0 },
  { 89, 228, 228.031019685, 0.0 },
  { 89, 229, 229.032947000, 0.0 },
  { 89, 230, 230.036327000, 0.0 },
  { 89, 231, 231.038393000, 0.0 },
  { 89, 232, 232.042034000, 0.0 },
  { 89, 233, 233.044346000, 0.0 },
  { 89, 234, 234.048139000, 0.0 },
  { 89, 235, 235.050840000, 0.0 },
  { 89, 236, 236.054988000, 0.0 },
  { 89, 237, 237.057993,    0.0 },
} },
{ 90, "Th", "Thorium", 232.0377, 2.06, 2.93, 1.3, {
  { 90, 208, 208.017915348, 0.0    },
  { 90, 209, 209.017601,    0.0    },
  { 90, 210, 210.015093515, 0.0    },
  { 90, 211, 211.014896923, 0.0    },
  { 90, 212, 212.013001570, 0.0    },
  { 90, 213, 213.013011470, 0.0    },
  { 90, 214, 214.011481480, 0.0    },
  { 90, 215, 215.011724640, 0.0    },
  { 90, 216, 216.011055933, 0.0    },
  { 90, 217, 217.013103443, 0.0    },
  { 90, 218, 218.013276248, 0.0    },
  { 90, 219, 219.015526432, 0.0    },
  { 90, 220, 220.015769866, 0.0    },
  { 90, 221, 221.018185757, 0.0    },
  { 90, 222, 222.018468220, 0.0    },
  { 90, 223, 223.020811083, 0.0    },
  { 90, 224, 224.021466137, 0.0    },
  { 90, 225, 225.023950975, 0.0    },
  { 90, 226, 226.024903699, 0.0    },
  { 90, 227, 227.027702546, 0.0    },
  { 90, 228, 228.028739741, 0.0    },
  { 90, 229, 229.031761357, 0.0    },
  { 90, 230, 230.033132267, 0.0002 },
  { 90, 231, 231.036302764, 0.0    },
  { 90, 232, 232.038053606, 0.9998 },
  { 90, 233, 233.041580126, 0.0    },
  { 90, 234, 234.043599801, 0.0    },
  { 90, 235, 235.047255000, 0.0    },
  { 90, 236, 236.049657000, 0.0    },
  { 90, 237, 237.053629000, 0.0    },
  { 90, 238, 238.056388,    0.0    },
  { 90, 239, 239.060655,    0.0    },
} },
{ 91, "Pa", "Protactinium", 231.03588, 2.00, 2.88, 1.5, {
  { 91, 211, 211.023674036, 0.0 },
  { 91, 212, 212.023184819, 0.0 },
  { 91, 213, 213.021099644, 0.0 },
  { 91, 214, 214.020891055, 0.0 },
  { 91, 215, 215.019113955, 0.0 },
  { 91, 216, 216.019134633, 0.0 },
  { 91, 217, 217.018309024, 0.0 },
  { 91, 218, 218.020021133, 0.0 },
  { 91, 219, 219.019949909, 0.0 },
  { 91, 220, 220.021769753, 0.0 },
  { 91, 221, 221.021873393, 0.0 },
  { 91, 222, 222.023687064, 0.0 },
  { 91, 223, 223.023980414, 0.0 },
  { 91, 224, 224.025617286, 0.0 },
  { 91, 225, 225.026147927, 0.0 },
  { 91, 226, 226.027948217, 0.0 },
  { 91, 227, 227.028803586, 0.0 },
  { 91, 228, 228.031050758, 0.0 },
  { 91, 229, 229.032095585, 0.0 },
  { 91, 230, 230.034539717, 0.0 },
  { 91, 231, 231.035882500, 1.0 },
  { 91, 232, 232.038590205, 0.0 },
  { 91, 233, 233.040246535, 0.0 },
  { 91, 234, 234.043305555, 0.0 },
  { 91, 235, 235.045399000, 0.0 },
  { 91, 236, 236.048668000, 0.0 },
  { 91, 237, 237.051023000, 0.0 },
  { 91, 238, 238.054637000, 0.0 },
  { 91, 239, 239.057260,    0.0 },
  { 91, 240, 240.061203,    0.0 },
  { 91, 241, 241.064134,    0.0 },
} },
{ 92, "U", "Uranium", 238.02891, 1.96, 2.71, 1.38, {
  { 92, 215, 215.026719774, 0.0      },
  { 92, 216, 216.024762829, 0.0      },
  { 92, 217, 217.024660,    0.0      },
  { 92, 218, 218.023504877, 0.0      },
  { 92, 219, 219.025009233, 0.0      },
  { 92, 220, 220.024706,    0.0      },
  { 92, 221, 221.026323297, 0.0      },
  { 92, 222, 222.026057957, 0.0      },
  { 92, 223, 223.027960754, 0.0      },
  { 92, 224, 224.027635913, 0.0      },
  { 92, 225, 225.029385050, 0.0      },
  { 92, 226, 226.029338669, 0.0      },
  { 92, 227, 227.031181124, 0.0      },
  { 92, 228, 228.031368959, 0.0      },
  { 92, 229, 229.033505976, 0.0      },
  { 92, 230, 230.033940114, 0.0      },
  { 92, 231, 231.036292180, 0.0      },
  { 92, 232, 232.037154765, 0.0      },
  { 92, 233, 233.039634294, 0.0      },
  { 92, 234, 234.040950296, 0.000054 },
  { 92, 235, 235.043928117, 0.007204 },
  { 92, 236, 236.045566130, 0.0      },
  { 92, 237, 237.048728309, 0.0      },
  { 92, 238, 238.050786936, 0.992742 },
  { 92, 239, 239.054291989, 0.0      },
  { 92, 240, 240.056592411, 0.0      },
  { 92, 241, 241.060330,    0.0      },
  { 92, 242, 242.062931,    0.0      },
  { 92, 243, 243.067075,    0.0      },
} },
{ 93, "Np", "Neptunium", 237, 1.90, 2.82, 1.36, {
  { 93, 219, 219.031601865, 0.0 },
  { 93, 220, 220.032716280, 0.0 },
  { 93, 221, 221.032110,    0.0 },
  { 93, 222, 222.033574706, 0.0 },
  { 93, 223, 223.032913340, 0.0 },
  { 93, 224, 224.034388030, 0.0 },
  { 93, 225, 225.033943422, 0.0 },
  { 93, 226, 226.035230364, 0.0 },
  { 93, 227, 227.034975012, 0.0 },
  { 93, 228, 228.036313,    0.0 },
  { 93, 229, 229.036287269, 0.0 },
  { 93, 230, 230.037828060, 0.0 },
  { 93, 231, 231.038243598, 0.0 },
  { 93, 232, 232.040107,    0.0 },
  { 93, 233, 233.040739421, 0.0 },
  { 93, 234, 234.042893245, 0.0 },
  { 93, 235, 235.044061518, 0.0 },
  { 93, 236, 236.046568296, 0.0 },
  { 93, 237, 237.048171640, 0.0 },
  { 93, 238, 238.050944603, 0.0 },
  { 93, 239, 239.052937538, 0.0 },
  { 93, 240, 240.056163778, 0.0 },
  { 93, 241, 241.058309671, 0.0 },
  { 93, 242, 242.061639548, 0.0 },
  { 93, 243, 243.064204,    0.0 },
  { 93, 244, 244.067891,    0.0 },
  { 93, 245, 245.070693,    0.0 },
} },
{ 94, "Pu", "Plutonium", 244, 1.87, 2.81, 1.28, {
  { 94, 221, 221.038572,    0.0 },
  { 94, 222, 222.037638,    0.0 },
  { 94, 223, 223.038777,    0.0 },
  { 94, 224, 224.037875,    0.0 },
  { 94, 225, 225.038970,    0.0 },
  { 94, 226, 226.038250,    0.0 },
  { 94, 227, 227.039474,    0.0 },
  { 94, 228, 228.038763325, 0.0 },
  { 94, 229, 229.040145099, 0.0 },
  { 94, 230, 230.039648313, 0.0 },
  { 94, 231, 231.041125946, 0.0 },
  { 94, 232, 232.041182133, 0.0 },
  { 94, 233, 233.042997411, 0.0 },
  { 94, 234, 234.043317489, 0.0 },
  { 94, 235, 235.045284609, 0.0 },
  { 94, 236, 236.046056661, 0.0 },
  { 94, 237, 237.048407888, 0.0 },
  { 94, 238, 238.049558175, 0.0 },
  { 94, 239, 239.052161596, 0.0 },
  { 94, 240, 240.053811740, 0.0 },
  { 94, 241, 241.056849651, 0.0 },
  { 94, 242, 242.058740979, 0.0 },
  { 94, 243, 243.062002068, 0.0 },
  { 94, 244, 244.064204401, 0.0 },
  { 94, 245, 245.067824554, 0.0 },
  { 94, 246, 246.070204172, 0.0 },
  { 94, 247, 247.074300,    0.0 },
} },
{ 95, "Am", "Americium", 243, 1.80, 2.83, 1.3, {
  { 95, 223, 223.045840,    0.0 },
  { 95, 224, 224.046442,    0.0 },
  { 95, 225, 225.045508,    0.0 },
  { 95, 226, 226.046130,    0.0 },
  { 95, 227, 227.045282,    0.0 },
  { 95, 228, 228.046001,    0.0 },
  { 95, 229, 229.045282534, 0.0 },
  { 95, 230, 230.046025,    0.0 },
  { 95, 231, 231.045529,    0.0 },
  { 95, 232, 232.046613,    0.0 },
  { 95, 233, 233.046468,    0.0 },
  { 95, 234, 234.047731,    0.0 },
  { 95, 235, 235.047906478, 0.0 },
  { 95, 236, 236.049427,    0.0 },
  { 95, 237, 237.049995,    0.0 },
  { 95, 238, 238.051982531, 0.0 },
  { 95, 239, 239.053022729, 0.0 },
  { 95, 240, 240.055298374, 0.0 },
  { 95, 241, 241.056827343, 0.0 },
  { 95, 242, 242.059547358, 0.0 },
  { 95, 243, 243.061379889, 0.0 },
  { 95, 244, 244.064282892, 0.0 },
  { 95, 245, 245.066452827, 0.0 },
  { 95, 246, 246.069774,    0.0 },
  { 95, 247, 247.072092,    0.0 },
  { 95, 248, 248.075752,    0.0 },
  { 95, 249, 249.078480,    0.0 },
} },
{ 96, "Cm", "Curium", 247, 1.69, 3.05, 1.3, {
  { 96, 231, 231.050746,    0.0 },
  { 96, 232, 232.049740,    0.0 },
  { 96, 233, 233.050771485, 0.0 },
  { 96, 234, 234.050158568, 0.0 },
  { 96, 235, 235.051545,    0.0 },
  { 96, 236, 236.051372112, 0.0 },
  { 96, 237, 237.052868988, 0.0 },
  { 96, 238, 238.053081606, 0.0 },
  { 96, 239, 239.054908519, 0.0 },
  { 96, 240, 240.055528233, 0.0 },
  { 96, 241, 241.057651218, 0.0 },
  { 96, 242, 242.058834187, 0.0 },
  { 96, 243, 243.061387329, 0.0 },
  { 96, 244, 244.062750622, 0.0 },
  { 96, 245, 245.065491047, 0.0 },
  { 96, 246, 246.067222016, 0.0 },
  { 96, 247, 247.070352678, 0.0 },
  { 96, 248, 248.072349086, 0.0 },
  { 96, 249, 249.075953992, 0.0 },
  { 96, 250, 250.078357541, 0.0 },
  { 96, 251, 251.082284988, 0.0 },
  { 96, 252, 252.084870,    0.0 },
} },
{ 97, "Bk", "Berkelium", 247, 1.68, 3.4, 1.3, {
  { 97, 233, 233.056652,    0.0 },
  { 97, 234, 234.057322,    0.0 },
  { 97, 235, 235.056651,    0.0 },
  { 97, 236, 236.057479,    0.0 },
  { 97, 237, 237.057123,    0.0 },
  { 97, 238, 238.058204,    0.0 },
  { 97, 239, 239.058239,    0.0 },
  { 97, 240, 240.059758,    0.0 },
  { 97, 241, 241.060098,    0.0 },
  { 97, 242, 242.061999,    0.0 },
  { 97, 243, 243.063005905, 0.0 },
  { 97, 244, 244.065178969, 0.0 },
  { 97, 245, 245.066359814, 0.0 },
  { 97, 246, 246.068671300, 0.0 },
  { 97, 247, 247.070305889, 0.0 },
  { 97, 248, 248.073141689, 0.0 },
  { 97, 249, 249.074983118, 0.0 },
  { 97, 250, 250.078317195, 0.0 },
  { 97, 251, 251.080760555, 0.0 },
  { 97, 252, 252.084310,    0.0 },
  { 97, 253, 253.086880,    0.0 },
  { 97, 254, 254.090600,    0.0 },
} },
{ 98, "Cf", "Californium", 251, 1.68, 3.05, 1.3, {
  { 98, 237, 237.062199272, 0.0 },
  { 98, 238, 238.061490,    0.0 },
  { 98, 239, 239.062482,    0.0 },
  { 98, 240, 240.062253447, 0.0 },
  { 98, 241, 241.063690,    0.0 },
  { 98, 242, 242.063754544, 0.0 },
  { 98, 243, 243.065475,    0.0 },
  { 98, 244, 244.065999447, 0.0 },
  { 98, 245, 245.068046755, 0.0 },
  { 98, 246, 246.068803685, 0.0 },
  { 98, 247, 247.070971348, 0.0 },
  { 98, 248, 248.072182905, 0.0 },
  { 98, 249, 249.074850428, 0.0 },
  { 98, 250, 250.076404494, 0.0 },
  { 98, 251, 251.079587171, 0.0 },
  { 98, 252, 252.081626507, 0.0 },
  { 98, 253, 253.085133723, 0.0 },
  { 98, 254, 254.087323575, 0.0 },
  { 98, 255, 255.091046,    0.0 },
  { 98, 256, 256.093442,    0.0 },
} },
{ 99, "Es", "Einsteinium", 252, 1.65, 2.7, 1.3, {
  { 99, 239, 239.068310,    0.0 },
  { 99, 240, 240.068949,    0.0 },
  { 99, 241, 241.068592,    0.0 },
  { 99, 242, 242.069567,    0.0 },
  { 99, 243, 243.069508,    0.0 },
  { 99, 244, 244.070881,    0.0 },
  { 99, 245, 245.071192,    0.0 },
  { 99, 246, 246.072806474, 0.0 },
  { 99, 247, 247.073621929, 0.0 },
  { 99, 248, 248.075469,    0.0 },
  { 99, 249, 249.076409,    0.0 },
  { 99, 250, 250.078611,    0.0 },
  { 99, 251, 251.079991431, 0.0 },
  { 99, 252, 252.082979173, 0.0 },
  { 99, 253, 253.084821241, 0.0 },
  { 99, 254, 254.088024337, 0.0 },
  { 99, 255, 255.090273504, 0.0 },
  { 99, 256, 256.093597,    0.0 },
  { 99, 257, 257.095979,    0.0 },
  { 99, 258, 258.099520,    0.0 },
} },
{ 100, "Fm", "Fermium", 257, 1.67, 3.0, 1.3, {
  { 100, 241, 241.074311,    0.0 },
  { 100, 242, 242.073430,    0.0 },
  { 100, 243, 243.074414,    0.0 },
  { 100, 244, 244.074036,    0.0 },
  { 100, 245, 245.075354,    0.0 },
  { 100, 246, 246.075353334, 0.0 },
  { 100, 247, 247.076944,    0.0 },
  { 100, 248, 248.077185451, 0.0 },
  { 100, 249, 249.078926042, 0.0 },
  { 100, 250, 250.079519765, 0.0 },
  { 100, 251, 251.081545130, 0.0 },
  { 100, 252, 252.082466019, 0.0 },
  { 100, 253, 253.085180945, 0.0 },
  { 100, 254, 254.086852424, 0.0 },
  { 100, 255, 255.089963495, 0.0 },
  { 100, 256, 256.091771699, 0.0 },
  { 100, 257, 257.095105419, 0.0 },
  { 100, 258, 258.097077,    0.0 },
  { 100, 259, 259.100596,    0.0 },
  { 100, 260, 260.102809,    0.0 },
} },
{ 101, "Md", "Mendelevium", 258, 1.73, 3.0, 1.3, {
  { 101, 244, 244.081157,    0.0 },
  { 101, 245, 245.080864,    0.0 },
  { 101, 246, 246.081713,    0.0 },
  { 101, 247, 247.081520,    0.0 },
  { 101, 248, 248.082607,    0.0 },
  { 101, 249, 249.082857155, 0.0 },
  { 101, 250, 250.084164934, 0.0 },
  { 101, 251, 251.084774287, 0.0 },
  { 101, 252, 252.086385000, 0.0 },
  { 101, 253, 253.087143,    0.0 },
  { 101, 254, 254.089590,    0.0 },
  { 101, 255, 255.091081702, 0.0 },
  { 101, 256, 256.093888,    0.0 },
  { 101, 257, 257.095537343, 0.0 },
  { 101, 258, 258.098433634, 0.0 },
  { 101, 259, 259.100445,    0.0 },
  { 101, 260, 260.103650,    0.0 },
  { 101, 261, 261.105828,    0.0 },
  { 101, 262, 262.109144,    0.0 },
} },
{ 102, "No", "Nobelium", 259, 1.76, 3.0, 1.3, {
  { 102, 248, 248.086623,    0.0 },
  { 102, 249, 249.087802,    0.0 },
  { 102, 250, 250.087565,    0.0 },
  { 102, 251, 251.088942,    0.0 },
  { 102, 252, 252.088966070, 0.0 },
  { 102, 253, 253.090562780, 0.0 },
  { 102, 254, 254.090954211, 0.0 },
  { 102, 255, 255.093196439, 0.0 },
  { 102, 256, 256.094281912, 0.0 },
  { 102, 257, 257.096884203, 0.0 },
  { 102, 258, 258.098205,    0.0 },
  { 102, 259, 259.100998364, 0.0 },
  { 102, 260, 260.102641,    0.0 },
  { 102, 261, 261.105696,    0.0 },
  { 102, 262, 262.107463,    0.0 },
  { 102, 263, 263.110714,    0.0 },
  { 102, 264, 264.112734,    0.0 },
} },
{ 103, "Lr", "Lawrencium", 266, 1.61, 3.0, -1.0, {
  { 103, 251, 251.094289,    0.0 },
  { 103, 252, 252.095048,    0.0 },
  { 103, 253, 253.095033850, 0.0 },
  { 103, 254, 254.096238813, 0.0 },
  { 103, 255, 255.096562399, 0.0 },
  { 103, 256, 256.098494024, 0.0 },
  { 103, 257, 257.099480,    0.0 },
  { 103, 258, 258.101753,    0.0 },
  { 103, 259, 259.102900,    0.0 },
  { 103, 260, 260.105504,    0.0 },
  { 103, 261, 261.106879,    0.0 },
  { 103, 262, 262.109615,    0.0 },
  { 103, 263, 263.111293,    0.0 },
  { 103, 264, 264.114198,    0.0 },
  { 103, 265, 265.116193,    0.0 },
  { 103, 266, 266.119874,    0.0 },
} },
{ 104, "Rf", "Rutherfordium", 267, 1.57, 3.0, -1.0, {
  { 104, 253, 253.100528,    0.0 },
  { 104, 254, 254.100055,    0.0 },
  { 104, 255, 255.101267,    0.0 },
  { 104, 256, 256.101151464, 0.0 },
  { 104, 257, 257.102916796, 0.0 },
  { 104, 258, 258.103429895, 0.0 },
  { 104, 259, 259.105601,    0.0 },
  { 104, 260, 260.106440,    0.0 },
  { 104, 261, 261.108769591, 0.0 },
  { 104, 262, 262.109923,    0.0 },
  { 104, 263, 263.112461,    0.0 },
  { 104, 264, 264.113876,    0.0 },
  { 104, 265, 265.116683,    0.0 },
  { 104, 266, 266.118236,    0.0 },
  { 104, 267, 267.121787,    0.0 },
  { 104, 268, 268.123968,    0.0 },
} },
{ 105, "Db", "Dubnium", 268, 1.49, 3.0, -1.0, {
  { 105, 255, 255.106919,    0.0 },
  { 105, 256, 256.107674,    0.0 },
  { 105, 257, 257.107520042, 0.0 },
  { 105, 258, 258.108972995, 0.0 },
  { 105, 259, 259.109491859, 0.0 },
  { 105, 260, 260.111297,    0.0 },
  { 105, 261, 261.111979,    0.0 },
  { 105, 262, 262.114067,    0.0 },
  { 105, 263, 263.114987,    0.0 },
  { 105, 264, 264.117297,    0.0 },
  { 105, 265, 265.118500,    0.0 },
  { 105, 266, 266.121032,    0.0 },
  { 105, 267, 267.122399,    0.0 },
  { 105, 268, 268.125669,    0.0 },
  { 105, 269, 269.127911,    0.0 },
  { 105, 270, 270.131399,    0.0 },
} },
{ 106, "Sg", "Seaborgium", 269, 1.43, 3.0, -1.0, {
  { 106, 258, 258.113040,    0.0 },
  { 106, 259, 259.114353,    0.0 },
  { 106, 260, 260.114383435, 0.0 },
  { 106, 261, 261.115948135, 0.0 },
  { 106, 262, 262.116338978, 0.0 },
  { 106, 263, 263.118299,    0.0 },
  { 106, 264, 264.118930,    0.0 },
  { 106, 265, 265.121089,    0.0 },
  { 106, 266, 266.121973,    0.0 },
  { 106, 267, 267.124323,    0.0 },
  { 106, 268, 268.125389,    0.0 },
  { 106, 269, 269.128495,    0.0 },
  { 106, 270, 270.130362,    0.0 },
  { 106, 271, 271.133782,    0.0 },
  { 106, 272, 272.135825,    0.0 },
  { 106, 273, 273.139475,    0.0 },
} },
{ 107, "Bh", "Bohrium", 270, 1.41, 3.0, -1.0, {
  { 107, 260, 260.121443,    0.0 },
  { 107, 261, 261.121395733, 0.0 },
  { 107, 262, 262.122654688, 0.0 },
  { 107, 263, 263.122916,    0.0 },
  { 107, 264, 264.124486,    0.0 },
  { 107, 265, 265.124955,    0.0 },
  { 107, 266, 266.126790,    0.0 },
  { 107, 267, 267.127499,    0.0 },
  { 107, 268, 268.129584,    0.0 },
  { 107, 269, 269.130411,    0.0 },
  { 107, 270, 270.133366,    0.0 },
  { 107, 271, 271.135115,    0.0 },
  { 107, 272, 272.138259,    0.0 },
  { 107, 273, 273.140294,    0.0 },
  { 107, 274, 274.143599,    0.0 },
  { 107, 275, 275.145766,    0.0 },
  { 107, 276, 276.149169,    0.0 },
  { 107, 277, 277.151477,    0.0 },
  { 107, 278, 278.154988,    0.0 },
} },
{ 108, "Hs", "Hassium", 269, 1.34, 3.0, -1.0, {
  { 108, 263, 263.128479,    0.0 },
  { 108, 264, 264.128356330, 0.0 },
  { 108, 265, 265.129791744, 0.0 },
  { 108, 266, 266.130048783, 0.0 },
  { 108, 267, 267.131678,    0.0 },
  { 108, 268, 268.132011,    0.0 },
  { 108, 269, 269.133649,    0.0 },
  { 108, 270, 270.134313,    0.0 },
  { 108, 271, 271.137082,    0.0 },
  { 108, 272, 272.138492,    0.0 },
  { 108, 273, 273.141458,    0.0 },
  { 108, 274, 274.143217,    0.0 },
  { 108, 275, 275.146530,    0.0 },
  { 108, 276, 276.148348,    0.0 },
  { 108, 277, 277.151772,    0.0 },
  { 108, 278, 278.153753,    0.0 },
  { 108, 279, 279.157274,    0.0 },
  { 108, 280, 280.159335,    0.0 },
} },
{ 109, "Mt", "Meitnerium", 277, 1.29, 3.0, -1.0, {
  { 109, 265, 265.135937,    0.0 },
  { 109, 266, 266.137062253, 0.0 },
  { 109, 267, 267.137189,    0.0 },
  { 109, 268, 268.138649,    0.0 },
  { 109, 269, 269.138809,    0.0 },
  { 109, 270, 270.140322,    0.0 },
  { 109, 271, 271.140741,    0.0 },
  { 109, 272, 272.143298,    0.0 },
  { 109, 273, 273.144695,    0.0 },
  { 109, 274, 274.147343,    0.0 },
  { 109, 275, 275.148972,    0.0 },
  { 109, 276, 276.151705,    0.0 },
  { 109, 277, 277.153525,    0.0 },
  { 109, 278, 278.156487,    0.0 },
  { 109, 279, 279.158439,    0.0 },
  { 109, 280, 280.161579,    0.0 },
  { 109, 281, 281.163608,    0.0 },
  { 109, 282, 282.166888,    0.0 },
} },
{ 110, "Ds", "Darmstadtium", 281, 1.28, 3.0, -1.0, {
  { 110, 267, 267.143726,    0.0 },
  { 110, 268, 268.143477,    0.0 },
  { 110, 269, 269.144750965, 0.0 },
  { 110, 270, 270.144586620, 0.0 },
  { 110, 271, 271.145951,    0.0 },
  { 110, 272, 272.146091,    0.0 },
  { 110, 273, 273.148455,    0.0 },
  { 110, 274, 274.149434,    0.0 },
  { 110, 275, 275.152085,    0.0 },
  { 110, 276, 276.153022,    0.0 },
  { 110, 277, 277.155763,    0.0 },
  { 110, 278, 278.157007,    0.0 },
  { 110, 279, 279.159984,    0.0 },
  { 110, 280, 280.161375,    0.0 },
  { 110, 281, 281.164545,    0.0 },
  { 110, 282, 282.166174,    0.0 },
  { 110, 283, 283.169437,    0.0 },
  { 110, 284, 284.171187,    0.0 },
} },
{ 111, "Rg", "Roentgenium", 282, 1.21, 3.0, -1.0, {
  { 111, 272, 272.153273, 0.0 },
  { 111, 273, 273.153393, 0.0 },
  { 111, 274, 274.155247, 0.0 },
  { 111, 275, 275.156088, 0.0 },
  { 111, 276, 276.158226, 0.0 },
  { 111, 277, 277.159322, 0.0 },
  { 111, 278, 278.161590, 0.0 },
  { 111, 279, 279.162880, 0.0 },
  { 111, 280, 280.165204, 0.0 },
  { 111, 281, 281.166757, 0.0 },
  { 111, 282, 282.169343, 0.0 },
  { 111, 283, 283.171101, 0.0 },
  { 111, 284, 284.173882, 0.0 },
  { 111, 285, 285.175771, 0.0 },
  { 111, 286, 286.178756, 0.0 },
} },
{ 112, "Cn", "Copernicium", 285, 1.22, 3.0, -1.0, {
  { 112, 276, 276.161418, 0.0 },
  { 112, 277, 277.163535, 0.0 },
  { 112, 278, 278.164083, 0.0 },
  { 112, 279, 279.166422, 0.0 },
  { 112, 280, 280.167102, 0.0 },
  { 112, 281, 281.169563, 0.0 },
  { 112, 282, 282.170507, 0.0 },
  { 112, 283, 283.173202, 0.0 },
  { 112, 284, 284.174360, 0.0 },
  { 112, 285, 285.177227, 0.0 },
  { 112, 286, 286.178691, 0.0 },
  { 112, 287, 287.181826, 0.0 },
  { 112, 288, 288.183501, 0.0 },
} },
{ 113, "Nh", "Nihonium", 286, 1.36, 3.0, -1.0, {
  { 113, 278, 278.170725, 0.0 },
  { 113, 279, 279.171187, 0.0 },
  { 113, 280, 280.173098, 0.0 },
  { 113, 281, 281.173710, 0.0 },
  { 113, 282, 282.175770, 0.0 },
  { 113, 283, 283.176666, 0.0 },
  { 113, 284, 284.178843, 0.0 },
  { 113, 285, 285.180106, 0.0 },
  { 113, 286, 286.182456, 0.0 },
  { 113, 287, 287.184064, 0.0 },
  { 113, 288, 288.186764, 0.0 },
  { 113, 289, 289.188461, 0.0 },
  { 113, 290, 290.191429, 0.0 },
} },
{ 114, "Fl", "Flerovium", 290, 1.43, 3.0, -1.0, {
  { 114, 284, 284.181192, 0.0 },
  { 114, 285, 285.183503, 0.0 },
  { 114, 286, 286.184226, 0.0 },
  { 114, 287, 287.186720, 0.0 },
  { 114, 288, 288.187781, 0.0 },
  { 114, 289, 289.190517, 0.0 },
  { 114, 290, 290.191875, 0.0 },
  { 114, 291, 291.194848, 0.0 },
} },
{ 115, "Mc", "Moscovium", 290, 1.62, 3.0, -1.0, {
  { 115, 287, 287.190820, 0.0 },
  { 115, 288, 288.192879, 0.0 },
  { 115, 289, 289.193971, 0.0 },
  { 115, 290, 290.196235, 0.0 },
  { 115, 291, 291.197725, 0.0 },
  { 115, 292, 292.200323, 0.0 },
} },
{ 116, "Lv", "Livermorium", 293, 1.75, 3.0, -1.0, {
  { 116, 289, 289.198023, 0.0 },
  { 116, 290, 290.198635, 0.0 },
  { 116, 291, 291.201014, 0.0 },
  { 116, 292, 292.201969, 0.0 },
  { 116, 293, 293.204583, 0.0 },
} },
{ 117, "Ts", "Tennessine", 294, 1.65, 3.0, -1.0, {
  { 117, 291, 291.205748, 0.0 },
  { 117, 292, 292.207861, 0.0 },
  { 117, 293, 293.208727, 0.0 },
  { 117, 294, 294.210840, 0.0 },
} },
{ 118, "Og", "Oganesson", 294, 1.57, 3.0, -1.0, {
  { 118, 293, 293.213423, 0.0 },
  { 118, 294, 294.213979, 0.0 },
  { 118, 295, 295.216178, 0.0 },
} }, },
  // clang-format on
  // GCOV_EXCL_START
    symbol_to_element_(static_cast<size_t>(kElementCount_ + 17) * 3),
    name_to_element_(static_cast<size_t>(kElementCount_ + 15) * 3) {
  // GCOV_EXCL_STOP
  // Canonical symbols and names.
  for (const Element &element: elements_) {
    symbol_to_element_[element.symbol()] = name_to_element_[element.name()] =
        &element;

    if (element.symbol().size() > 1) {
      absl::StrAppend(&symb_name_buf_, absl::AsciiStrToUpper(element.symbol()));
    }

    absl::StrAppend(&symb_name_buf_, absl::AsciiStrToLower(element.symbol()),
                    absl::AsciiStrToUpper(element.name()),
                    absl::AsciiStrToLower(element.name()));
  }

  symb_name_buf_.shrink_to_fit();

  std::string_view buffer = symb_name_buf_;
  size_t pos = 0;
  for (const Element &element: elements_) {
    size_t len = element.symbol().size();
    if (element.symbol().size() > 1) {
      symbol_to_element_[buffer.substr(pos, len)] = &element;
      pos += len;
    }
    symbol_to_element_[buffer.substr(pos, len)] = &element;
    pos += len;

    len = element.name().size();
    name_to_element_[buffer.substr(pos, len)] = &element;
    pos += len;
    name_to_element_[buffer.substr(pos, len)] = &element;
    pos += len;
  }

  ABSL_DCHECK(pos == symb_name_buf_.size());

// NOLINTNEXTLINE(*-macro-usage)
#define NURI_ELEMENT_ADD_EXTRA(number, s1, s2, s3, n1, n2, n3)                 \
  symbol_to_element_[#s1] = symbol_to_element_[#s2] =                          \
      symbol_to_element_[#s3] = name_to_element_[#n1] =                        \
          name_to_element_[#n2] = name_to_element_[#n3] = &elements_[number]

  // Non-canonical symbols for convenience.

  // RDKit use "*" and some others use "X", "Du" for dummy atom.
  NURI_ELEMENT_ADD_EXTRA(0, *, X, x, Du, DU, du);

  // Systematic element symbols & names (104-118).

  NURI_ELEMENT_ADD_EXTRA(104,            //
                         Unq, UNQ, unq,  //
                         Unnilquadium, UNNILQUADIUM, unnilquadium);
  NURI_ELEMENT_ADD_EXTRA(105,            //
                         Unp, UNP, unp,  //
                         Unnilpentium, UNNILPENTIUM, unnilpentium);
  NURI_ELEMENT_ADD_EXTRA(106,            //
                         Unh, UNH, unh,  //
                         Unnilhexium, UNNILHEXIUM, unnilhexium);
  NURI_ELEMENT_ADD_EXTRA(107,            //
                         Uns, UNS, uns,  //
                         Unnilseptium, UNNILSEPTIUM, unnilseptium);
  NURI_ELEMENT_ADD_EXTRA(108,            //
                         Uno, UNO, uno,  //
                         Unniloctium, UNNILOCTIUM, unniloctium);
  NURI_ELEMENT_ADD_EXTRA(109,            //
                         Une, UNE, une,  //
                         Unnilennium, UNNILENNIUM, unnilennium);
  NURI_ELEMENT_ADD_EXTRA(110,            //
                         Uun, UUN, uun,  //
                         Ununnilium, UNUNNILIUM, ununnilium);
  NURI_ELEMENT_ADD_EXTRA(111,            //
                         Uuu, UUU, uuu,  //
                         Unununium, UNUNUNIUM, unununium);
  NURI_ELEMENT_ADD_EXTRA(112,            //
                         Uub, UUB, uub,  //
                         Ununbium, UNUNBIUM, ununbium);
  NURI_ELEMENT_ADD_EXTRA(113,            //
                         Uut, UUT, uut,  //
                         Ununtrium, UNUNTRIUM, ununtrium);
  NURI_ELEMENT_ADD_EXTRA(114,            //
                         Uuq, UUQ, uuq,  //
                         Ununquadium, UNUNQUADIUM, ununquadium);
  NURI_ELEMENT_ADD_EXTRA(115,            //
                         Uup, UUP, uup,  //
                         Ununpentium, UNUNPENTIUM, ununpentium);
  NURI_ELEMENT_ADD_EXTRA(116,            //
                         Uuh, UUH, uuh,  //
                         Ununhexium, UNUNHEXIUM, ununhexium);
  NURI_ELEMENT_ADD_EXTRA(117,            //
                         Uus, UUS, uus,  //
                         Ununseptium, UNUNSEPTIUM, ununseptium);
  NURI_ELEMENT_ADD_EXTRA(118,            //
                         Uuo, UUO, uuo,  //
                         Ununoctium, UNUNOCTIUM, ununoctium);
#undef NURI_ELEMENT_ADD_EXTRA

  symbol_to_element_.rehash(0);
  name_to_element_.rehash(0);
}
}  // namespace nuri
